/***********************************************************/
/*                                                         */
/* D3DStructures.h                                         */
/*                                                         */
/* created 20.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef D3DSTRUCTURES_H
#define D3DSTRUCTURES_H

#include <d3d9.h>

#include "MathStructures.h"
#include "DynamicArray.h"

// www.infinitecode.com 3d library
namespace I3D
{
	using IMATH::Vector;
	using IMATH::Vector2D;

#ifndef RELEASE
	#define RELEASE(x) if(x){(x)->Release();}
#endif

#define I3D_FVFVERTEX (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1)
#define I3D_FVFLVERTEX (D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1)
#define I3D_FVFVERTEXTEX2 (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX2)

class Vertex
{
public:
	union
	{
		struct
		{
			float x, y, z, nx, ny, nz, tu, tv;
		};
		struct
		{
			Vector v, n;
			Vector2D t;
		};
		float Data[8];
	};

	Vertex() {}

	Vertex(const Vector &Vec, const Vector &Norm, const Vector2D &Tex)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		nx = Norm.x;
		ny = Norm.y;
		nz = Norm.z;
		tu = Tex.x;
		tv = Tex.y;
	}

	void Set(const Vector &Vec, const Vector &Norm, const Vector2D &Tex)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		nx = Norm.x;
		ny = Norm.y;
		nz = Norm.z;
		tu = Tex.x;
		tv = Tex.y;
	}
};

class LVertex
{
public:
	union
	{
		struct
		{
			float x, y, z;
			unsigned int c; // RGBA color
			float tu, tv;
		};
		struct
		{
			Vector v;
			unsigned int c; // RGBA color
			Vector2D t;
		};
		float Data[8];
	};

	LVertex() {}

	LVertex(const Vector &Vec, unsigned int Color, const Vector2D &Tex)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		c = Color;
		tu = Tex.x;
		tv = Tex.y;
	}

	void Set(const Vector &Vec, unsigned int Color, const Vector2D &Tex)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		c = Color;
		tu = Tex.x;
		tv = Tex.y;
	}
};

class VertexTex2
{
public:
	union
	{
		struct
		{
			float x, y, z, nx, ny, nz, tu0, tv0, tu1, tv1;
		};
		struct
		{
			Vector v, n;
			Vector2D t0, t1;
		};
		float Data[10];
	};

	VertexTex2() {}

	VertexTex2(const Vector &Vec, const Vector &Norm, const Vector2D &Tex0, const Vector2D &Tex1)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		nx = Norm.x;
		ny = Norm.y;
		nz = Norm.z;
		tu0 = Tex0.x;
		tv0 = Tex0.y;
		tu1 = Tex1.x;
		tv1 = Tex1.y;
	}

	void Set(const Vector &Vec, const Vector &Norm, const Vector2D &Tex0, const Vector2D &Tex1)
	{
		x = Vec.x;
		y = Vec.y;
		z = Vec.z;
		nx = Norm.x;
		ny = Norm.y;
		nz = Norm.z;
		tu0 = Tex0.x;
		tv0 = Tex0.y;
		tu1 = Tex1.x;
		tv1 = Tex1.y;
	}
};

class Matrix: public D3DMATRIX
{
	friend inline Vector &operator *=(Vector &v, const Matrix &m);

public:
	Matrix(void) {}

	Matrix(float x1, float x2, float x3, float x4
		, float y1, float y2, float y3, float y4
		, float z1, float z2, float z3, float z4
		, float a1, float a2, float a3, float a4)
	{
		m[0][0] = x1;
		m[0][1] = x2;
		m[0][2] = x3;
		m[0][3] = x4;
		m[1][0] = y1;
		m[1][1] = y2;
		m[1][2] = y3;
		m[1][3] = y4;
		m[2][0] = z1;
		m[2][1] = z2;
		m[2][2] = z3;
		m[2][3] = z4;
		m[3][0] = a1;
		m[3][1] = a2;
		m[3][2] = a3;
		m[3][3] = a4;
	}

	void Set(float x1, float x2, float x3, float x4
		, float y1, float y2, float y3, float y4
		, float z1, float z2, float z3, float z4
		, float a1, float a2, float a3, float a4)
	{
		m[0][0] = x1;
		m[0][1] = x2;
		m[0][2] = x3;
		m[0][3] = x4;
		m[1][0] = y1;
		m[1][1] = y2;
		m[1][2] = y3;
		m[1][3] = y4;
		m[2][0] = z1;
		m[2][1] = z2;
		m[2][2] = z3;
		m[2][3] = z4;
		m[3][0] = a1;
		m[3][1] = a2;
		m[3][2] = a3;
		m[3][3] = a4;
	}

	Vector operator * (const Vector &V) const
	{
		Vector r;
		// apply
		r.x = V.x * m[0][0] + V.y * m[1][0] + V.z * m[2][0] + m[3][0];
		r.y = V.x * m[0][1] + V.y * m[1][1] + V.z * m[2][1] + m[3][1];
		r.z = V.x * m[0][2] + V.y * m[1][2] + V.z * m[2][2] + m[3][2];
		// scale
		float s = m[3][3];
		s += V.x * m[0][3];
		s += V.y * m[1][3];
		s += V.z * m[2][3];
		if(s != 0.0f){
			s = 1/s;
			r.x *= s;
			r.y *= s;
			r.z *= s;
		}
		return r;
	}

	Matrix operator *= (const Matrix &M)
	{
		*this = M * *this;
		return *this;
	}

	Matrix operator * (const Matrix &M2) const
	{
		Matrix Ret;
		//Ret.m[0][0] = m[0][0] * M2.m[0][0] + 
		for(int i = 0; i < 4; i++)
			for(int j = 0; j < 4; j++)
			{
				Ret.m[i][j] = 0;
				for(int k = 0; k < 4; k++)
					Ret.m[i][j] += m[k][j] * M2.m[i][k];
			}
		return Ret;
	}

/*	Matrix Invert(void) const
	{
		Matrix Ori = *this;
		Matrix Ret;
		Ret.SetIdentityMatrix();
		int j;
		// make triangular matrix
		// diagonal
		for(j = 0; j < 4; j++)
			if(Ori.m[j][j] != 0)
			{
				// create a 0
				float Mul = 1.0f / Ori.m[j][j];
				int i;
				// for columns
				for(i = 0; i < 4; i++)
				{
					Ori.m[j][i] *= Mul;
					Ret.m[j][i] *= Mul;
				}
				// annihilate the other first numbers in the same column
				if(j < 3)
					// for rows
					for(i = j + 1; i < 4; i++)
					{
						float First = Ori.m[i][j];
						// for columns
						for(int k = j; k < 4; k++)
						{
							Ori.m[i][k] += First * Ori.m[j][k];
							Ret.m[i][k] += First * Ret.m[j][k];
						}
					}
			}
		// make diagonal matrix
		// diagonal
		for(j = 3; j > 0; j--)
		{
			// for rows
			for(int i = j - 1; i >= 0; i--)
			{
				float Mul = Ori.m[i][j];
				// for columns
				for(int k = 0; k < 4; k++)
				{
					// annihilate other numbers in the same row
					Ori.m[i][k] -= Mul * Ori.m[j][k];
					Ret.m[i][k] -= Mul * Ret.m[j][k];
				}
			}
		}
		return Ret;
	}*/


	// programmed by pal18
	Matrix Transpose(void)
	{
		Matrix Ret;
		Ret._11 = _11;
		Ret._21 = _12;
		Ret._31 = _13;
		Ret._41 = _14;

		Ret._12 = _21;
		Ret._22 = _22;
		Ret._32 = _23;
		Ret._42 = _24;

		Ret._13 = _31;
		Ret._23 = _32;
		Ret._33 = _33;
		Ret._43 = _34;

		Ret._14 = _41;
		Ret._24 = _42;
		Ret._34 = _43;
		Ret._44 = _44;
		return Ret;
	}

	// programmed by pal18
	Matrix Invert()
	{
		Matrix Ret;
		Matrix TransposedMtx = Transpose();
		float fTemp[12], fDet; // coupled cofactors, matrix determinant
		// pairs for first 8 cofactors
		fTemp[0] = TransposedMtx._33 * TransposedMtx._44;
		fTemp[1] = TransposedMtx._34 * TransposedMtx._43;
		fTemp[2] = TransposedMtx._32 * TransposedMtx._44;
		fTemp[3] = TransposedMtx._34 * TransposedMtx._42;
		fTemp[4] = TransposedMtx._32 * TransposedMtx._43;
		fTemp[5] = TransposedMtx._33 * TransposedMtx._42;
		fTemp[6] = TransposedMtx._31 * TransposedMtx._44;
		fTemp[7] = TransposedMtx._34 * TransposedMtx._41;
		fTemp[8] = TransposedMtx._31 * TransposedMtx._43;
		fTemp[9] = TransposedMtx._33 * TransposedMtx._41;
		fTemp[10] = TransposedMtx._31 * TransposedMtx._42;
		fTemp[11] = TransposedMtx._32 * TransposedMtx._41;
		// calculate first 8 cofactors
		Ret._11  = fTemp[0]*TransposedMtx._22 + fTemp[3]*TransposedMtx._23 + fTemp[4] *TransposedMtx._24;
		Ret._11 -= fTemp[1]*TransposedMtx._22 + fTemp[2]*TransposedMtx._23 + fTemp[5] *TransposedMtx._24;
		Ret._12  = fTemp[1]*TransposedMtx._21 + fTemp[6]*TransposedMtx._23 + fTemp[9] *TransposedMtx._24;
		Ret._12 -= fTemp[0]*TransposedMtx._21 + fTemp[7]*TransposedMtx._23 + fTemp[8] *TransposedMtx._24;
		Ret._13  = fTemp[2]*TransposedMtx._21 + fTemp[7]*TransposedMtx._22 + fTemp[10]*TransposedMtx._24;
		Ret._13 -= fTemp[3]*TransposedMtx._21 + fTemp[6]*TransposedMtx._22 + fTemp[11]*TransposedMtx._24;
		Ret._14  = fTemp[5]*TransposedMtx._21 + fTemp[8]*TransposedMtx._22 + fTemp[11]*TransposedMtx._23;
		Ret._14 -= fTemp[4]*TransposedMtx._21 + fTemp[9]*TransposedMtx._22 + fTemp[10]*TransposedMtx._23;
		Ret._21  = fTemp[1]*TransposedMtx._12 + fTemp[2]*TransposedMtx._13 + fTemp[5] *TransposedMtx._14;
		Ret._21 -= fTemp[0]*TransposedMtx._12 + fTemp[3]*TransposedMtx._13 + fTemp[4] *TransposedMtx._14;
		Ret._22  = fTemp[0]*TransposedMtx._11 + fTemp[7]*TransposedMtx._13 + fTemp[8] *TransposedMtx._14;
		Ret._22 -= fTemp[1]*TransposedMtx._11 + fTemp[6]*TransposedMtx._13 + fTemp[9] *TransposedMtx._14;
		Ret._23  = fTemp[3]*TransposedMtx._11 + fTemp[6]*TransposedMtx._12 + fTemp[11]*TransposedMtx._14;
		Ret._23 -= fTemp[2]*TransposedMtx._11 + fTemp[7]*TransposedMtx._12 + fTemp[10]*TransposedMtx._14;
		Ret._24  = fTemp[4]*TransposedMtx._11 + fTemp[9]*TransposedMtx._12 + fTemp[10]*TransposedMtx._13;
		Ret._24 -= fTemp[5]*TransposedMtx._11 + fTemp[8]*TransposedMtx._12 + fTemp[11]*TransposedMtx._13;
		// pairs for second 8 cofactors
		fTemp[0] = TransposedMtx._13 * TransposedMtx._24;
		fTemp[1] = TransposedMtx._14 * TransposedMtx._23;
		fTemp[2] = TransposedMtx._12 * TransposedMtx._24;
		fTemp[3] = TransposedMtx._14 * TransposedMtx._22;
		fTemp[4] = TransposedMtx._12 * TransposedMtx._23;
		fTemp[5] = TransposedMtx._13 * TransposedMtx._22;
		fTemp[6] = TransposedMtx._11 * TransposedMtx._24;
		fTemp[7] = TransposedMtx._14 * TransposedMtx._21;
		fTemp[8] = TransposedMtx._11 * TransposedMtx._23;
		fTemp[9] = TransposedMtx._13 * TransposedMtx._21;
		fTemp[10] = TransposedMtx._11 * TransposedMtx._22;
		fTemp[11] = TransposedMtx._12 * TransposedMtx._21;
		// calculate second 8 cofactors
		Ret._31  = fTemp[0] *TransposedMtx._42 + fTemp[3] *TransposedMtx._43 + fTemp[4] *TransposedMtx._44;
		Ret._31 -= fTemp[1] *TransposedMtx._42 + fTemp[2] *TransposedMtx._43 + fTemp[5] *TransposedMtx._44;
		Ret._32  = fTemp[1] *TransposedMtx._41 + fTemp[6] *TransposedMtx._43 + fTemp[9] *TransposedMtx._44;
		Ret._32 -= fTemp[0] *TransposedMtx._41 + fTemp[7] *TransposedMtx._43 + fTemp[8] *TransposedMtx._44;
		Ret._33  = fTemp[2] *TransposedMtx._41 + fTemp[7] *TransposedMtx._42 + fTemp[10]*TransposedMtx._44;
		Ret._33 -= fTemp[3] *TransposedMtx._41 + fTemp[6] *TransposedMtx._42 + fTemp[11]*TransposedMtx._44;
		Ret._34  = fTemp[5] *TransposedMtx._41 + fTemp[8] *TransposedMtx._42 + fTemp[11]*TransposedMtx._43;
		Ret._34 -= fTemp[4] *TransposedMtx._41 + fTemp[9] *TransposedMtx._42 + fTemp[10]*TransposedMtx._43;
		Ret._41  = fTemp[2] *TransposedMtx._33 + fTemp[5] *TransposedMtx._34 + fTemp[1] *TransposedMtx._32;
		Ret._41 -= fTemp[4] *TransposedMtx._34 + fTemp[0] *TransposedMtx._32 + fTemp[3] *TransposedMtx._33;
		Ret._42  = fTemp[8] *TransposedMtx._34 + fTemp[0] *TransposedMtx._31 + fTemp[7] *TransposedMtx._33;
		Ret._42 -= fTemp[6] *TransposedMtx._33 + fTemp[9] *TransposedMtx._34 + fTemp[1] *TransposedMtx._31;
		Ret._43  = fTemp[6] *TransposedMtx._32 + fTemp[11]*TransposedMtx._34 + fTemp[3] *TransposedMtx._31;
		Ret._43 -= fTemp[10]*TransposedMtx._34 + fTemp[2] *TransposedMtx._31 + fTemp[7] *TransposedMtx._32;
		Ret._44  = fTemp[10]*TransposedMtx._33 + fTemp[4] *TransposedMtx._31 + fTemp[9] *TransposedMtx._32;
		Ret._44 -= fTemp[8] *TransposedMtx._32 + fTemp[11]*TransposedMtx._33 + fTemp[5] *TransposedMtx._31;

		fDet = TransposedMtx._11*Ret._11 + TransposedMtx._12*Ret._12 + TransposedMtx._13*Ret._13 + TransposedMtx._14*Ret._14;
		fDet = 1 / fDet;

		Ret._11 *= fDet;
		Ret._12 *= fDet;
		Ret._13 *= fDet;
		Ret._14 *= fDet;

		Ret._21 *= fDet;
		Ret._22 *= fDet;
		Ret._23 *= fDet;
		Ret._24 *= fDet;

		Ret._31 *= fDet;
		Ret._32 *= fDet;
		Ret._33 *= fDet;
		Ret._34 *= fDet;

		Ret._41 *= fDet;
		Ret._42 *= fDet;
		Ret._43 *= fDet;
		Ret._44 *= fDet;

		return Ret;
	}

	void SetIdentityMatrix(void)
	{
		for(int i = 0; i < 4; i++)
			for(int j = 0; j < 4; j++)
				m[i][j] = ((i == j) ? 1.0f : 0.0f);
	}

	static Matrix IdentityMatrix(void)
	{
		Matrix Mtx;
		Mtx.SetIdentityMatrix();
		return Mtx;
	}

	void SetTranslationMatrix(const Vector &Translation)
	{
		SetIdentityMatrix();
		m[3][0] = Translation.x;
		m[3][1] = Translation.y;
		m[3][2] = Translation.z;
	}

	static Matrix TranslationMatrix(const Vector &Translation)
	{
		Matrix Mtx;
		Mtx.SetTranslationMatrix(Translation);
		return Mtx;
	}

	void SetScaleTranslationMatrix(const Vector &Scale, const Vector &Translation)
	{
		SetIdentityMatrix();
		m[0][0] = Scale.x;
		m[1][1] = Scale.y;
		m[2][2] = Scale.z;
		m[3][0] = Translation.x;
		m[3][1] = Translation.y;
		m[3][2] = Translation.z;
	}

	static Matrix ScaleTranslationMatrix(const Vector &Scale, const Vector &Translation)
	{
		Matrix Mtx;
		Mtx.SetScaleTranslationMatrix(Scale, Translation);
		return Mtx;
	}

	void SetScaleMatrix(const Vector &Scale)
	{
		SetIdentityMatrix();
		m[0][0] = Scale.x;
		m[1][1] = Scale.y;
		m[2][2] = Scale.z;
	}

	static Matrix ScaleMatrix(const Vector &Scale)
	{
		Matrix Mtx;
		Mtx.SetScaleMatrix(Scale);
		return Mtx;
	}

	void SetProjectionMatrix(void)
	{
		SetIdentityMatrix();
		m[2][3] = 1.0f;
	}

	static Matrix ProjectionMatrix(void)
	{
		Matrix Mtx;
		Mtx.SetProjectionMatrix();
		return Mtx;
	}

	void SetRotationXMatrix(float Angle)
	{
		SetIdentityMatrix();
		m[1][1] =  (float)cos(Angle);
		m[1][2] =  (float)sin(Angle);
		m[2][2] =  (float)cos(Angle);
		m[2][1] = -(float)sin(Angle);
	}

	static Matrix RotationXMatrix(float Angle)
	{
		Matrix Mtx;
		Mtx.SetRotationXMatrix(Angle);
		return Mtx;
	}

	void SetRotationXMatrix(float Angle, float y, float z)
	{
		SetTranslationMatrix(IMATH::Vector(0, -y, -z));
		Matrix R;
		R.SetRotationXMatrix(Angle);
		*this = R * *this;
		R.SetTranslationMatrix(IMATH::Vector(0, y, z));
		*this = R * *this;
	}

	static Matrix RotationXMatrix(float Angle, float y, float z)
	{
		Matrix Mtx;
		Mtx.SetRotationXMatrix(Angle, y, z);
		return Mtx;
	}

	void SetRotationYMatrix(float Angle)
	{
		SetIdentityMatrix();
		m[0][0] =  (float)cos(Angle);
		m[0][2] = -(float)sin(Angle);
		m[2][2] =  (float)cos(Angle);
		m[2][0] =  (float)sin(Angle);
	}

	static Matrix RotationYMatrix(float Angle)
	{
		Matrix Mtx;
		Mtx.SetRotationYMatrix(Angle);
		return Mtx;
	}

	void SetRotationYMatrix(float Angle, float x, float z)
	{
		SetTranslationMatrix(IMATH::Vector(-x, 0, -z));
		Matrix R;
		R.SetRotationYMatrix(Angle);
		*this = R * *this;
		R.SetTranslationMatrix(IMATH::Vector(x, 0, z));
		*this = R * *this;
	}

	static Matrix RotationYMatrix(float Angle, float x, float z)
	{
		Matrix Mtx;
		Mtx.SetRotationYMatrix(Angle, x, z);
		return Mtx;
	}

	void SetRotationZMatrix(float Angle)
	{
		SetIdentityMatrix();
		m[0][0] =  (float)cos(Angle);
		m[0][1] =  (float)sin(Angle);
		m[1][1] =  (float)cos(Angle);
		m[1][0] = -(float)sin(Angle);
	}

	static Matrix RotationZMatrix(float Angle)
	{
		Matrix Mtx;
		Mtx.SetRotationZMatrix(Angle);
		return Mtx;
	}

	void SetRotationZMatrix(float Angle, float x, float y)
	{
		SetTranslationMatrix(IMATH::Vector(-x, -y, 0));
		Matrix R;
		R.SetRotationZMatrix(Angle);
		*this = R * *this;
		R.SetTranslationMatrix(IMATH::Vector(x, y, 0));
		*this = R * *this;
	}

	static Matrix RotationZMatrix(float Angle, float x, float y)
	{
		Matrix Mtx;
		Mtx.SetRotationZMatrix(Angle, x, y);
		return Mtx;
	}

	// Translation:		Position of object
	// Angle:			XZ Angle of object
	// Normal:			Normal of terrain
	void SetTangentMatrix(const Vector &Translation, float Angle, const Vector &Normal)
	{
		// TODO: optimize for speed!
		// normal in tank's coordinate system
		Matrix Mtx;
		Mtx.SetRotationYMatrix(-Angle);
		Vector Normal2 = Mtx * Normal;

		// z = Vector pointing to the front
		SetIdentityMatrix();
		Vector z(0, -Normal2.z, Normal2.y);
		z = Vector::Normalize(z);
		// x = Vector pointing to the right
		Vector x = Vector::CrossProduct(Normal2, z);

		// turn matrix
		Mtx.SetRotationYMatrix(Angle);
		x = Mtx * x;
		z = Mtx * z;
		Normal2 = Mtx * Normal2;

		m[1][0] = Normal2.x;
		m[1][1] = Normal2.y;
		m[1][2] = Normal2.z;
		m[2][0] = z.x;
		m[2][1] = z.y;
		m[2][2] = z.z;
		m[0][0] = x.x;
		m[0][1] = x.y;
		m[0][2] = x.z;
		m[3][0] = Translation.x;
		m[3][1] = Translation.y;
		m[3][2] = Translation.z;
	}

	// Translation:		Position of object
	// Angle:			XZ Angle of object
	// Normal:			Normal of terrain
	// PitchRollCenter: Center of Rotation
	void SetTangentMatrix(const Vector &Translation, float Angle, const Vector &Normal, const Vector &PitchRollCenter, const Vector &TurnCenter)
	{
		// TODO: optimize for speed!
		Matrix Mtx;

		// normal in tank's coordinate system
		Mtx.SetRotationYMatrix(-Angle);
		Vector Normal2 = Mtx * Normal;

		// z = Vector pointing to the front
		Vector z(0, -Normal2.z, Normal2.y);
		z = Vector::Normalize(z);

		// x = Vector pointing to the right
		Vector x = Vector::CrossProduct(Normal2, z);

		// pitch, roll
		SetIdentityMatrix();
		Mtx.SetTranslationMatrix(-PitchRollCenter);
		*this = Mtx * *this;

		Mtx.m[1][0] = Normal2.x;
		Mtx.m[1][1] = Normal2.y;
		Mtx.m[1][2] = Normal2.z;
		Mtx.m[2][0] = z.x;
		Mtx.m[2][1] = z.y;
		Mtx.m[2][2] = z.z;
		Mtx.m[0][0] = x.x;
		Mtx.m[0][1] = x.y;
		Mtx.m[0][2] = x.z;
		Mtx.m[3][0] = 0;
		Mtx.m[3][1] = 0;
		Mtx.m[3][2] = 0;
		*this = Mtx * *this;

		Mtx.SetTranslationMatrix(PitchRollCenter);
		*this = Mtx * *this;

		// turn
		/*Mtx.SetTranslationMatrix(-TurnCenter);
		*this = Mtx * *this;*/

		Mtx.SetRotationYMatrix(Angle, TurnCenter.x, TurnCenter.z);
		*this = Mtx * *this;

		/*Mtx.SetTranslationMatrix(TurnCenter);
		*this = Mtx * *this;*/

		// translate
		Mtx.SetTranslationMatrix(Translation);
		*this = Mtx * *this;
	}

	// Translation:	Position of object
	void SetNormalMatrix(const Vector &Translation, const Vector &x, const Vector &y, const Vector &z)
	{
		m[0][0] = x.x;
		m[0][1] = x.y;
		m[0][2] = x.z;
		m[1][0] = y.x;
		m[1][1] = y.y;
		m[1][2] = y.z;
		m[2][0] = z.x;
		m[2][1] = z.y;
		m[2][2] = z.z;
		m[3][0] = Translation.x;
		m[3][1] = Translation.y;
		m[3][2] = Translation.z;
	}

	static Matrix NormalMatrix(const Vector &Translation, const Vector &x, const Vector &y, const Vector &z)
	{
		Matrix Mtx;
		Mtx.SetNormalMatrix(Translation, x, y, z);
		return Mtx;
	}

	// Translation:	Position of object
	// Angle:		XZ Angle of object
	// Pitch:		YZ Angle of object
	// Normal:		Normal of terrain
	void SetAnglePitchMatrix(const Vector &Translation, float Angle, float Pitch, const Vector &Normal)
	{
		// TODO: optimize for speed!
		Matrix Mtx;
		Mtx.SetRotationYMatrix(-Angle);
		Vector Normal2 = Mtx * Normal;

		SetIdentityMatrix();
		Vector z(0, -Normal2.z, Normal2.y);
		z = Vector::Normalize(z);
		Vector x = Vector::CrossProduct(Normal2, z);

		Mtx.SetRotationXMatrix(-Pitch);
		x = Mtx * x;
		z = Mtx * z;
		Normal2 = Mtx * Normal2;

		Mtx.SetRotationYMatrix(Angle);
		x = Mtx * x;
		z = Mtx * z;
		Normal2 = Mtx * Normal2;

		m[1][0] = Normal2.x;
		m[1][1] = Normal2.y;
		m[1][2] = Normal2.z;
		m[2][0] = z.x;
		m[2][1] = z.y;
		m[2][2] = z.z;
		m[0][0] = x.x;
		m[0][1] = x.y;
		m[0][2] = x.z;
		m[3][0] = Translation.x;
		m[3][1] = Translation.y;
		m[3][2] = Translation.z;
	}

	// Translation:		Position of object
	// Angle:			XZ Angle of object
	// Pitch:			YZ Angle of object
	// Normal:			Normal of terrain
	// PitchRollCenter: Center of Normal Rotation
	// PitchCenter:		Center of "Pitch" Rotation
	void SetAnglePitchMatrix(const Vector &Translation, float Angle, float Angle2, float Pitch, float Pitch2, const Vector &Normal, const Vector &PitchRollCenter, const Vector &TurnCenter, const Vector &PitchCenter, const Vector &TurnCenter2)
	{
		// TODO: optimize for speed!
		Matrix Mtx;

		// normal in tank's coordinate system
		Mtx.SetRotationYMatrix(-Angle);
		Vector Normal2 = Mtx * Normal;

		// z = Vector pointing to the front
		Vector z(0, -Normal2.z, Normal2.y);
		z = Vector::Normalize(z);

		// x = Vector pointing to the right
		Vector x = Vector::CrossProduct(Normal2, z);

		// pitch, roll
		SetIdentityMatrix();
		Mtx.SetTranslationMatrix(-PitchRollCenter);
		*this = Mtx * *this;

		Mtx.m[1][0] = Normal2.x;
		Mtx.m[1][1] = Normal2.y;
		Mtx.m[1][2] = Normal2.z;
		Mtx.m[2][0] = z.x;
		Mtx.m[2][1] = z.y;
		Mtx.m[2][2] = z.z;
		Mtx.m[0][0] = x.x;
		Mtx.m[0][1] = x.y;
		Mtx.m[0][2] = x.z;
		Mtx.m[3][0] = 0;
		Mtx.m[3][1] = 0;
		Mtx.m[3][2] = 0;
		*this = Mtx * *this;

		Mtx.SetTranslationMatrix(PitchRollCenter);
		*this = Mtx * *this;

		// Pitch2, Angle2
		Mtx.SetRotationXMatrix(-Pitch2, PitchCenter.y, PitchCenter.z);
		*this = Mtx * *this;
		Mtx.SetRotationYMatrix(Angle2, TurnCenter2.x, TurnCenter2.z);
		*this = Mtx * *this;

		// tank turn, pitch
		/*Mtx.SetTranslationMatrix(-TurnCenter);
		*this = Mtx * *this;*/

		Mtx.SetRotationXMatrix(-Pitch);
		*this = Mtx * *this;
		Mtx.SetRotationYMatrix(Angle, TurnCenter.x, TurnCenter.z);
		*this = Mtx * *this;

		/*Mtx.SetTranslationMatrix(TurnCenter);
		*this = Mtx * *this;*/

		// translate
		Mtx.SetTranslationMatrix(Translation);
		*this = Mtx * *this;
	}

};

inline Vector &operator *=(Vector &v, const Matrix &m)
{
	// apply
	const float x = v.x * m.m[0][0] + v.y * m.m[1][0] + v.z * m.m[2][0] + m.m[3][0];
	const float y = v.x * m.m[0][1] + v.y * m.m[1][1] + v.z * m.m[2][1] + m.m[3][1];
	const float z = v.x * m.m[0][2] + v.y * m.m[1][2] + v.z * m.m[2][2] + m.m[3][2];
	// scale
	float s = v.x * m.m[0][3] + v.y * m.m[1][3] + v.z * m.m[2][3] + m.m[3][3];
	if(s != 0.0f){
		s = 1/s;
		v.x = x * s;
		v.y = y * s;
		v.z = z * s;
	}
	else
	{
		v.x = x;
		v.y = y;
		v.z = z;
	}
	return v;
}

}

#endif
