/***********************************************************/
/*                                                         */
/* Game.cpp                                                */
/*                                                         */
/* created 13.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#include "Game.h"

#include "MathStructures.h"
#include "LogFile.h"
#include "MessageWindow.h"
#include "Text.h"
#include "DataStructures.h"
#include "IMath.h"
#include "IDMissionGoal.h"
#include "IDType.h"
#include "IDTank.h"
#include "SpawnInterface.h"
#include "IDParticleFX.h"
#include "IDType.h"
#include "Airfield.h"
#include "IDSide.h"

#include <time.h>
#include <cassert>

// Tank 4
namespace T4
{

void Game::CICommand::Command(const char *Command)
{
	try
	{
		if(!stricmp(Command, "MissionComplete"))
		{
			Game->Music.Play("Music\\Song 1.wav");
			// Mission Complete
			Game->State = GAME_STATE_DONE;
			I3D::Menu::MenuSpecs MSpecs;
			MSpecs.DS = &Game->DS;
			MSpecs.Directory = "Menus";
			MSpecs.MenuFile = "MissionComplete.ini";
			MSpecs.Visible = true;
			MSpecs.CommandInterface = this;
			Game->Menu.Create(&MSpecs);
		}
		else if(!stricmp(Command, "MissionFailed"))
		{
			Game->Music.Play("Music\\Song 1.wav");
			// Mission Lost
			Game->State = GAME_STATE_DONE;
			I3D::Menu::MenuSpecs MSpecs;
			MSpecs.DS = &Game->DS;
			MSpecs.Directory = "Menus";
			MSpecs.MenuFile = "MissionFailed.ini";
			MSpecs.Visible = true;
			MSpecs.CommandInterface = this;
			Game->Menu.Create(&MSpecs);
		}
		else if(!stricmp(Command, "NextMission"))
		{
			Game->Menu.SetVisible(false);
			const char *Next = Game->MLib.Missions[Game->Mission]->Next;
			if(Next && *Next)
			{
				Game->Mission = Game->MLib.GetMissionNum(Next);
				Game->LoadMission();
			}
			else
				Game->LoadCredits();
		}
		else if(!stricmp(Command, "MissionRepeat"))
		{
			Game->Menu.SetVisible(false);
			Game->LoadMission();
		}
		else if(!stricmp(Command, "Credits"))
		{
			Game->Menu.SetVisible(false);
			Game->LoadCredits();
		}
		else if(!stricmp(Command, "Quit"))
		{
			PostQuitMessage(0);
		}
		else if(!stricmp(Command, "NewGamePA"))
		{
			Game->Menu.SetVisible(false);
			Game->Mission = Game->MLib.GetMissionNum(Game->MLib.FirstMissionPA);
			Game->LoadMission();
		}
		else if(!stricmp(Command, "NewGameIS"))
		{
			Game->Menu.SetVisible(false);
			Game->Mission = Game->MLib.GetMissionNum(Game->MLib.FirstMissionIS);
			Game->LoadMission();
		}
		else if(!stricmp(Command, "AbortMission"))
		{
			Game->LoadState(GAME_STATE_MAINMENU); // destroys Command
		}
		else if(!stricmp(Command, "ToggleFullscreen"))
		{
			Game->ToggleFullscreen();
		}
	}
	catch(char *s)
	{
		throw s;
	}
	catch(...)
	{
		char str[1024];
		sprintf(str, "Game::CICommand::Command");
		throw str;
	}
}

const char *Game::CIData::GetData(const char *Str)
{
	if(!stricmp(Str, "orders"))
	{
		return Game->MLib.Missions[Game->Mission]->Orders;
	}
	return 0;
}

void Game::LoadState(int State)
{
	if(State == GAME_STATE_MAINMENU)
	{
		// menu
		I3D::Menu::MenuSpecs MSpecs;
		MSpecs.DS = &DS;
		MSpecs.Directory = "Menus";
		MSpecs.MenuFile = "MainMenu.ini";
		MSpecs.Visible = true;
		MSpecs.CommandInterface = &ICommand;
		Menu.Create(&MSpecs);
	}
	Game::State = State;
}

void Game::Create(int WindowStyle)
{
	try
	{
		Music.Play("Music\\TankSquad Theme.wav");
		//Music.Play("E:\\Musik\\Klassik\\Beethoven 9 II Molto vivace - Boston Leinsdorf.mp3");
		ICommand.Create(this);
		IData.Create(this);
		srand(time(0));
		WindowStyle = SW_SHOWNORMAL;
		ShowCursor(false);
		// clear the error output file
		ITL::LogFile::Clear("Error.txt");
		FlagFullscreen = false;
		Reinitialize = false;
	}
	catch(char *s)
	{
		throw(s);
	}
	catch(...)
	{
		throw("Game::Create#initializing(0)");
	}
	// Create the main window and initialize D3D
	try
	{
		Wnd.Create(WindowStyle, "Tank 4", 800, 600);
	}
	catch(exception& Ex)
	{
		ITL::LogFile::Msg("Error.txt", (char *)Ex.what());
		ITL::MessageWindow::Message((char *)Ex.what(), "Error during D3D Window initialisation");
		exit(-1);
	}
	int State = 1;
	try
	{
		Wnd.AddListener(this);
		State = 2;
		DS.Create(Wnd.D3DDevice, &Sp);
		State = 3;
		Create3D();
		Wrld.AddListener(this);
		State = 4;
		LoadingScreen("Mouse"); // no beginscene() done yet?
		State = 5;
		Mouse.Create(&DS, Wnd.GetHWnd());
		// story
		StoryRnd.Create(&DS);
		// mission
		State = 6;
		LoadingScreen("Mission Library");
		MLib.Create("Missions");
		State = 8;
		Mission = MLib.GetMissionNum(MLib.FirstMissionPA);
		State = 9;
		LoadState(GAME_STATE_MAINMENU);
		// Sound Manager
		SndMgr.Create(Wnd.GetHWnd());
		// profiler
		Prf.Create(&DS);
		Prf.SetName(0, "Sp, Handlers");
		Prf.SetName(1, "World::Physics");
		Prf.SetName(2, "MissionGoals");
		Prf.SetName(3, "ClrScr, Matrices");
		Prf.SetName(4, "World::Render");
		Prf.SetName(5, "Overlay");
		Prf.SetName(6, "Present");
		Prf.SetName(7, "Windows");
		Prf.SetName(8, "Cursor");
		Prf.SetName(9, "BeginScene");
		Prf.SetName(10, "TestCoop");
		State = 10;

		I3D::Button::Specs MSpecs;
		MSpecs.DS = &DS;
		MSpecs.Pos = IMATH::Vector2D(0.94f, 0.94f);
		MSpecs.Size = IMATH::Vector2D(0.05f, 0.05f);
		MaximizeButton.Create(MSpecs);
	}
	catch(char *s)
	{
		throw(s);
	}
	catch(...)
	{
		char Str[1024];
		sprintf(Str, "Game::Create#initializing(%d)", State);
		throw(Str);
	}
}

void Game::LoadingScreen(const char *Str)
{
	DS.D3DDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER | D3DCLEAR_TARGET, 0x00000000, 1.0f, 0);
	D3DMATRIX Matrix = {1,0,0,0,  0,1,0,0,  0,0,1,0,  0,0,0,1};
	// world and view matrix
	DS.D3DDevice->SetTransform(D3DTS_WORLD, &Matrix);
	DS.D3DDevice->SetTransform(D3DTS_VIEW, &Matrix);
	DS.D3DDevice->SetTransform(D3DTS_PROJECTION, &Matrix);
	// loading text
	char Text[1024] = "loading...";
	HRESULT BeginSceneResult = DS.D3DDevice->BeginScene();
	I3D::Text::DrawString(&DS, Text, -0.05f * 0.5f * strlen(Text), 0, 0.05f, 0.05f);
	I3D::Text::DrawString(&DS, Str, -0.05f * 0.5f * strlen(Str), -0.1f, 0.05f, 0.05f);
	if(BeginSceneResult == D3D_OK)
		DS.D3DDevice->EndScene(); // end scene, if we needed to begin one

	// display
	DS.D3DDevice->Present(NULL, NULL, NULL, NULL);
}

void Game::Create3D(void)
{
	DS.D3DDevice = Wnd.D3DDevice;
	DS.D3DDevice->SetRenderState(D3DRS_LIGHTING,TRUE);
	DS.D3DDevice->SetRenderState(D3DRS_ZENABLE,TRUE);
	DS.D3DDevice->SetSamplerState(0,D3DSAMP_MAGFILTER, D3DTEXF_LINEAR); 
	DS.D3DDevice->SetSamplerState(0,D3DSAMP_MINFILTER, D3DTEXF_LINEAR); 
	DS.D3DDevice->SetSamplerState(0,D3DSAMP_MIPFILTER, D3DTEXF_POINT); 
	DS.D3DDevice->SetSamplerState(1,D3DSAMP_MAGFILTER, D3DTEXF_LINEAR); 
	DS.D3DDevice->SetSamplerState(1,D3DSAMP_MINFILTER, D3DTEXF_LINEAR); 
	DS.D3DDevice->SetSamplerState(1,D3DSAMP_MIPFILTER, D3DTEXF_POINT); 
    //DS.D3DDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
    //DS.D3DDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    //DS.D3DDevice->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_POINT);
    //DS.D3DDevice->SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
    //DS.D3DDevice->SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
    //DS.D3DDevice->SetTextureStageState(1, D3DTSS_MIPFILTER, D3DTEXF_POINT);
	DS.D3DDevice->SetRenderState(D3DRS_CULLMODE,D3DCULL_CW);
	// D3DRS_AMBIENT is now handled in World
//	DS.D3DDevice->SetRenderState(D3DRS_AMBIENT, 0x00000000);
//	DS.D3DDevice->SetRenderState(D3DRS_AMBIENT, 0x40404040);
//	DS.D3DDevice->SetRenderState(D3DRS_AMBIENT, 0x80808080);

	// for debugging:
	//DS.D3DDevice->SetRenderState(D3DRS_AMBIENT, 0xFFFFFFFF);
	if(D3DFlags & GAME_D3DFLAG_WIREFRAME)
		DS.D3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);

	D3DMATERIAL9 Material;
	memset(&Material, 0, sizeof(D3DMATERIAL9));
	Material.Diffuse.r = 1.0f;
	Material.Diffuse.g = 1.0f;
	Material.Diffuse.b = 1.0f;
	Material.Diffuse.a = 1.0f;
	Material.Ambient.r = 1.0f;
	Material.Ambient.g = 1.0f;
	Material.Ambient.b = 1.0f;
	Material.Ambient.a = 1.0f;
	Material.Specular.r = 1.0f;// direct light reflections
	Material.Specular.g = 1.0f;
	Material.Specular.b = 1.0f;
	Material.Specular.a = 1.0f;
	Material.Power = 0.0f;
	//  Material.power = (float)10.0;
	DS.D3DDevice->SetMaterial(&Material);
	Wrld.InitLight(&DS);
	AsyncReady = true;

	// load mouse cursors
	MouseTextures[0] = I3D::TextureManager::GetTexture(&DS, "Textures\\cursor_move.bmp", 0, I3D::TM_FLAG_ALPHACC0);
	MouseTextures[1] = I3D::TextureManager::GetTexture(&DS, "Textures\\cursor_impassable.bmp", 0, I3D::TM_FLAG_ALPHACC0);
	MouseTextures[2] = I3D::TextureManager::GetTexture(&DS, "Textures\\cursor_select.bmp", 0, I3D::TM_FLAG_ALPHACC0);
	MouseTextures[3] = I3D::TextureManager::GetTexture(&DS, "Textures\\cursor_attack.bmp", 0, I3D::TM_FLAG_ALPHACC0);
}

void Game::Init3D(bool recurse)
{
	int Stage = 0;
	try
	{
		Stage = 1;
		DS.D3DDevice = Wnd.D3DDevice;
		Stage = 2;
		if(DS.D3DDevice && recurse)
		{
			Stage = 3;
			I3D::TextureManager::Clear();
			I3D::Text::Init3D(&DS, true);
			LoadingScreen("reinitializing Direct3D");
			Stage = 4;
		}
		LoadingScreen("reinitializing Direct3D (Game)");
		Create3D();
		if(DS.D3DDevice && recurse)
		{
			LoadingScreen("reinitializing Direct3D (Mouse)");
			Mouse.Init3D(&DS, true);
			LoadingScreen("reinitializing Direct3D (Menu)");
			Menu.Init3D(&DS, true);
			Stage = 5;
			MaximizeButton.Init3D(&DS, true);
			Stage = 6;
			StoryRnd.Init3D(&DS, true);
			Stage = 7;
			LoadingScreen("reinitializing Direct3D (World)");
			Wrld.Init3D(&DS, true);
			LoadingScreen("reinitializing Direct3D (done)");
			Stage = 8;
		}
	}
	catch(char *s)
	{
		throw(s);
	}
	catch(...)
	{
		// unknown exceptions
		char Str[1024];
		sprintf(Str, "Unknown Exception in Game::Init3D#%d", Stage);
		throw(Str);
	}
}

void Game::LoadCredits(void)
{
	Music.Play("Music\\TankSquad Theme.wav");
	Music.Enqueue("Music\\Song 2.wav");
	Credits.Create("Libs\\Credits.txt");
	State = GAME_STATE_CREDITS;
}

void Game::LoadMission(void)
{
	int Error = 0;
	try
	{
		Error = 1;
		int i;
		Music.Play("Music\\Song 2.wav");
		LoadingScreen("World");
		Error = 100;
		IGAME::World::WorldSpecs WS;
		WS.DS = &DS;
		WS.SndMgr = &SndMgr;
		Error = 101;
		WS.Seed = MLib.Missions[Mission]->Seed;
		Error = 102;
		for(i = 0; i < MLib.Missions[Mission]->Ways.Count(); i++)
			WS.Ways[i] = &MLib.Missions[Mission]->Ways[i];
		Error = 11;
		for(i = 0; i < MLib.Missions[Mission]->TerrainDscs.Count(); i++)
			WS.TerrainDscs[i] = &MLib.Missions[Mission]->TerrainDscs[i];
		Error = 12;
		for(i = 0; i < MLib.Missions[Mission]->Airfields.Count(); i++)
			WS.Airfields[i] = MLib.Missions[Mission]->Airfields[i];
		Error = 13;
		for(i = 0; i < MLib.Missions[Mission]->Decals.Count(); i++)
			WS.Decals[i] = MLib.Missions[Mission]->Decals[i];
		Error = 2;
		Wrld.Create(&WS);
		Wrld.SetCameraPos(MLib.Missions[Mission]->InitialCameraPos);
		Wrld.SetCameraAngle(MLib.Missions[Mission]->InitialCameraAngle);
		Error = 3;
		for(i = 0; i < MLib.Missions[Mission]->Objects.Count(); i++)
		{
			char Str[1024];
			sprintf(Str, "Mission Object %d of %d", i + 1, MLib.Missions[Mission]->Objects.Count());
			LoadingScreen(Str);
			IGAME::SpawnInterface::SpawnInterfaceSpecs SS;
			IGAME::MissionObject *MObj = &MLib.Missions[Mission]->Objects[i];
			SS.Type = MObj->ObjType;
			SS.ID = MObj->Type;
			SS.Owner = MObj->Owner;
			SS.Pos = IMATH::Vector(MObj->x, Wrld.GetSurfaceY(MObj->x, MObj->z), MObj->z);
			SS.Target = 0;
			SS.Angle = MObj->Angle;
			SS.AI = MObj->AI;
			SS.TargetX = MObj->TargetX;
			SS.TargetZ = MObj->TargetZ;
			SS.Commandable = MObj->Commandable;
			SS.Highlight = MObj->Highlight;
			SS.MaxSpeed = MObj->MaxSpeed;
			strcpy(SS.Name, MObj->Name);
			Wrld.GetSpawnInterface()->Spawn(&SS);
		}
		Error = 4;
		State = GAME_STATE_ACTIVE;
		MissionStartTime = DS.Sp->GameTime;
		LoadStory("MainStory");
		Error = 5;
		// show orders
		if(*MLib.Missions[Mission]->Orders)
		{
			I3D::Menu::MenuSpecs MSpecs;
			MSpecs.DS = &DS;
			MSpecs.Directory = "Menus";
			MSpecs.MenuFile = "Orders.ini";
			MSpecs.Visible = true;
			MSpecs.CommandInterface = &ICommand;
			MSpecs.DataIf = &IData;
			Menu.Create(&MSpecs);
		}
		Error = 6;
	}
	catch(char *s)
	{
		throw s;
	}
	catch(...)
	{
		char str[1024];
		sprintf(str, "Game::LoadMission#%d", Error);
		throw str;
	}
}

void Game::LoadStory(const char *Name)
{
	// Story
	int Num = MLib.Missions[Mission]->GetStoryNum(Name);
	if(Num == -1)
	{
		StoryItem = 0;
		StoryRnd.SetBumper(0);
		StoryRnd.SetTicker("");
		StoryRnd.SetTopic("");
	}
	else
	{
		StoryItem = MLib.Missions[Mission]->StoryItems[Num];
		StoryRnd.SetBumper(StoryItem->BumperID);
		StoryRnd.SetTicker(StoryItem->NewsTicker);
		StoryRnd.SetTickerStyle(StoryItem->TickerStyle);
		StoryRnd.SetTopic(StoryItem->Topic);
		StoryRnd.SetText(StoryItem->Text);
		StoryRnd.SetTimeDelta(StoryItem->TimeDelta);
	}
	StoryTime = Sp.GameTime;
}

void Game::CheckStory(void)
{
	if(StoryItem)
	{
		if(Sp.GameTime - StoryTime > /*StoryItem->TimeDeltaStory*/ StoryRnd.GetTimeDelta())
			LoadStory(StoryItem->Next);
	}
}

// called asynchronously!
void Game::OnEvent(I3D::D3DWEvent Event)
{
	try
	{
		if(!AsyncReady) return;
		switch(Event.GetID())
		{
		case I3D::D3DW_EVENT_RESET:
			Reinitialize = true;
			break;
		default:
			Events[Events.Count()] = Event;
		}
	}
	catch(char *s)
	{
		throw(s);
	}
	catch(...)
	{
		throw("Game::OnEvent");
	}
}

void Game::EventHandler(void)
{
	if(Events.Count())
	{
		for(int i = 0; i < Events.Count(); i++)
		{
			switch(Events[i].GetID())
			{
			/*case I3D::D3DW_EVENT_KEYDOWN:
				KeyboardInput(Events[i].GetInfo());
				break;*/
			case I3D::D3DW_EVENT_WINDOWSEVENT:
				switch(Events[i].GetInfo())
				{
				case WM_KEYDOWN:
					KeyboardHandler(Events[i].GetWParam());
					break;
				/*case WM_LBUTTONDOWN:
					break;
				case WM_MOUSEWHEEL:
					int Wheel = (short)(Events[i].GetWParam() >> 16);
					break;*/
				}
				break;
			}
		}
		Events.Clear();
	}
}

void Game::KeyboardHandler(char c)
{
	switch(c)
	{
	case 27:
		// Escape
		if(StoryItem)
		{
			LoadStory(StoryItem->Next);
		}
		else if(State == GAME_STATE_CREDITS)
		{
			LoadState(GAME_STATE_MAINMENU);
		}
		else if(!Menu.GetVisible())
		{
			I3D::Menu::MenuSpecs MSpecs;
			MSpecs.DS = &DS;
			MSpecs.Directory = "Menus";
			if(GetIngame())
				MSpecs.MenuFile = "MainGame.ini";
			else
				MSpecs.MenuFile = "MainMenu.ini";
			MSpecs.Visible = true;
			MSpecs.CommandInterface = &ICommand;
			MSpecs.DataIf = &IData;
			Menu.Create(&MSpecs);
		}
		break;
	case 't':
	case 'T':
			I3D::Menu::MenuSpecs MSpecs;
			MSpecs.DS = &DS;
			MSpecs.Directory = "Menus";
			MSpecs.MenuFile = "Orders.ini";
			MSpecs.Visible = true;
			MSpecs.CommandInterface = &ICommand;
			MSpecs.DataIf = &IData;
			Menu.Create(&MSpecs);
		break;
	case 'w':
	case 'W':
		D3DFlags ^= GAME_D3DFLAG_WIREFRAME;
		if(D3DFlags & GAME_D3DFLAG_WIREFRAME)
			DS.D3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
		else
			DS.D3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
		break;
	case 'u':
	case 'U':
		Flags ^= FLAG_DEBUG;
		break;
	case 'p':
	case 'P':
		Sp.Paused = !Sp.Paused;
		break;
	case 's':
	case 'S':
		Prf.SetShow(!Prf.GetShow());
		break;
	case 'q':
	case 'Q':
		if(GetIngame())
			Wrld.SetQTreeProfShow(!Wrld.GetQTreeProfShow());
		break;
	case 'r':
	case 'R':
		if(GetIngame())
			Wrld.SetProfShow(!Wrld.GetProfShow());
		break;
	case 'e':
	case 'E':
		if(GetIngame())
			Wrld.SetPhysicsProfShow(!Wrld.GetPhysicsProfShow());
		break;
	case VK_NUMPAD4:
	case VK_LEFT:
		if(GetIngame())
			Wrld.MoveCamera(I3D::Vector(-4, 0, 0));
		break;
	case VK_NUMPAD6:
	case VK_RIGHT:
		if(GetIngame())
			Wrld.MoveCamera(I3D::Vector(4, 0, 0));
		break;
	case VK_NUMPAD8:
	case VK_UP:
		if(GetIngame())
			Wrld.MoveCamera(I3D::Vector(0, 0, 4));
		break;
	case VK_NUMPAD2:
	case VK_DOWN:
		if(GetIngame())
			Wrld.MoveCamera(I3D::Vector(0, 0, -4));
		break;
	case VK_NUMPAD7:
	case VK_HOME:
		if(GetIngame())
			Wrld.MoveCamera(I3D::Vector(0, 4, 0));
		break;
	case VK_NUMPAD1:
	case VK_END:
		if(GetIngame())
			Wrld.MoveCamera(I3D::Vector(0, -4, 0));
		break;
	case VK_ADD:
		if(GetIngame())
			Wrld.MoveCameraDirect(I3D::Vector(0, 0, 4));
		break;
	case VK_SUBTRACT:
		if(GetIngame())
			Wrld.MoveCameraDirect(I3D::Vector(0, 0, -4));
		break;
	case VK_NUMPAD9:
	case VK_PRIOR:
		if(GetIngame())
			Wrld.RotateCamera(I3D::Vector(-0.02f, 0, 0));
		break;
	case VK_NUMPAD3:
	case VK_NEXT:
		if(GetIngame())
			Wrld.RotateCamera(I3D::Vector(0.02f, 0, 0));
		break;
	case VK_NUMPAD0:
	case VK_INSERT:
		if(GetIngame())
			Wrld.RotateCamera(I3D::Vector(0, -0.02f, 0));
		break;
	case VK_DELETE:
	case VK_DECIMAL:
		if(GetIngame())
			Wrld.RotateCamera(I3D::Vector(0, 0.02f, 0));
		break;
	case VK_DIVIDE:
		if(GetIngame())
			Wrld.RotateCamera(I3D::Vector(0, 0, 0.02f));
		break;
	case VK_MULTIPLY:
		if(GetIngame())
			Wrld.RotateCamera(I3D::Vector(0, 0, -0.02f));
		break;
	}
}

void Game::MouseHandler(void)
{
	Scrolling();
	I3D::MouseCursor::MouseCursorEvent Event;
	while(Mouse.GetEvent(&Event))
	{
		if(GetFocus() == Wnd.GetHWnd())
			switch(Event.Code)
			{
			case I3D::MC_CLICK:
				if(MaximizeButton.Click(IMATH::Vector2D(Event.x, Event.y)))
					Wnd.ToggleFullscreen();
				else if(!(StoryItem && StoryRnd.Click(IMATH::Vector2D(Event.x, Event.y))))
				{
					if(!Menu.Click(IMATH::Vector2D(Event.x, Event.y)))
						if(GetIngame())
							Wrld.Click(IMATH::Vector2D(Event.x, Event.y));
				}
				break;
			case I3D::MC_RIGHTCLICK:
				if(GetIngame())
					Wrld.RightClick(IMATH::Vector2D(Event.x, Event.y));
				break;
			case I3D::MC_DRAG:
				if(GetIngame())
					Wrld.Drag(IMATH::Vector2D(Event.x, Event.y), IMATH::Vector2D(Event.x2, Event.y2));
				break;
			case I3D::MC_DRAGSTART:
				if(GetIngame())
					Wrld.DragStart(IMATH::Vector2D(Event.x, Event.y));
				break;
			case I3D::MC_DRAGEND:
				if(GetIngame())
					Wrld.DragEnd(IMATH::Vector2D(Event.x, Event.y), IMATH::Vector2D(Event.x2, Event.y2));
				break;
			}
	}
}

char *Game::DebugInfo()
{
	char *Info = new char[2048];
	// TODO: create a son of a god of a debugmessage
	strcpy(Info, "Debug Data:\r\n");
	strcat(Info, "Flags:\r\n");
	// Generate some 1337 looking hex codes (to impress friends, ...)
	for(int Block = 0; Block < 8; Block++)
	{
		if(Block)
		{
			char *Last = Info + strlen(Info);
			*Last = ' ';
			Last[1] = 0;
		}
		for(int Digit = 0; Digit < 2; Digit++)
		{
			char *Last = Info + strlen(Info);
			int Value = rand()%16;
			if(Value < 10)
				*Last = 48 + Value;
			else
				*Last = 55 + Value;
			Last[1] = 0;
		}
	}
	strcat(Info, "\r\nMemory:");
	// Generate even more 1337 looking hex codes
	for(int Line = 0; Line < 4; Line++)
	{
		char *Last = Info + strlen(Info);
		*Last = '\r';
		Last[1] = '\n';
		Last[2] = 0;
		for(int Block = 0; Block < 4; Block++)
		{
			if(Block)
			{
				char *Last = Info + strlen(Info);
				*Last = ' ';
				Last[1] = 0;
			}
			for(int Digit = 0; Digit < 8; Digit++)
			{
				char *Last = Info + strlen(Info);
				int Value = rand()%16;
				if(Value < 10)
					*Last = 48 + Value;
				else
					*Last = 55 + Value;
				Last[1] = 0;
			}
		}
	}
	strcat(Info, "\r\nshit happens.");
/*	strcat(Info, "lets see what the preprocessor did...\r\n");
	for(int Block = 0; Block < 256; Block++)
	{
		switch(rand() % 5)
		{
		case 0: strcat(Info, "05 "); break;
		case 1: strcat(Info, "FE "); break;
		case 2: strcat(Info, "0E "); break;
		case 3: strcat(Info, "D8 "); break;
		case 4: strcat(Info, "5A "); break;
		}
	}
	strcat(Info, "\r\nhmm so it did quite a lot of stuff... strange, huh?");*/
	strcat(Info, "\r\nwell maybe you just try restarting the game.\r\ngood luck :)");
	return Info;
}

void Game::Scrolling(void)
{
	bool Inside = true;
	if(Mouse.GetX() < -1 || Mouse.GetX() > 1
	|| Mouse.GetY() < -1 || Mouse.GetY() > 1)
		Inside = false;
	if(Inside)
	{
		if(GetIngame())
		{
			float d = Sp.Delta;
			if(Mouse.GetX() < -0.9f)
				Wrld.MoveCamera(IMATH::Vector(-48 * d, 0, 0));
			if(Mouse.GetX() > 0.9f)
				Wrld.MoveCamera(IMATH::Vector( 48 * d, 0, 0));
			if(Mouse.GetY() < -0.9f)
				Wrld.MoveCamera(IMATH::Vector(0, 0, -48 * d));
			if(Mouse.GetY() > 0.9f)
				Wrld.MoveCamera(IMATH::Vector(0, 0,  48 * d));
			static float LastZ = Mouse.GetZ();
			float dZ = Mouse.GetZ() - LastZ;
			static float InterpolateddZ = 0;
			InterpolateddZ *= 0.9f;
			InterpolateddZ += 0.1f * dZ;
			if(absf(InterpolateddZ) > 0.01f)
				Wrld.MoveCameraDirect(IMATH::Vector(0, 0, 0.1f * InterpolateddZ));
			LastZ = Mouse.GetZ();
		}
	}
}

void Game::ClearScreen(void)
{
	if(D3DFlags & GAME_D3DFLAG_WIREFRAME
		|| (StoryItem && !StoryItem->Render)
		|| State == GAME_STATE_MAINMENU
		|| State == GAME_STATE_CREDITS)
			DS.D3DDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER | D3DCLEAR_TARGET, 0x00000000, 1.0f, 0);
	else
		DS.D3DDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER | D3DCLEAR_TARGET, 0x0040B0FF, 1.0f, 0);
}

void Game::Set3DMatrices(void)
{
	D3DMATRIX Matrix = {1,0,0,0,  0,1,0,0,  0,0,1,0,  0,0,0,1};
	DS.D3DDevice->SetTransform(D3DTS_WORLD, &Matrix);
	DS.D3DDevice->SetTransform(D3DTS_VIEW, &Matrix);
	DS.D3DDevice->SetTransform(D3DTS_PROJECTION, &Matrix);
	DS.D3DDevice->GetTransform(D3DTS_VIEW, &DS.ViewMtx);
	// projection matrix
	// distance eye->monitor in meters
	const float EyeDistance = 0.6f;
	// (vertical) monitor size in meters
	const float MonitorSize = 0.33f;
	// minimum distance of objects in game world space (for z-buffer optimization)
	const float MinDistance = 30.0f;
	// perspective projection: [x, y, z] /= (1 + z)
	D3DMATRIX PMatrix = {1,0,0,0,  0,1,0,0,  0,0,1,1,  0,0,0,MinDistance};
	DS.D3DDevice->MultiplyTransform(D3DTS_PROJECTION, &PMatrix);
	// adjust to monitor size and position
	// screen 3:4 = 0.75f
	D3DMATRIX TMatrix3 = {0.75f * 2.0f / MonitorSize,0,0,0,  0,2.0f / MonitorSize,0,0,  0,0,1.0f / EyeDistance,0,  0,0,0,1};
	DS.D3DDevice->MultiplyTransform(D3DTS_PROJECTION, &TMatrix3);
	DS.D3DDevice->GetTransform(D3DTS_PROJECTION, &DS.ProjMtx);
}

void Game::DisplayFPS(void)
{
	if(Sp.DeltaLongTerm)
	{
		char Buf[1024];
		itoa((int)(1.0f / Sp.DeltaLongTerm), Buf, 10);
		strcat(Buf, " fps");
		I3D::Text::DrawString(&DS, Buf, 0.5f, 0.97f, 0.05f, 0.05f);
	}
}

bool Game::GetIngame(void) const
{
	return State == GAME_STATE_ACTIVE || State == GAME_STATE_DONE;
}

void Game::DebugScreenRender(I3D::D3DStruct *DS)
{
	char Str[1024];
	sprintf(Str, "t = %f\n%s", Sp.GameTime, DebugInfo());
	I3D::Text::DrawString(DS, Str, -1.0f, 0.96f, 0.04f, 0.04f, 1.0f);
}

bool Game::GoalComplete(IGAME::MissionLib::Goal *Goal)
{
	int i;
	bool Complete;
	IGAME::Tank *Tnk;
	switch(Goal->Type)
	{
	case IGAME::MLIB_GOAL_TRUE:
		return true;
	case IGAME::MLIB_GOAL_FALSE:
		return false;
	case IGAME::MLIB_GOAL_AND:
		Complete = true;
		for(i = 0; i < Goal->Params.Count(); i++)
			if(!GoalComplete(Goal->Params[i]))
				Complete = false;
		return Complete;
		break;
	case IGAME::MLIB_GOAL_OR:
		Complete = false;
		for(i = 0; i < Goal->Params.Count(); i++)
			if(GoalComplete(Goal->Params[i]))
				Complete = true;
		return Complete;
		break;
	case IGAME::MLIB_GOAL_DESTROY:
		return !Wrld.GetTank(Goal->Object);
		break;
	case IGAME::MLIB_GOAL_DESTROYALL:
		return Wrld.GetTankNumber(Goal->Owner) == 0;
		break;
	case IGAME::MLIB_GOAL_TIME:
		return DS.Sp->GameTime > MissionStartTime + Goal->Value;
		break;
	case IGAME::MLIB_GOAL_STORY:
		if(StoryItem)
			return !stricmp(StoryItem->Name, Goal->Object);
		else
			return *Goal->Object == 0;
		break;
	case IGAME::MLIB_GOAL_POS:
		{
			Tnk = Wrld.GetTank(Goal->Object);
			return Tnk && (Tnk->GetPos().x > Goal->x1 && Tnk->GetPos().x < Goal->x2 && Tnk->GetPos().z > Goal->z1 && Tnk->GetPos().z < Goal->z2);
			break;
		}
	case IGAME::MLIB_GOAL_OWNERPOS:
		{
			IMATH::Vector2DXZ Area[2];
			Area[0].x = Goal->x1;
			Area[0].z = Goal->z1;
			Area[1].x = Goal->x2;
			Area[1].z = Goal->z2;
			return Wrld.GetTankNumber(Goal->Owner, Area) != 0;
			break;
		}
	default:
		throw("Game::GoalComplete: invalid mission goal type");
		return false;
	}
}

void Game::CheckEvents(void)
{
	int State = 0;
/*	static float LastEvents = 0;
	if(DS.Sp->GameTime - LastEvents > 2.0f)
	{
		LastEvents = DS.Sp->GameTime;*/
		try
		{
			State = 1;
			for(int i = 0; i < MLib.Missions[Mission]->Events.Count(); ++i)
			{
				State = 2;
				if(GoalComplete(MLib.Missions[Mission]->GetGoal(MLib.Missions[Mission]->Events[i]->Trigger)))
				{
					State = 3;
					ICommand.Command(MLib.Missions[Mission]->Events[i]->Action);
					State = 4;
				}
				State = 5;
			}
			State = 6;
		}
		catch(char *s)
		{
			throw s;
		}
		catch(...)
		{
			char str[1024];
			sprintf(str, "Game::CheckEvents#%d", State);
			throw str;
		}
	/*	if(GoalComplete(MLib.Missions[Mission]->MainGoal))
		{
			ICommand.Command("MissionComplete");
		}
		if(GoalComplete(MLib.Missions[Mission]->MainGoalEnemy))
		{
			ICommand.Command("MissionFailed");
		}*/
	//}
}

void Game::CheckAirfields(void)
{
	const int PointsAirfield = 10;
	ISL::DynamicArray<IGAME::Airfield *> *Airfields = &MLib.Missions[Mission]->Airfields;
	for(int i = 0; i < Airfields->Count(); ++i)
	{
		IMATH::Vector2DXZ Area[2];
		Area[0].x = (*Airfields)[i]->GetPos().x - 144;
		Area[0].z = (*Airfields)[i]->GetPos().z - 48;
		Area[1].x = (*Airfields)[i]->GetPos().x + 144;
		Area[1].z = (*Airfields)[i]->GetPos().z + 48;
		int Player = Wrld.GetTankNumber(2, Area);
		int Enemy = Wrld.GetTankNumber(3, Area);
		if(Player && !Enemy)
			(*Airfields)[i]->SetOwner(2);
		else if(!Player && Enemy)
			(*Airfields)[i]->SetOwner(3);
		for(int Owner = 2; Owner <= 3; ++Owner)
			if((*Airfields)[i]->GetOwner() == Owner && (*Airfields)[i]->GetPoints(Owner) >= PointsAirfield && (DS.Sp->GameTime - (*Airfields)[i]->GetLastPlane()) > 40)
			{
				(*Airfields)[i]->SetPoints(Owner, (*Airfields)[i]->GetPoints(Owner) - PointsAirfield);
				(*Airfields)[i]->SetLastPlane(DS.Sp->GameTime);
				// spawn reinforcements
				IGAME::SpawnInterface::SpawnInterfaceSpecs SS;
				SS.Type = IGAME::IDTYPE_TANK;
				SS.ID = ((MLib.Missions[Mission]->Side == IGAME::SIDE_ISRAEL && (*Airfields)[i]->GetOwner() == 2)
					|| (MLib.Missions[Mission]->Side != IGAME::SIDE_ISRAEL && (*Airfields)[i]->GetOwner() != 2)) ? IGAME::IDT_TANK_US_C160 : IGAME::IDT_TANK_PAL_C160;
				SS.Pos = IMATH::Vector(-Wrld.GetSize() / 2.0f, 30, (*Airfields)[i]->GetLandingPos().z);
				SS.Owner = Owner;
				SS.Commandable = false;
				SS.Angle = 0.5f * PI;
				SS.AI = IGAME::Tank::TNK_AI_LAND;
				SS.TargetX = (*Airfields)[i]->GetLandingPos().x + 0.01f * (rand() % 101 - 50);
				SS.TargetZ = (*Airfields)[i]->GetLandingPos().z + 0.01f * (rand() % 101 - 50);
				Wrld.GetSpawnInterface()->Spawn(&SS);
			}
	}
}

void Game::MainLoop(void)
{
	int Stage = 0;
	try
	{
		Prf.Stop(7);
		if(FlagFullscreen)
		{
			Wnd.ToggleFullscreen();
			FlagFullscreen = false;
		}
		if(Reinitialize)
		{
			Reinitialize = false;
			Init3D(true);
		}
		Prf.Stop(10);

		if(DS.D3DDevice && DS.D3DDevice->TestCooperativeLevel() == D3D_OK)
		{
			Prf.Stop(10);
			Stage = 1;
			Prf.Start(0);
			// handlers
			EventHandler();
			MouseHandler();
			// end of debug
			const bool RenderWorld = !StoryItem || StoryItem->Render;
			const bool RenderMouse = true;//!StoryItem || StoryItem->Render || Menu.GetVisible();
			// physics
			Sp.Update();
			Prf.Stop(0);
			Stage = 2;
			Prf.Start(8);
			CheckStory();
			Stage = 51;
			// mouse cursor
			if(GetIngame())
			{
				switch(Wrld.GetMouseTarget(Mouse.GetPos()))
				{
				case IGAME::World::MOUSE_TARGET_NOTHING:
					Mouse.SetCursorTexture(MouseTextures[0]);
					break;
				case IGAME::World::MOUSE_TARGET_IMPASSABLE:
					Mouse.SetCursorTexture(MouseTextures[1]);
					break;
				case IGAME::World::MOUSE_TARGET_FRIEND:
					Mouse.SetCursorTexture(MouseTextures[2]);
					break;
				case IGAME::World::MOUSE_TARGET_ENEMY:
					Mouse.SetCursorTexture(MouseTextures[3]);
					break;
				}
			}
			else
				Mouse.SetStdCursor();
			Stage = 52;
			Prf.Stop(8);
			if(GetIngame())
			{
				// debug: pfx at mouse position
				/*IMATH::Vector2DXZ Pos = Wrld.GetXZMousePos(IMATH::Vector2D(Mouse.GetX(), Mouse.GetY()));
				// spawn particle effect
				IGAME::SpawnInterface::SpawnInterfaceSpecs SS;
				SS.Type = IGAME::IDTYPE_PARTICLEFX;
				SS.ID = 4;
				SS.Pos = IMATH::Vector(Pos.x, Wrld.GetSurfaceY(Pos.x, Pos.z), Pos.z);
				Wrld.GetSpawnInterface()->Spawn(&SS);*/
				Stage = 3;
				if(RenderWorld)
				{
					// physics
					Prf.Start(1);
					Wrld.Physics(&DS);
					Prf.Stop(1);
					Stage = 4;
					Prf.Start(2);
					Stage = 41;
					static float LastEvents = 0;
					if(DS.Sp->GameTime - LastEvents > 1.0f)
					{
						if(State == GAME_STATE_ACTIVE)
							CheckEvents();
						CheckAirfields();
						LastEvents = DS.Sp->GameTime;
					}
					Prf.Stop(2);
					Stage = 5;
				}
			}
			Stage = 50;
			Prf.Start(9);
			// rendering
			HRESULT BeginSceneResult = DS.D3DDevice->BeginScene();
			if(BeginSceneResult == D3D_OK)
			{
				Prf.Stop(9);
				Stage = 53;
				Prf.Start(3);
				Stage = 6;
				ClearScreen();
				Prf.Stop(3);
				Stage = 7;
				if(GetIngame())
				{
					if(RenderWorld)
					{
						Stage = 70;
						Prf.Start(4);
						// setup for rendering
						Set3DMatrices();
						Stage = 8;
						// render
						Wrld.Render(&DS);
						Prf.Stop(4);
						Stage = 9;
					}
					// story
					if(StoryItem)
					{
						Stage = 75;
						// destroy light
						Wrld.DestroyLight(&DS);
						Stage = 76;
						StoryRnd.Render(&DS);
						Stage = 77;
						// reset
						Stage = 78;
						Wrld.InitLight(&DS);
						Stage = 79;
						if(D3DFlags & GAME_D3DFLAG_WIREFRAME)
							DS.D3DDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
					}
					// spawn pfx at mouse pos
					/*static float LastSpawn = DS.Sp->GameTime;
					if(DS.Sp->GameTime > LastSpawn + 0.1f)
					{
						IGAME::SpawnInterface::SpawnInterfaceSpecs SS;
						SS.Type = IGAME::IDTYPE_PARTICLEFX;
						SS.ID = IGAME::IDPFX_SMALLRISINGSMOKE;
						IMATH::Vector2D Pos = Mouse.GetPos();
						IMATH::Vector2DXZ XZPos = Wrld.GetSurfaceXZ(Pos);
						SS.Pos = IMATH::Vector(XZPos.x, Wrld.GetSurfaceY(XZPos.x, XZPos.z), XZPos.z);
						Wrld.GetSpawnInterface()->Spawn(&SS);
						LastSpawn = DS.Sp->GameTime;
					}*/
				}
				Stage = 20;

				// overlay
				if(State == GAME_STATE_CREDITS)
				{
					Credits.Render(&DS);
				}
				Prf.Start(5);
				Stage = 21;
				Menu.Render(&DS);
				Stage = 22;
				DisplayFPS();
				Stage = 23;
				MaximizeButton.Render(&DS);
				Stage = 24;
				if(RenderMouse)
				{
					Mouse.Render(&DS);
				}
				Stage = 25;
				Prf.Stop(5);
				Prf.Render(&DS);
				Stage = 10;
				// debug
				if(Flags & FLAG_DEBUG)
				{
					DebugScreenRender(&DS);
				}
				// display
				Prf.Start(6);
				DS.D3DDevice->EndScene();
				if(DS.D3DDevice->Present(NULL, NULL, NULL, NULL) == D3DERR_DEVICELOST)
					Wnd.ResetDirect3D();
				Prf.Stop(6);
				Stage = 11;
			}
			else
			{
				assert(BeginSceneResult != D3DERR_INVALIDCALL);
				Prf.Stop(9);
			}
			Stage = 12;
		}
		else
			Prf.Stop(10);
		Prf.Start(7);
	}
	catch(char *s)
	{
		throw(s);
	}
	catch(...)
	{
		// unknown exceptions
		char Str[1024];
		sprintf(Str, "Unknown Exception in Game::MainLoop#%d", Stage);
		throw(Str);
	}
}

}
