/***********************************************************/
/*                                                         */
/* MathStructures.h                                        */
/*                                                         */
/* created 22.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef MATHSTRUCTURES_H
#define MATHSTRUCTURES_H

#include <math.h>

// www.infinitecode.com math library
namespace IMATH
{

class Vector2D
{
public:
	union
	{
		struct
		{
			float x, y;
		};
		float Data[2];
	};

	Vector2D() {}

	Vector2D(float x, float y)
	{
		this->x = x;
		this->y = y;
	}

	Vector2D &operator += (const Vector2D &V)
	{
		x += V.x;
		y += V.y;
		return *this;
	}

	const Vector2D operator - (const Vector2D &V) const
	{
		return Vector2D(x - V.x, y - V.y);
	}

	const Vector2D operator + (const Vector2D &V) const
	{
		return Vector2D(x + V.x, y + V.y);
	}

	float Length(void) const
	{
		return (float)sqrt(x * x + y * y);
	}

	float LengthSq(void) const
	{
		return x * x + y * y;
	}
};

class Vector2DXZ
{
public:
	union
	{
		struct
		{
			float x, z;
		};
		float Data[2];
	};

	Vector2DXZ() {}

	Vector2DXZ(float x, float z)
	{
		this->x = x;
		this->z = z;
	}

	Vector2DXZ Normalize(void) const
	{
		float VLength = Length();
		if(VLength)
		{
			float OOLength = 1.0f / VLength;
			return Vector2DXZ(OOLength * x, OOLength * z);
		}
		else
			return Vector2DXZ(0, 0);
	}

	bool operator == (Vector2DXZ V)
	{
		return x == V.x && z == V.z;
	}

	const Vector2DXZ operator - (const Vector2DXZ &V) const
	{
		return Vector2DXZ(x - V.x, z - V.z);
	}

	const Vector2DXZ operator + (const Vector2DXZ &V) const
	{
		return Vector2DXZ(x + V.x, z + V.z);
	}

	Vector2DXZ &operator += (const Vector2DXZ &V)
	{
		x += V.x;
		z += V.z;
		return *this;
	}

	Vector2DXZ &operator -= (const Vector2DXZ &V)
	{
		x -= V.x;
		z -= V.z;
		return *this;
	}

	Vector2DXZ operator * (float Scalar)
	{
		return Vector2DXZ(Scalar * x, Scalar * z);
	}

	Vector2DXZ operator / (const float Scalar) const
	{
		float Rec;
		if(Scalar)
			Rec = 1 / Scalar;
		else
			Rec = 1;
		return Vector2DXZ(x * Rec, z * Rec);
	}

	const Vector2DXZ operator - () const
	{
		return Vector2DXZ(-x, -z);
	}

	float Length(void) const
	{
		return (float)sqrt(x * x + z * z);
	}

	float LengthSq(void) const
	{
		return x * x + z * z;
	}
};

class Vector
{
public:
	union
	{
		struct
		{
			float x, y, z;
		};
		float Data[3];
	};

	Vector() {}

	Vector(float x, float y, float z)
	{
		this->x = x;
		this->y = y;
		this->z = z;
	}

	float Length(void) const
	{
		return (float)sqrt(x * x + y * y + z * z);
	}

	float LengthSq(void)
	{
		return x * x + y * y + z * z;
	}

	static Vector CrossProduct(Vector V1, Vector V2)
	{
		return Vector(V1.y * V2.z - V1.z * V2.y, V1.z * V2.x - V1.x * V2.z, V1.x * V2.y - V1.y * V2.x);
	}

	static Vector Normalize(Vector V)
	{
		float Len = (float)sqrt(V.x * V.x + V.y * V.y + V.z * V.z);
		if(Len)
		{
			float OOLen = 1 / Len;
			return Vector(OOLen * V.x, OOLen * V.y, OOLen * V.z);
		}
		else
			return Vector(0, 0, 0);
	}

	Vector Normalize(void)
	{
		float Len = (float)sqrt(x * x + y * y + z * z);
		if(Len)
		{
			float OOLen = 1 / Len;
			return Vector(OOLen * x, OOLen * y, OOLen * z);
		}
		else
			return Vector(0, 0, 0);
	}

	Vector operator + (const Vector &V2) const
	{
		return Vector(x + V2.x, y + V2.y, z + V2.z);
	}

	Vector operator * (const float Scalar) const
	{
		return Vector(x * Scalar, y * Scalar, z * Scalar);
	}

	Vector operator / (const float Scalar) const
	{
		float Rec;
		if(Scalar)
			Rec = 1 / Scalar;
		else
			Rec = 1;
		return Vector(x * Rec, y * Rec, z * Rec);
	}

	Vector operator - (const Vector& V2) const
	{
		return Vector(x - V2.x, y - V2.y, z - V2.z);
	}

	Vector operator += (const Vector& V2)
	{
		x += V2.x;
		y += V2.y;
		z += V2.z;
		return *this;
	}

	Vector operator -= (const Vector& V2)
	{
		x -= V2.x;
		y -= V2.y;
		z -= V2.z;
		return *this;
	}

	Vector operator - () const
	{
		return Vector(-x, -y, -z);
	}

	bool operator == (const Vector V2) const
	{
		return x == V2.x && y == V2.y && z == V2.z;
	}

//	friend Vector ::operator * (float Scalar, Vector& V);
};

// multiplication with a scalar value
IMATH::Vector operator * (float Scalar, const IMATH::Vector &V);
IMATH::Vector2D operator * (float Scalar, const IMATH::Vector2D &V);
IMATH::Vector2DXZ operator * (float Scalar, const IMATH::Vector2DXZ &V);

}

#endif
