/***********************************************************/
/*                                                         */
/* MissionEditor.h                                         */
/*                                                         */
/* created 15.03.2003                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef MISSIONEDITOR_H
#define MISSIONEDITOR_H

#include "D3DWindow.h"
#include "D3DStruct.h"
#include "EventListener.h"
#include "DynamicArray.h"
#include "Model.h"
#include "RenderableModel.h"
#include "Input.h"
#include "MouseCursor.h"
#include "Terrain.h"
#include "QuadTree.h"
#include "Speed.h"
#include "MissionLib.h"
#include "SpawnInterface.h"
#include "TankLib.h"
#include "World.h"
#include "Brush.h"
#include "MOPropertiesIf.h"
#include "DecalLib.h"
#include "Button.h"
#include "TextField.h"
#include "StoryRenderer.h"
#include "DialogWindow.h"
#include "Menu.h"

// www.infinitecode.com game library
namespace IGAME
{

enum
{
	GAME_D3DFLAG_WIREFRAME = 1,
	GAME_D3DFLAG_NOCULLING = 2,
	GAME_D3DFLAG_COLORKEY = 4,
	GAME_D3DFLAG_ALPHABLEND = 8,
	GAME_D3DFLAG_WHITEBACK = 16
};

class MissionEditor: public IEV::EventListener<I3D::D3DWEvent>, public IEV::EventListener<const char *>
{
	friend class GenMap;
	friend class LandscapeTree;

	class MOPropertiesImpl: public MOPropertiesIf
	{
		MissionEditor *ME;

	public:
		void SetBase(MissionEditor *ME)
		{
			MOPropertiesImpl::ME = ME;
		}

		virtual void AIMove(const char *Name, IMATH::Vector2DXZ Pos);
		virtual void AIAttack(const char *Name, const char *Target);
		virtual void SetPos(const char *Name, IMATH::Vector Pos);
		virtual void SetWayPos(const char *Name, int ID, IMATH::Vector2DXZ Pos);
		void SetTerrainPos(const char *Name, int ID, IMATH::Vector2DXZ Pos);
	} MOPropImpl;

	class MouseEvent
	{
	public:
		enum
		{
			CLICK,
			RIGHTCLICK,
			DRAG_END,
			DRAG_START,
			DRAG
		} Event;
		IMATH::Vector2D Pos, Down;
	};

	class DialogHandler: public ITL::DialogWindow::Handler
	{
	protected:
		ITL::DialogWindow *Dlg;
		MissionEditor *ME;

	public:
		DialogHandler(ITL::DialogWindow *DW, MissionEditor *ME)
			: Dlg(DW)
		{
			DialogHandler::ME = ME;
		}

		virtual void OnEvent(ITL::DlgWndEvent Ev) {}
	};

	class StoryItemDialogHandler: public DialogHandler
	{
		int ID;

	public:
		StoryItemDialogHandler(int ID, ITL::DialogWindow *DW, MissionEditor *ME)
			: DialogHandler(DW, ME)
		{
			StoryItemDialogHandler::ID = ID;
		}

		virtual void OnEvent(ITL::DlgWndEvent Ev);
	};

	class OrdersDialogHandler: public DialogHandler
	{
	public:
		OrdersDialogHandler(ITL::DialogWindow *DW, MissionEditor *ME)
			: DialogHandler(DW, ME) {}

		virtual void OnEvent(ITL::DlgWndEvent Ev);
	};

	class CIData: public DataInterface
	{
		MissionEditor *ME;

	public:
		void Create(MissionEditor *ME)
		{
			CIData::ME = ME;
		}

		virtual const char *GetData(const char *Str);
	} IData;

	struct FreeArea
	{
		bool Successful;
		IMATH::Vector2DXZ Area[2];
	};

	// street direction info
	struct StreetDirectionInfo
	{
		enum
		{
			OBSTACLE_IN_WORLD = 1,
			END_OF_WORLD = 2
		};
		float MaxLength;
		int BlockingID, BlockingWorldID;
	};

	IMATH::Vector2D MouseDown;
	ISL::DynamicArray<MouseEvent> MouseEvents;
	bool MouseDrag;

	I3D::D3DWindow Wnd;
	I3D::D3DStruct DS;
	int D3DFlags;
	HINSTANCE Instance;
	//int BackgroundColor;
	IIN::Input In;
	I3D::MouseCursor Mouse;
	ITL::Speed Sp;
	ISL::DynamicArray<I3D::D3DWEvent, 0, 4> Events;
	// libraries
	IGAME::MissionLib MLib;
	int Mission;
	IGAME::Brush Brush;
	TankLib TLib;
	DecalLib DLib;
	IGAME::World Wrld;
	LPDIRECT3DTEXTURE9 MouseTextures[4];
	I3D::Button MaximizeButton;
	IGAME::StoryRenderer StoryRnd;
	bool Reinitialize;
	I3D::TextField InfoBox;
	IGAME::MissionLib::StoryItem *StoryItem;
	float StoryTime;
	ISL::DynamicArray<ITL::DialogWindow::Handler *> Handlers;
	ISL::DynamicArray<ITL::DialogWindow *> Dialogs;
	I3D::Menu Menu;
	// mission generation
public:
	struct GenStreet
	{
		int x, z, Dir;
		GenStreet() {}
		GenStreet(int new_x, int new_z, int new_Dir): x(new_x), z(new_z), Dir(new_Dir) {}
		bool operator ==(const GenStreet &a) {return a.x == x && a.z == z && a.Dir == Dir;}
		bool is_connected_with(GenStreet &a) {return (a.x == x && a.z == z) || ((abs(a.x - x) == 32 && a.z == z && (a.Dir & Dir & World::STREET_HORZ)) || (abs(a.z - z) == 32 && a.x == x && (a.Dir & Dir & World::STREET_VERT)));}
		int GetDir() { return Dir; }
	};
private:
	friend class GenStreetC;
	friend struct Conn;
	ISL::DynamicArray<GenStreet> GenStreets;
	ofstream GenerationLog;

	virtual void OnEvent(I3D::D3DWEvent Data);
	void EventHandler(void);
	void KeyboardInput(char c);
	void MouseClick(void);
	void Init3D(bool recurse);
	void Create3D(void);
	void MenuCheck(int ID);
	void MenuCheck(int ID, bool Value);
	void MouseHandler(void);
	void Scrolling(void);
	void LoadingScreen(const char *Str);
	void DisplayFPS(void);
	void LoadMission(void);
	void CreateWindowMenu(HWND Window);
	void NewMission(void);
	void Compile();
	void RenderInfoBox(I3D::D3DStruct *DS);
	void LoadStory(const char *Name);
	void CheckStory();
	void StoryItemDialog(int ID);
	void OrdersDialog();
	void GenerateStreet(GenMap *Map, int x1, int z1, int x2, int z2);
	//bool GenerateAbstractStreet(GenMap *Map, bool Linked, bool Middle);
	IMATH::Vector2DXZ FindStreetLink(GenMap *Map);
	void GenerateLinkedStreet(GenMap *Map, const IMATH::Vector2DXZ &Link);
	bool GenerateMainStreet(GenMap *Map);
	bool GenerateSideStreet(GenMap *Map);
	int MissionEditor::GetBuildingID(int Type, int Owner);
	void GenerateBuilding(GenMap *Map, int Type, int Owner);
	void GenerateCity(GenMap *Map);
	void GrowCity(GenMap *Map);
	void GrowMainStreet(GenMap *Map, bool Highway);
	void GrowStreets(GenMap *Map);
	StreetDirectionInfo MissionEditor::TestStreetDirection(GenMap *Map, const IMATH::Vector2DXZ &Pos, const IMATH::Vector2DXZ &Direction);
	void GenerateLandscape(GenMap *Map);
	void GenerateBaseStreets(GenMap *Map);
	void GenerateBaseBuildings(GenMap *Map, int Owner);
	bool GenerateAirbaseAirfield(GenMap *Map, int Owner);
	void GenerateAirbaseBuildings(GenMap *Map, int Owner);
	void GenerateBase(GenMap *Map, int Owner);
	void GenerateAirbase(GenMap *Map, int Owner);
	FreeArea FindFreeAreaXZ(GenMap *Map, int MinSizeX, int MinSizeZ, int MaxSizeX, int MaxSizeZ);
	FreeArea FindFreeArea(GenMap *Map, int MinSize, int MaxSize);
	bool GenerateAreaAndAirbase(GenMap *Map, int Owner);
	bool IsConnected(GenMap *Map, GenStreet &a, GenStreet &b);
	void GenerateConnection(GenMap *Map, GenStreet &a_world, GenStreet &b_world);
	void GenerateConnectedStreets(GenMap *Map);
	void GenerateMissionCaptureAirbases(GenMap *Map);
	void GenerateMissionProtectConvoy(GenMap *Map);
	void GenerateMissionDestroyConvoy(GenMap *Map);
	void GenerateMissionCaptureCity(GenMap *Map);
	void GenerateMissionCaptureBase(GenMap *Map);
	void GenerateMissionMovePrototype(GenMap *Map);
	void GenerateMissionMoveFleet(GenMap *Map);
	void GenerateMissionDestroyFleet(GenMap *Map);
	void GenerateTank(GenMap *Map, int Type, int Owner);

public:
	MissionEditor(void)
		:D3DFlags(0), GenerationLog("generation_log.txt")
	{
		MOPropImpl.SetBase(this);
		//ShowCursor(false);
	}

	~MissionEditor(void)
	{
		Wnd.DeleteListener(this);
		//ShowCursor(true);
	}

	void OnEvent(const char *Event)
	{
		LoadingScreen(Event);
	}

	std::ofstream *GetGenerationLog()
	{
		return &GenerationLog;
	}

	void SetPause(bool Pause) { Sp.Paused = Pause; }
	bool GetPause() { return Sp.Paused; }
	void SetNextBrushID(int Value) { Brush.SetNextID(Value); }
	int GetNextBrushID() { return Brush.GetNextID(); }

	void Create(HINSTANCE ThisInstance, int WindowStyle);
	// this is called continuously while the application is running
	void MainLoop(void);
	void OnViewChange(void);
	void AddTerrainNoHMap(I3D::TerrainDsc::TerrainDscSpecs &Specs) { Wrld.AddTerrainNoHMap(&Specs); }
};


#define MAP_RES 4

class GenMap
{
	int *Map;
	bool *StreetMap;

public:
	int Width, Height;
	MissionEditor *Editor;
	IMATH::Vector2DXZ Pos[2];

	GenMap(const IMATH::Vector2DXZ *Pos, MissionEditor *ME)
	{
		/*GenMap::Pos[0].x = (float)((((int)Pos[0].x + 31) >> 5) << 5);
		GenMap::Pos[0].z = (float)((((int)Pos[0].z + 31) >> 5) << 5);
		GenMap::Pos[1].x = (float)((((int)Pos[1].x) >> 5) << 5);
		GenMap::Pos[1].z = (float)((((int)Pos[1].z) >> 5) << 5);*/
		GenMap::Pos[0].x = (float)(int)Pos[0].x;
		GenMap::Pos[0].z = (float)(int)Pos[0].z;
		GenMap::Pos[1].x = (float)(int)Pos[1].x;
		GenMap::Pos[1].z = (float)(int)Pos[1].z;
		// order
		if(GenMap::Pos[0].x > GenMap::Pos[1].x)
		{
			float swap = GenMap::Pos[0].x;
			GenMap::Pos[0].x = GenMap::Pos[1].x;
			GenMap::Pos[1].x = swap;
		}
		if(GenMap::Pos[0].z > GenMap::Pos[1].z)
		{
			float swap = GenMap::Pos[0].z;
			GenMap::Pos[0].z = GenMap::Pos[1].z;
			GenMap::Pos[1].z = swap;
		}
		*ME->GetGenerationLog() << "pos: " << (int)GenMap::Pos[0].x << ", " << (int)GenMap::Pos[0].z << "->" << (int)GenMap::Pos[1].x << ", " << (int)GenMap::Pos[1].z << std::endl;
		Editor = ME;
		//Width = ((int)(Pos[1].x - Pos[0].x) >> 5) << 5;
		Width = (int)(GenMap::Pos[1].x - GenMap::Pos[0].x);
		if(Width < MAP_RES)
			Width = MAP_RES;
		//Height = ((int)(Pos[1].z - Pos[0].z) >> 5) << 5;
		Height = (int)(GenMap::Pos[1].z - GenMap::Pos[0].z);
		if(Height < MAP_RES)
			Height = MAP_RES;
		*ME->GetGenerationLog() << "width x height: " << Width << "x" << Height << std::endl;
		Map = new int[(Width / MAP_RES) * (Height / MAP_RES)];
		StreetMap = new bool[(Width / MAP_RES) * (Height / MAP_RES)];
		FillMap();
		NearMap();
	}

	~GenMap()
	{
		delete[] Map;
	}

	void FillMap()
	{
		for(int i = 0; i < Width / MAP_RES; ++i)
			for(int j = 0; j < Height / MAP_RES; ++j)
				Map[j * (Width / MAP_RES) + i] = Editor->Wrld.GetLandscape(Pos[0] + IMATH::Vector2DXZ((float)(i * MAP_RES), (float)(j * MAP_RES)));
	}

	void NearMap()
	{
		for(int i = 0; i < Width / MAP_RES; ++i)
			for(int j = 0; j < Height / MAP_RES; ++j)
				StreetMap[j * (Width / MAP_RES) + i] = Editor->Wrld.GetStreetAccessible(Pos[0] + IMATH::Vector2DXZ((float)(i * MAP_RES), (float)(j * MAP_RES)));
	}

	MissionEditor *GetME()
	{
		return Editor;
	}

	Decal *GetDecal(int x, int y)
	{
		return Editor->Wrld.GetDecal(Pos[0] + IMATH::Vector2DXZ((float)x, (float)y));
	}

	int Get(int x, int y) const
	{
		x /= MAP_RES;
		y /= MAP_RES;
		if(x >= 0 && x < Width / MAP_RES && y >= 0 && y < Height / MAP_RES)
		{
			return Map[y * (Width / MAP_RES) + x];
		}
		else
			return World::NOTHING;
	}

	void Set(int x, int y, int Value) const
	{
		x /= MAP_RES;
		y /= MAP_RES;
		if(x >= 0 && x < Width / MAP_RES && y >= 0 && y < Height / MAP_RES)
		{
			Map[y * (Width / MAP_RES) + x] = '0' + Value;
		}
	}

	int GetStreet(int x, int y)
	{
		x /= MAP_RES;
		y /= MAP_RES;
		if(x >= 0 && x < Width / MAP_RES && y >= 0 && y < Height / MAP_RES)
		{
			return StreetMap[y * (Width / MAP_RES) + x];
		}
		else
			return false;
	}

	int GetStreetXOffset()
	{
		int x = (int)Pos[0].x % 32;
		if(x)
			return 32 - x;
		else
			return 0;
	}

	int GetStreetZOffset()
	{
		int z = (int)Pos[0].z % 32;
		if(z)
			return 32 - z;
		else
			return 0;
	}

	void AddDecal(Decal *D)
	{
		IMATH::Vector2DXZ Pos = D->GetPos() - GenMap::Pos[0];
		for(int i = (int)(Pos.x - D->GetSizeX() / 2); i <= (int)(Pos.x + D->GetSizeX() / 2); ++i)
			for(int j = (int)(Pos.z - D->GetSizeZ() / 2); j <= (int)(Pos.z + D->GetSizeZ() / 2); ++j)
                if(i >= 0 && i / MAP_RES < Width / MAP_RES && j >= 0 && j / MAP_RES < Height / MAP_RES)
					if(D->IsStreet())
					{
						switch(D->GetID())
						{
						case 0:
							Map[(j / MAP_RES) * (Width / MAP_RES) + i / MAP_RES] = World::STREET_HORZ;
							break;
						case 1:
							Map[(j / MAP_RES) * (Width / MAP_RES) + i / MAP_RES] = World::STREET_VERT;
							break;
						case 2:
							Map[(j / MAP_RES) * (Width / MAP_RES) + i / MAP_RES] = World::STREET_CROSS;
							break;
						}
					}
					else
						Map[(j / MAP_RES) * (Width / MAP_RES) + i / MAP_RES] = World::DECAL;
	}

	void AddWay(I3D::Way *W)
	{
		IMATH::Vector2DXZ Diff = W->GetPos(1) - W->GetPos(0);
		IMATH::Vector2DXZ Orth(Diff.z, -Diff.x);
		float Length = Diff.Length();
		for(int i = 0; i < Length; i++)
		{
			IMATH::Vector2DXZ V = W->GetPos(0) + ((float)i / Length) * Diff - Pos[0];
			for(int j = -(int)W->GetSize() - 4; j < (int)W->GetSize() + 4; j++)
			{
				IMATH::Vector2DXZ Vt = V + Orth.Normalize() * (float)j;
				int x = (int)Vt.x;
				int z = (int)Vt.z;
				if(x >= 0 && x / MAP_RES < Width / MAP_RES && z >= 0 && z / MAP_RES < Height / MAP_RES)
					Map[(z / MAP_RES) * (Width / MAP_RES) + x / MAP_RES] = World::WAY;
			}
		}
	}

	void AddTank(Tank *T)
	{
		IMATH::Vector2DXZ Pos = IMATH::Vector2DXZ(T->GetPos().x, T->GetPos().z) - GenMap::Pos[0];
		for(int i = (int)(Pos.x - T->GetSize()) - 4; i <= (int)(Pos.x + T->GetSize()) + 4; ++i)
			for(int j = (int)(Pos.z - T->GetSize()) - 4; j <= (int)(Pos.z + T->GetSize()) + 4; ++j)
                if(i >= 0 && i / MAP_RES < Width / MAP_RES && j >= 0 && j / MAP_RES < Height / MAP_RES)
					Map[(j / MAP_RES) * (Width / MAP_RES) + i / MAP_RES] = World::VEHICLE;
	}
};

std::ostream &operator << (std::ostream &os, const GenMap &Map);

}

#endif
