/***********************************************************/
/*                                                         */
/* MissionLib.h                                            */
/*                                                         */
/* created 01.03.2003                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef MISSIONLIB_H
#define MISSIONLIB_H

#include "IniLoader.h"
#include "DynamicArray.h"
#include "MissionObject.h"
#include "MathStructures.h"
#include "Way.h"
#include "TerrainDsc.h"
#include "Airfield.h"
#include "Decal.h"

// www.infinitecode.com game library
namespace IGAME
{

class MissionLib
{
public:
	#define MISL_STRMEM 1024

	class Goal
	{
	public:
		Goal(void): Owner(-1), Type(-1)
		{
			*Object = 0;
		}

		int Type, Owner;
		ISL::DynamicArray<char *, 0 ,1> ParamsTxt; // pointers are only valid in MissionLib::Create!
		ISL::DynamicArray<Goal *, 0 ,1> Params;
		char Name[MISL_STRMEM], Object[MISL_STRMEM];
		float Value, x1, x2, z1, z2;
	};

	class Event
	{
	public:
		char Name[MISL_STRMEM], Trigger[MISL_STRMEM], Action[MISL_STRMEM];
	};

	class StoryItem
	{
	public:
		int SubType, BumperID, TickerStyle;
		bool Render;
		char Next[MISL_STRMEM], NewsTicker[MISL_STRMEM], Name[MISL_STRMEM], Topic[MISL_STRMEM], Text[MISL_STRMEM];
		float TimeDelta;
		IMATH::Vector Camera, CameraAngle;

		StoryItem(void): TimeDelta(0), Camera(0, 0, 0), CameraAngle(0, 0, 0)
			, BumperID(0), Render(true), TickerStyle(0)
		{
			*Next = 0;
			*NewsTicker = 0;
			*Topic = 0;
			*Text = 0;
		}

		static const char *IniToPrintable(const char *Str);
		static const char *PrintableToIni(const char *Str);
	};

	class Mission
	{
	public:
		enum MissionGoal
		{
			// flags
			MG_PLAYER = 1,
			MG_ENEMY = 2,
			// types
			MG_TYPES_MASK = 4 | 8 | 16 | 32 | 64 | 128,
			MG_DESTROY = 4,
			MG_DESTROYALLPLAYER = 8,
			MG_DESTROYALLENEMY = 16,
			MG_MOVETO = 32,
			MG_ANYPLAYERMOVETO = 64,
			MG_ANYENEMYMOVETO = 128,
		};

		struct MissionObjectSpecs
		{
			int Type, ID, Owner;
			IMATH::Vector2DXZ Pos;
			char Name[1024];
			float Angle;
		};

		char Name[MISL_STRMEM], FileName[MISL_STRMEM], Next[MISL_STRMEM], Orders[MISL_STRMEM];
		ISL::DynamicArray<MissionObject, 0, 1> Objects;
		ISL::DynamicArray<I3D::Way, 0, 1> Ways;
		ISL::DynamicArray<I3D::TerrainDsc, 0, 1> TerrainDscs;
		ISL::DynamicArray<IGAME::Airfield *> Airfields;
		ISL::DynamicArray<IGAME::Decal *> Decals;
		ISL::DynamicArray<Goal *, 0, 1> Goals;
		ISL::DynamicArray<Event *, 0, 1> Events;
		ISL::DynamicArray<StoryItem *, 0, 1> StoryItems;
		Goal *MainGoal, *MainGoalEnemy;
		int NextID, Seed, Side;
		IMATH::Vector InitialCameraPos, InitialCameraAngle;

		Mission(void) :MainGoal(0), NextID(0), InitialCameraPos(0, 0, 0), InitialCameraAngle(0, 0, 0), Seed(0)
		{
			*Next = 0;
			*Orders = 0;
		}

		~Mission(void);

		void SetInitialCameraPos(const IMATH::Vector &V)
		{
			InitialCameraPos = V;
		}

		void SetInitialCameraAngle(const IMATH::Vector &V)
		{
			InitialCameraAngle = V;
		}

		MissionObject *GetMissionObject(const char *Name);
		I3D::Way *GetWay(const char *Name);
		I3D::TerrainDsc *GetTerrainDsc(const char *Name);
		void AddObject(MissionObjectSpecs *MOS);
		void UpdateObject(const char *Name, MissionObjectSpecs *MOS);
		void AddWay(I3D::Way::WaySpecs *WS);
		void AddAirfield(Airfield *AF);
		void AddDecal(Decal *D);
		void AddTerrainDsc(I3D::TerrainDsc::TerrainDscSpecs *TS);
		Event *Mission::GetEventComplete();
		Event *Mission::GetEventFailed();
		Goal *GetGoalComplete(Event *EvComplete);
		Goal *GetGoalFailed(Event *EvFailed);
		Goal *FindGoal(Goal *Tree, bool Conjunctions, int Type, const char *Tank = 0);
		void AddMissionGoal(int Type, const char *Tank = 0, const IMATH::Vector2DXZ *Area = 0);
		void RemoveMissionGoal(int Type, const char *Tank = 0);
		bool TestMissionGoal(int Type, const char *Tank = 0);
		void DeleteObject(const char *Name);
		void Save(const char *Directory);
		int GetStoryNum(const char *Name);
		Goal *GetGoal(const char *Name);
	};

	char FirstMissionPA[MISL_STRMEM], FirstMissionIS[MISL_STRMEM];
	ISL::DynamicArray<Mission *, 0, 1> Missions;

	~MissionLib(void);

	void Create(const char *Directory);
	int GetMissionNum(const char *FileName);

};

}

#endif
