/***********************************************************/
/*                                                         */
/* ParticleEffect.h                                        */
/*                                                         */
/* created 17.02.2003                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef PARTICLEEFFECT_H
#define PARTICLEEFFECT_H

#include "MathStructures.h"
#include "ManagedObject.h"
#include "SpaceManagerInterface.h"
#include "ParticleFXLib.h"
#include "RenderStruct.h"
#include "D3DStructures.h"
#include "D3DStruct.h"

// www.infinitecode.com 3d library
namespace I3D
{

class ParticleEffect: public I3D::ManagedObject
{
	static int TypeID;

	IMATH::Vector Pos;
public:

	struct ParticleEffectSpecs
	{
		IMATH::Vector Pos;
		I3D::SpaceManagerInterface *SM;
		class ParticleFXRenderer *PFXRenderer;
		float Time, Delay;
		int ID;
		class WorldPFXInterface *IWorldPFX;
		ParticleFXLib *PFXLib;
		I3D::D3DStruct *DS;
	};

private:
	ParticleEffectSpecs Specs;
	I3D::RenderStruct<I3D::LVertex, I3D_FVFLVERTEX, 4, 4> RS;

public:

	float LifeTime, RelTime, Rotation, RotationDir;
	int Offset;
	bool Loaded;
	LPDIRECT3DTEXTURE9 OverrideTexture;

	virtual void SetPos(const IMATH::Vector &Value)
	{
		Pos = Value;
	}

	virtual const IMATH::Vector &GetPos(void) const
	{
		return Pos;
	}

	void SetTime(float Time)
	{
		Specs.Time = Time;
	}

	virtual void *GetBase(void) const
	{
		return (void *)this;
	}

	float Function(int FctNum, float t)
	{
		switch(FctNum)
		{
		default:
		case 0:	return 1.0f;
				break;
		case 1: return t;
				break;
		case 2: return 1 - t;
				break;
		case 3: return(4 * t * (1 - t));
				break;
		case 4: return 2 * (float)sqrt(t * (1 - t)); // half circle
				break;
		case 5: return 0.3f + 0.7f * t;
				break;
		case 6: return 1.0f - 0.7f * t;
				break;
		}
	}

	void Init3D(D3DStruct *DS, bool recurse)
	{
	}

	void Render()
	{
		ParticleFXLib::Effect *Lib = &Specs.PFXLib->Effects[Specs.ID];
		int XFrames = Lib->XFrames;
		int YFrames = Lib->YFrames;
		float OOXFrames = 1.0f / XFrames;
		float OOYFrames = 1.0f / YFrames;
		unsigned char Light = (int)(0xFF * (Lib->LightFunction ? Function(Lib->LightFunction, RelTime) : 1));
		unsigned int RGBA = Light * 0x01010101;
		int Frame = ((int)(XFrames * YFrames * RelTime) + Offset) % (XFrames * YFrames);
		float FrameX = (float)(Frame % XFrames) * OOXFrames;
		float FrameY = (float)(Frame / XFrames) * OOYFrames;
		IMATH::Vector PFXPos = Specs.DS->ViewMtx * Pos /*+ Vector(0, 0, -8)*/;
		float Size = 8.0f * Lib->Size
			* (Lib->SizeFunction ? Function(Lib->SizeFunction, RelTime) : 1);
		/*float x = Size * (float)cos(Rotation);
		float y = Size * (float)sin(Rotation);*/
		I3D::Matrix M;
		M.SetRotationZMatrix(Rotation, 0, 0);
		M *= I3D::Matrix::ScaleTranslationMatrix(IMATH::Vector(Size, Size, 1), PFXPos);
		Specs.DS->D3DDevice->SetTransform(D3DTS_WORLD, &M);
		// create vertices
		//int FirstVertex = RS->GetVertexNumber();
		RS.RenderList[0].c = RGBA;
		RS.RenderList[0].tu = FrameX;
		RS.RenderList[0].tv = FrameY;
		RS.RenderList[1].c = RGBA;
		RS.RenderList[1].tu = FrameX + OOXFrames;
		RS.RenderList[1].tv = FrameY;
		RS.RenderList[2].c = RGBA;
		RS.RenderList[2].tu = FrameX + OOXFrames;
		RS.RenderList[2].tv = FrameY + OOYFrames;
		RS.RenderList[3].c = RGBA;
		RS.RenderList[3].tu = FrameX;
		RS.RenderList[3].tv = FrameY + OOYFrames;
		RS.CreateVBuf(Specs.DS->D3DDevice);
		RS.RenderIndex(Specs.DS->D3DDevice);
	}

	void RenderToRS(I3D::RenderStruct<I3D::LVertex, I3D_FVFLVERTEX, 64, 64> *RS)
	{
		ParticleFXLib::Effect *Lib = &Specs.PFXLib->Effects[Specs.ID];
		int XFrames = Lib->XFrames;
		int YFrames = Lib->YFrames;
		float OOXFrames = 1.0f / XFrames;
		float OOYFrames = 1.0f / YFrames;
		unsigned char Light = (int)(0xFF * (Lib->LightFunction ? Function(Lib->LightFunction, RelTime) : 1));
		unsigned int RGBA = Light * 0x01010101;
		int Frame = ((int)(XFrames * YFrames * RelTime) + Offset) % (XFrames * YFrames);
		float FrameX = (float)(Frame % XFrames) * OOXFrames;
		float FrameY = (float)(Frame / XFrames) * OOYFrames;
		IMATH::Vector PFXPos = Specs.DS->ViewMtx * Pos /*+ Vector(0, 0, -8)*/;
		float Size = 8.0f * Lib->Size
			* (Lib->SizeFunction ? Function(Lib->SizeFunction, RelTime) : 1);
		float x = Size * (float)cos(Rotation);
		float y = Size * (float)sin(Rotation);
		/*I3D::Matrix M;
		M.SetRotationZMatrix(Rotation, 0, 0);
		M *= I3D::Matrix::ScaleTranslationMatrix(IMATH::Vector(Size, Size, 1), PFXPos);
		Specs.DS->D3DDevice->SetTransform(D3DTS_WORLD, &M);*/
		// create vertices
		int FirstVertex = RS->GetVertexNumber();
		RS->RenderList[FirstVertex + 0].x = PFXPos.x - x - y;
		RS->RenderList[FirstVertex + 0].y = PFXPos.y + x - y;
		RS->RenderList[FirstVertex + 0].z = PFXPos.z;
		RS->RenderList[FirstVertex + 0].c = RGBA;
		RS->RenderList[FirstVertex + 0].tu = FrameX;
		RS->RenderList[FirstVertex + 0].tv = FrameY;
		RS->RenderList[FirstVertex + 1].x = PFXPos.x - x + y;
		RS->RenderList[FirstVertex + 1].y = PFXPos.y - x - y;
		RS->RenderList[FirstVertex + 1].z = PFXPos.z;
		RS->RenderList[FirstVertex + 1].c = RGBA;
		RS->RenderList[FirstVertex + 1].tu = FrameX + OOXFrames;
		RS->RenderList[FirstVertex + 1].tv = FrameY;
		RS->RenderList[FirstVertex + 2].x = PFXPos.x + x + y;
		RS->RenderList[FirstVertex + 2].y = PFXPos.y - x + y;
		RS->RenderList[FirstVertex + 2].z = PFXPos.z;
		RS->RenderList[FirstVertex + 2].c = RGBA;
		RS->RenderList[FirstVertex + 2].tu = FrameX + OOXFrames;
		RS->RenderList[FirstVertex + 2].tv = FrameY + OOYFrames;
		RS->RenderList[FirstVertex + 3].x = PFXPos.x + x - y;
		RS->RenderList[FirstVertex + 3].y = PFXPos.y + x + y;
		RS->RenderList[FirstVertex + 3].z = PFXPos.z;
		RS->RenderList[FirstVertex + 3].c = RGBA;
		RS->RenderList[FirstVertex + 3].tu = FrameX;
		RS->RenderList[FirstVertex + 3].tv = FrameY + OOYFrames;
		//int FirstIndex = RS->GetIndexNumber();
		RS->AddIndex(FirstVertex + 0);
		RS->AddIndex(FirstVertex + 1);
		RS->AddIndex(FirstVertex + 3);
		RS->AddIndex(FirstVertex + 1);
		RS->AddIndex(FirstVertex + 2);
		RS->AddIndex(FirstVertex + 3);
	}

	virtual void SetVisible(bool Visible);

	void SetTexture(LPDIRECT3DTEXTURE9 Texture)
	{
		OverrideTexture = Texture;
	}

	virtual void *GetType(void) const
	{
		return &TypeID;
	}

	int GetID(void) const
	{
		return Specs.ID;
	}

	ParticleEffect(void): Loaded(false) {}
	~ParticleEffect(void);

	void Create(ParticleEffectSpecs *Specs);
	void Physics(D3DStruct *DS);

};

}

#endif
