/***********************************************************/
/*                                                         */
/* QTNode.cpp                                              */
/*                                                         */
/* created 11.08.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#include "QTNode.h"
#include "QuadTree.h"

#include "MathStructures.h"
#include "Waterlevel.h"
#include "TextureManager.h"

#include <fstream>
#include <cmath>

namespace I3D
{

inline float absf(float x)
{
	return x >= 0 ? x : -x;
}

QTNode::QTNode(TreeCreateStruct *TCS)
{
//	int i;
	Flags = TCS->Flags & QT_CHILD_MASK;
	Parent = TCS->Parent;
	Level = TCS->Level;
	Tree = TCS->Tree;
	if(Level < Tree->Specs.TexDepth)
		Flags |= QT_FLAG_SUBDIV;
	// set bounding box x and z coordinates
	if(TCS->Parent)
	{
		switch(TCS->Flags & QT_CHILD_MASK)
		{
		case QT_CHILD_NW:
			BBox.Data[BB_LEFT] = Parent->BBox.Data[BB_LEFT];
			BBox.Data[BB_RIGHT] = 0.5f*(Parent->BBox.Data[BB_LEFT] + Parent->BBox.Data[BB_RIGHT]);
			BBox.Data[BB_NEAR] = 0.5f*(Parent->BBox.Data[BB_NEAR] + Parent->BBox.Data[BB_FAR]);
			BBox.Data[BB_FAR] = Parent->BBox.Data[BB_FAR];
			break;
		case QT_CHILD_NE:
			BBox.Data[BB_LEFT] = 0.5f*(Parent->BBox.Data[BB_LEFT] + Parent->BBox.Data[BB_RIGHT]);
			BBox.Data[BB_RIGHT] = Parent->BBox.Data[BB_RIGHT];
			BBox.Data[BB_NEAR] = 0.5f*(Parent->BBox.Data[BB_NEAR] + Parent->BBox.Data[BB_FAR]);
			BBox.Data[BB_FAR] = Parent->BBox.Data[BB_FAR];
			break;
		case QT_CHILD_SW:
			BBox.Data[BB_LEFT] = Parent->BBox.Data[BB_LEFT];
			BBox.Data[BB_RIGHT] = 0.5f*(Parent->BBox.Data[BB_LEFT] + Parent->BBox.Data[BB_RIGHT]);
			BBox.Data[BB_NEAR] = Parent->BBox.Data[BB_NEAR];
			BBox.Data[BB_FAR] = 0.5f*(Parent->BBox.Data[BB_NEAR] + Parent->BBox.Data[BB_FAR]);
			break;
		case QT_CHILD_SE:
			BBox.Data[BB_LEFT] = 0.5f*(Parent->BBox.Data[BB_LEFT] + Parent->BBox.Data[BB_RIGHT]);
			BBox.Data[BB_RIGHT] = Parent->BBox.Data[BB_RIGHT];
			BBox.Data[BB_NEAR] = Parent->BBox.Data[BB_NEAR];
			BBox.Data[BB_FAR] = 0.5f*(Parent->BBox.Data[BB_NEAR] + Parent->BBox.Data[BB_FAR]);
			break;
		}
	}
	else
	{
		// root node
		BBox.Data[BB_LEFT] = -0.5f * TCS->Tree->Specs.Size;
		BBox.Data[BB_RIGHT] = 0.5f * TCS->Tree->Specs.Size;
		BBox.Data[BB_NEAR] = -0.5f * TCS->Tree->Specs.Size;
		BBox.Data[BB_FAR] = 0.5f * TCS->Tree->Specs.Size;
	}
	// set bounding box y coordinates
	BBox.Data[BB_BOTTOM] = 0; // these dummy values are replaced later
	BBox.Data[BB_TOP] = 0;
	// outer bounding box for node visibility calculation
	VBox.Data[BB_LEFT] = BBox.Data[BB_LEFT] - 8.0f;
	VBox.Data[BB_RIGHT] = BBox.Data[BB_RIGHT] + 8.0f;
	VBox.Data[BB_NEAR] = BBox.Data[BB_NEAR] - 8.0f;
	VBox.Data[BB_FAR] = BBox.Data[BB_FAR] + 8.0f;
	VBox.Data[BB_BOTTOM] = BBox.Data[BB_BOTTOM] - 8.0f;
	VBox.Data[BB_TOP] = BBox.Data[BB_TOP] + 8.0f;
	// set center
	Center = Vector(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT]),
	                0.5f * (BBox.Data[BB_BOTTOM] + BBox.Data[BB_TOP]),
					0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	// create children
	if(Level >= TCS->Tree->Specs.Depth)
	{
		Child[QT_CHILD_NW] = NULL;
		Child[QT_CHILD_NE] = NULL;
		Child[QT_CHILD_SW] = NULL;
		Child[QT_CHILD_SE] = NULL;
	}
	else
	{
		// new level
		TCS->Level++;
		TCS->Parent = this;
		// create new TCSs
		TreeCreateStruct NewTCS[4];
		// copy
		NewTCS[0] = *TCS;
		NewTCS[1] = *TCS;
		NewTCS[2] = *TCS;
		NewTCS[3] = *TCS;
		// modify for subnodes
		NewTCS[0].Flags = QT_CHILD_NW;
		NewTCS[1].Flags = QT_CHILD_NE;
		NewTCS[2].Flags = QT_CHILD_SW;
		NewTCS[3].Flags = QT_CHILD_SE;
		// create subnodes
		if(TCS->Level < TCS->Tree->Specs.TexDepth)
		{
			Child[QT_CHILD_NW] = new QTNode(&NewTCS[0]);
			Child[QT_CHILD_NE] = new QTNode(&NewTCS[1]);
			Child[QT_CHILD_SW] = new QTNode(&NewTCS[2]);
			Child[QT_CHILD_SE] = new QTNode(&NewTCS[3]);
		}
		else if(TCS->Level == TCS->Tree->Specs.TexDepth)
		{
			Child[QT_CHILD_NW] = new QTNodeTerraTex(&NewTCS[0]);
			Child[QT_CHILD_NE] = new QTNodeTerraTex(&NewTCS[1]);
			Child[QT_CHILD_SW] = new QTNodeTerraTex(&NewTCS[2]);
			Child[QT_CHILD_SE] = new QTNodeTerraTex(&NewTCS[3]);
		}
		else if(TCS->Level > TCS->Tree->Specs.TexDepth)
		{
			Child[QT_CHILD_NW] = new QTNodeTerra(&NewTCS[0]);
			Child[QT_CHILD_NE] = new QTNodeTerra(&NewTCS[1]);
			Child[QT_CHILD_SW] = new QTNodeTerra(&NewTCS[2]);
			Child[QT_CHILD_SE] = new QTNodeTerra(&NewTCS[3]);
		}
	}
}

void QTNode::ClearSubdivisions(void)
{
	if(Flags & QT_FLAG_SUBDIV)
	{
		Flags &= ~QT_FLAG_SUBDIV;
		for(int i = 0; i < 4; i++)
			if(Child[i])
				Child[i]->ClearSubdivisions();
	}
}

void QTNode::Visibility(D3DStruct *DS)
{
	if(Level < Tree->Specs.TexDepth)
	{
		CalculateScreenBBox(DS);
		Flags |= QT_FLAG_SUBDIV;
	}
	for(int i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->Visibility(DS);
}

void QTNode::BalanceSubTrees(D3DStruct *DS)
{
	for(int i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->BalanceSubTrees(DS);
}

void QTNode::CreateScene(D3DStruct *DS)
{
	for(int i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->CreateScene(DS);
}

void QTNode::CreateTerrain(void)
{
	using namespace IMATH;
	int i;
	// neighbors
	// north
	Vector2DXZ Pos = Vector2DXZ(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT])
	                          , 1.5f * BBox.Data[BB_FAR] - 0.5f * BBox.Data[BB_NEAR]);
	Neighbors[QT_NEIGHBOR_N] = Tree->GetNode(&Pos, Level);
	// south
	Pos = Vector2DXZ(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT])
	               , 1.5f * BBox.Data[BB_NEAR] - 0.5f * BBox.Data[BB_FAR]);
	Neighbors[QT_NEIGHBOR_S] = Tree->GetNode(&Pos, Level);
	// west
	Pos = Vector2DXZ(1.5f * BBox.Data[BB_LEFT] - 0.5f * BBox.Data[BB_RIGHT]
	               , 0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	Neighbors[QT_NEIGHBOR_W] = Tree->GetNode(&Pos, Level);
	// east
	Pos = Vector2DXZ(1.5f * BBox.Data[BB_RIGHT] - 0.5f * BBox.Data[BB_LEFT]
	               , 0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	Neighbors[QT_NEIGHBOR_E] = Tree->GetNode(&Pos, Level);
	// children
	for(i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->CreateTerrain();
	// BBox, VBox
	if(Level < Tree->Specs.Depth)
	{
		BBox.Data[BB_BOTTOM] = Child[0]->BBox.Data[BB_BOTTOM];
		BBox.Data[BB_TOP] = Child[0]->BBox.Data[BB_TOP];
		for(i = 0; i < 4; i++)
		{
			if(Child[i]->BBox.Data[BB_BOTTOM] < BBox.Data[BB_BOTTOM])
				BBox.Data[BB_BOTTOM] = Child[i]->BBox.Data[BB_BOTTOM];
			if(Child[i]->BBox.Data[BB_TOP] > BBox.Data[BB_TOP])
				BBox.Data[BB_TOP] = Child[i]->BBox.Data[BB_TOP];
		}
		VBox.Data[BB_LEFT] = BBox.Data[BB_LEFT] - 8.0f;
		VBox.Data[BB_RIGHT] = BBox.Data[BB_RIGHT] + 8.0f;
		VBox.Data[BB_NEAR] = BBox.Data[BB_NEAR] - 8.0f;
		VBox.Data[BB_FAR] = BBox.Data[BB_FAR] + 8.0f;
		VBox.Data[BB_BOTTOM] = BBox.Data[BB_BOTTOM] - 8.0f;
		VBox.Data[BB_TOP] = BBox.Data[BB_TOP] + 8.0f;
		// set center
		Center = Vector(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT]),
						0.5f * (BBox.Data[BB_BOTTOM] + BBox.Data[BB_TOP]),
						0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	}
}

QTNode *QTNode::GetNode(const IMATH::Vector2DXZ *Pos, int Level)
{
	if(BBox.IsIn(Pos))
	{
		if(Level == QTNode::Level)
			return this;
		else
			for(int i = 0; i < 4; i++)
				if(Child[i]->BBox.IsIn(Pos))
					return Child[i]->GetNode(Pos, Level);
	}
	return NULL;
}

void QTNode::AddManagedObject(ManagedObject *Obj)
{
	if(!Obj)
		throw("QTNode::AddManagedObject added NULL Object");
	for(int i = 0; i < 4; i++)
	{
		IMATH::Vector2DXZ Pos = IMATH::Vector2DXZ(Obj->GetPos().x, Obj->GetPos().z);
		if(Child[i]->BBox.IsIn(&Pos))
			Child[i]->AddManagedObject(Obj);
	}
}

void QTNode::CreateTextures(D3DStruct *DS)
{
	for(int i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->CreateTextures(DS);
}

void QTNode::ReconstructMipMaps(D3DStruct *DS)
{
	for(int i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->ReconstructMipMaps(DS);
}

void QTNode::RenderWater(D3DStruct *DS)
{
	for(int i = 0; i < 4; i++)
		if(Child[i])
			Child[i]->RenderWater(DS);
}

void QTNode::Init3D(D3DStruct *DS, bool recurse)
{
	if(recurse)
		for(int i = 0; i < 4; i++)
			if(Child[i])
				Child[i]->Init3D(DS, true);
}

int QTNode::DeleteManagedObject(ManagedObject *Obj)
{
	int NumDeleted = 0;
	for(int i = 0; i < 4; i++)
	{
		IMATH::Vector2DXZ Pos = IMATH::Vector2DXZ(Obj->GetPos().x, Obj->GetPos().z);
		//if(Child[i]->BBox.IsIn(&Pos))
			NumDeleted += Child[i]->DeleteManagedObject(Obj);
	}
	return NumDeleted;
}

void QTNode::CalculateScreenBBox(D3DStruct *DS)
{
	// outline vectors of the bounding box
	Vector Vec(BBox.Data[BB_LEFT], BBox.Data[BB_BOTTOM], BBox.Data[BB_NEAR]);
	Vec *= DS->ProjViewMtx;
	ScreenBBox.Data[BB_LEFT] = Vec.x;
	ScreenBBox.Data[BB_RIGHT] = Vec.x;
	ScreenBBox.Data[BB_BOTTOM] = Vec.y;
	ScreenBBox.Data[BB_TOP] = Vec.y;
	ScreenBBox.Data[BB_NEAR] = Vec.z;
	ScreenBBox.Data[BB_FAR] = Vec.z;
	for(int i = BB_LEFT; i <= BB_RIGHT; ++i)
		for(int j = BB_BOTTOM; j <= BB_TOP; ++j)
			for(int k = BB_NEAR; k <= BB_FAR; ++k)
			{
				if(i == BB_LEFT && j == BB_BOTTOM && k == BB_NEAR)
					continue;
				Vector Vec(BBox.Data[i], BBox.Data[j], BBox.Data[k]);
				Vec *= DS->ProjViewMtx;
				if(Vec.z < 0 || Vec.z > 1)
				{
					Vec.x = -Vec.x;
					Vec.y = -Vec.y;
				}
				if(Vec.x < ScreenBBox.Data[BB_LEFT])
					ScreenBBox.Data[BB_LEFT] = Vec.x;
				if(Vec.x > ScreenBBox.Data[BB_RIGHT])
					ScreenBBox.Data[BB_RIGHT] = Vec.x;
				if(Vec.y < ScreenBBox.Data[BB_BOTTOM])
					ScreenBBox.Data[BB_BOTTOM] = Vec.y;
				if(Vec.y > ScreenBBox.Data[BB_TOP])
					ScreenBBox.Data[BB_TOP] = Vec.y;
				if(Vec.z < ScreenBBox.Data[BB_NEAR])
					ScreenBBox.Data[BB_NEAR] = Vec.z;
				if(Vec.z > ScreenBBox.Data[BB_FAR])
					ScreenBBox.Data[BB_FAR] = Vec.z;
			}

	/*Vector Vec[8];
	Vec[0] = Vector(BBox.Data[BB_LEFT ], BBox.Data[BB_BOTTOM], BBox.Data[BB_NEAR]);
	Vec[1] = Vector(BBox.Data[BB_RIGHT], BBox.Data[BB_BOTTOM], BBox.Data[BB_NEAR]);
	Vec[2] = Vector(BBox.Data[BB_LEFT ], BBox.Data[BB_TOP   ], BBox.Data[BB_NEAR]);
	Vec[3] = Vector(BBox.Data[BB_RIGHT], BBox.Data[BB_TOP   ], BBox.Data[BB_NEAR]);
	Vec[4] = Vector(BBox.Data[BB_LEFT ], BBox.Data[BB_BOTTOM], BBox.Data[BB_FAR ]);
	Vec[5] = Vector(BBox.Data[BB_RIGHT], BBox.Data[BB_BOTTOM], BBox.Data[BB_FAR ]);
	Vec[6] = Vector(BBox.Data[BB_LEFT ], BBox.Data[BB_TOP   ], BBox.Data[BB_FAR ]);
	Vec[7] = Vector(BBox.Data[BB_RIGHT], BBox.Data[BB_TOP   ], BBox.Data[BB_FAR ]);
	Vector Low, High;
	for(int i = 0; i < 8; i++)
	{
		//Vec[i] = DS->ViewMtx * Vec[i];
		Vec[i] = DS->ProjViewMtx * Vec[i];
		if(Vec[i].z < 0 || Vec[i].z > 1)
		{
			//Vec[i].z = 1.0f;
			Vec[i].x = -Vec[i].x;
			Vec[i].y = -Vec[i].y;
		}
		if(i == 0)
		{
			// initialize
			Low = Vec[i];//Vector(Vec[i].x, Vec[i].y, Vec[i].z);
			High = Vec[i];//Vector(Vec[i].x, Vec[i].y, Vec[i].z);
		}
		else
		{
			if(Vec[i].x < Low.x) Low.x = Vec[i].x;
			else if(Vec[i].x > High.x) High.x = Vec[i].x;
			if(Vec[i].y < Low.y) Low.y = Vec[i].y;
			else if(Vec[i].y > High.y) High.y = Vec[i].y;
			if(Vec[i].z < Low.z) Low.z = Vec[i].z;
			else if(Vec[i].z > High.z) High.z = Vec[i].z;
		}
	}
	ScreenBBox = BoundingBox(Low.x, High.x, Low.y, High.y, Low.z, High.z);*/
}

bool QTNode::IsInScreenBBox(const IMATH::Vector2D &Pos)
{
	if(!(Flags & QT_FLAG_SUBDIV))
		CalculateScreenBBox(Tree->Specs.DS);
	return ScreenBBox.IsIn(&Pos) || (ScreenBBox.Data[BB_NEAR] < 0 && ScreenBBox.Data[BB_FAR] > 0);
}

IMATH::Vector2D QTNode::ToScreenCoords(const IMATH::Vector &V)
{
	IMATH::Vector Screen3D = V;
	Screen3D *= Tree->Specs.DS->ProjViewMtx;
	if(Screen3D.z < 0 || Screen3D.z > 1)
	{
		Screen3D.x = -Screen3D.x;
		Screen3D.y = -Screen3D.y;
	}
	IMATH::Vector2D Screen(Screen3D.x, Screen3D.y);
	return Screen;
}

IMATH::Vector2D QTNode::GetScreenCenter()
{
	IMATH::Vector ScreenCenter3D = Center;
	ScreenCenter3D *= Tree->Specs.DS->ProjViewMtx;
	if(ScreenCenter3D.z < 0 || ScreenCenter3D.z > 1)
	{
		ScreenCenter3D.x = -ScreenCenter3D.x;
		ScreenCenter3D.y = -ScreenCenter3D.y;
	}
	ScreenCenter = IMATH::Vector2D(ScreenCenter3D.x, ScreenCenter3D.y);
	return ScreenCenter;
}

IMATH::Vector QTNode::GetBBoxCenter()
{
	return IMATH::Vector((BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT])/2,
		(BBox.Data[BB_TOP] + BBox.Data[BB_BOTTOM])/2,
		(BBox.Data[BB_NEAR] + BBox.Data[BB_FAR])/2);
}

IMATH::Vector2D QTNode::GetScreenBBoxCenter()
{
	if(!(Flags & QT_FLAG_SUBDIV))
		CalculateScreenBBox(Tree->Specs.DS);
	return IMATH::Vector2D((ScreenBBox.Data[BB_LEFT] + ScreenBBox.Data[BB_RIGHT])/2,
		(ScreenBBox.Data[BB_TOP] + ScreenBBox.Data[BB_BOTTOM])/2);
}

BoundingBox QTNode::GetScreenBBox()
{
	if(!(Flags & QT_FLAG_SUBDIV))
		CalculateScreenBBox(Tree->Specs.DS);
	return ScreenBBox;
}

QTNode *QTNode::GetNode(const IMATH::Vector2D *Pos)
{
	if(!Child[0])
		return this;
	else
	{
		QTNode *NearestChildNearest = 0;
		float ChildDistSq = 1E8;
		for(int i = 0; i < 4; i++)
		{
			QTNode *CNearest = Child[i];
			QTNode *DeepNearest;
			float DistSq = 1E8;
			BoundingBox BB = CNearest->GetScreenBBox();
			/*IMATH::Vector2D NW = CNearest->GetScreenCorner(QT_INDEX_NW);
			IMATH::Vector2D NE = CNearest->GetScreenCorner(QT_INDEX_NE);
			IMATH::Vector2D SW = CNearest->GetScreenCorner(QT_INDEX_SW);
			IMATH::Vector2D SE = CNearest->GetScreenCorner(QT_INDEX_SE);
			//IMATH::Vector2D C = DeepNearest->GetScreenCenter();
			float Diam1 = (NW - SE).LengthSq();
			float Diam2 = (NE - SW).LengthSq();
			float DiamSq = max(Diam1, Diam2) / 2;*/
			//DiamSq = DiamSq * DiamSq;
			/*if((NW - *Pos).LengthSq() < DiamSq
				|| (NE - *Pos).LengthSq() < DiamSq
				|| (SW - *Pos).LengthSq() < DiamSq
				|| (SE - *Pos).LengthSq() < DiamSq)*/
			if(Pos->x > BB.Data[BB_LEFT] && Pos->x < BB.Data[BB_RIGHT]
				&& Pos->y > BB.Data[BB_BOTTOM] && Pos->y < BB.Data[BB_TOP])
			{
				DeepNearest = Child[i]->GetNode(Pos);
				DistSq = (DeepNearest->GetScreenCenter() - *Pos).LengthSq();
			}
			else
				continue;
			if(DistSq < ChildDistSq)
			{
				ChildDistSq = DistSq;
				NearestChildNearest = DeepNearest;
			}
		}
		return NearestChildNearest ? NearestChildNearest : this;
	}
}

XZPos QTNode::GetXZPos(const IMATH::Vector2D &Pos)
{
	/*float Dist = 1000000000.0f	
	float NearestDist = (GetScreenCenter() - Pos).Length();
	Node *NearestNode = this;*/
/*	XZPos ChRet[4];
	bool FoundChild = false, HasChild = false;
	int HighestLevel = -2, Index = -2;
	float z = 1000000000.0f, CenterDist = 1000000000.0f;
	// find child
	for(int i = 0; i < 4; i++)
		if(Child[i])
		{
			HasChild = true;
			if(Child[i]->IsInScreenBBox(Pos))
			{
				ChRet[i] = Child[i]->GetXZPos(Pos);
				if(ChRet[i].Level > HighestLevel)
				{
					if(Child[i]->ScreenBBox.Data[BB_NEAR] < z)
					{
						//if((Pos - Child[i]->ScreenBBox.GetCenterXY()).Length() < CenterDist)
						{
							HighestLevel = ChRet[i].Level;
							Index = i;
							CenterDist = (Pos - Child[i]->ScreenBBox.GetCenterXY()).Length();
							z = Child[i]->ScreenBBox.Data[BB_NEAR];
						}
					}
				}
				FoundChild = true;
			}
		}
	if(FoundChild)
	{
		// return child result
		return ChRet[Index];
	}*/
	/*else if(HasChild)
	{
		// no child with matching screenbbox found, but has children
		// find nearest child
		int Index = -1;
		float MinLength = -1;
		for(int i = 0; i < 4; i++)
			if(Child[i])
				if(Index == -1 || (Pos - Child[i]->GetScreenBBoxCenter()).Length() < MinLength)
				{
					MinLength = (Pos - Child[i]->GetScreenBBoxCenter()).Length();
					Index = i;
				}
		// return result of nearest child
		return Child[Index]->GetXZPos(Pos);
	}*/
	{
		//std::ofstream log("affine.txt", std::ios::app);
		// node has no children
		// get neighbor positions
		ISL::DynamicArray<NodeInfo> Nodes;
		//Nodes.Add(NodeInfo(GetScreenBBoxCenter(), GetBBoxCenter()));
		int i;
		/*for(i = 0; i < 4; i++) // BEWARE: only 3 neighbors
			if(Nodes.Count() < 3 && Neighbors[i])
				Nodes.Add(NodeInfo(Neighbors[i]->GetScreenCenter(), Neighbors[i]->Center));
		if(Nodes.Count() < 3)
		{
			Nodes.Add(NodeInfo(GetScreenCenter(), Center));
			if(Nodes.Count() < 3)
				for(int i = 0; i < 4; i++)
					if(Child[i] && Nodes.Count() < 3)
						Nodes.Add(NodeInfo(Child[i]->GetScreenCenter(), Child[i]->Center));
		}*/
		Nodes.Add(NodeInfo(GetScreenCenter(), Center));
		Nodes.Add(NodeInfo(ToScreenCoords(Center + IMATH::Vector(10, 0, 0)), Center + IMATH::Vector(10, 0, 0)));
		Nodes.Add(NodeInfo(ToScreenCoords(Center + IMATH::Vector(0, 0, 10)), Center + IMATH::Vector(0, 0, 10)));
		/*Nodes.Add(NodeInfo(ToScreenCoords(IMATH::Vector(0, 0, 0)), IMATH::Vector(0, 0, 0)));
		Nodes.Add(NodeInfo(ToScreenCoords(IMATH::Vector(1000, 0, 0)), IMATH::Vector(1000, 0, 0)));
		Nodes.Add(NodeInfo(ToScreenCoords(IMATH::Vector(0, 0, 1000)), IMATH::Vector(0, 0, 1000)));*/
		// calculate coefficients
		/*for(i = 0; i < Nodes.Count(); i++)
			Nodes[i].Coeff = 1.0f / Nodes.Count();*/
		Nodes[0].Coeff = 1;
		Nodes[1].Coeff = 0;
		Nodes[2].Coeff = 0;
		if(Nodes.Count() == 3)
		{
			// Gaussian Algorithm A*l = s
			// last line l_1 + l_2 + l_3 = 1 for affine combination
			float A[3][3], s[3];
			A[0][0] = Nodes[0].ScreenCenter.x;
			A[1][0] = Nodes[0].ScreenCenter.y;
			A[2][0] = 1;
			A[0][1] = Nodes[1].ScreenCenter.x;
			A[1][1] = Nodes[1].ScreenCenter.y;
			A[2][1] = 1;
			A[0][2] = Nodes[2].ScreenCenter.x;
			A[1][2] = Nodes[2].ScreenCenter.y;
			A[2][2] = 1;
			s[0] = Pos.x;
			s[1] = Pos.y;
			s[2] = 1;
			/*log << "pos: " << Pos.x << ", " << Pos.y << std::endl;
			for(int j = 0; j < 3; ++j)
			{
				for(int l = 0; l < 3; ++l)
					log << A[j][l] << " ";
				log << " " << s[j] << std::endl;
			}*/
			
			bool found_nonzero = true;
			for(int j = 0; j < 3; ++j)
			{
				if(A[j][j] == 0)
				{
					// find row to swap
					found_nonzero = false;
					for(int k = j + 1; k < 3; ++k)
						if(A[k][j] != 0)
						{
							// swap values of A[j][.] and A[k][.]
							for(int l = j; l < 3; ++l)
							{
								float swap = A[j][l];
								A[j][l] = A[k][l];
								A[k][l] = swap;
							}
							// swap values of s[j] and s[k]
							float swap = s[j];
							s[j] = s[k];
							s[k] = swap;
							found_nonzero = true;
							break;
						}
					if(!found_nonzero)
						break;
				}
				// transform pivot element to 1
				for(int l = j + 1; l < 3; ++l)
					A[j][l] /= A[j][j];
				s[j] /= A[j][j];
				A[j][j] = 1;
				// eliminate column under pivot element
				for(int k = j + 1; k < 3; ++k)
				{
					for(int l = j + 1; l < 3; ++l)
						A[k][l] -= A[k][j] * A[j][l];
					s[k] -= A[k][j] * s[j];
					A[k][j] = 0;
				}
				//
				/*log << std::endl;
				for(int j = 0; j < 3; ++j)
				{
					for(int l = 0; l < 3; ++l)
						log << A[j][l] << " ";
					log << " " << s[j] << std::endl;
				}*/
			}
			// result
			if(found_nonzero)
			{
				// construct identity matrix
				s[1] -= A[1][2] * s[2];
				A[1][2] = 0;
				s[0] -= A[0][2] * s[2];
				A[0][2] = 0;
				s[0] -= A[0][1] * s[1];
				A[0][1] = 0;
				for(int j = 0; j < 3; ++j)
					Nodes[j].Coeff = s[j];
			}
			/*log << std::endl;
			for(int j = 0; j < 3; ++j)
			{
				for(int l = 0; l < 3; ++l)
					log << A[j][l] << " ";
				log << " " << s[j] << std::endl;
			}*/
		}
		// debug
		//std::ofstream deb;
		//if(!IsInScreenBBox(Pos))
		//{
		//	deb.open("debug.txt");
		//	deb << "Level: " << Level << std::endl;
		//}
		// get affine combination by adaptive iteration
/*		IMATH::Vector2D AverageScreenPos;
		for(int k = 0; k < 20 && (Pos-AverageScreenPos).Length() > 0.001f; ++k)
		{
			AverageScreenPos = IMATH::Vector2D(0, 0);
			for(i = 0; i < Nodes.Count(); i++)
				AverageScreenPos += Nodes[i].Coeff * Nodes[i].ScreenCenter;
			//AverageScreenPos = (1.0f / Nodes.Count()) * AverageScreenPos;
			//float AvError = 0;
			for(i = 0; i < Nodes.Count(); i++)
			{
				// Distance
				//float Factor = (Nodes[i].ScreenCenter - AverageScreenPos).Length();
				//float Factor = (Nodes[i].ScreenCenter - Pos).Length();
				float Factor = 0;
				for(int j = 0; j < Nodes.Count(); j++)
					Factor += (Nodes[j].ScreenCenter - Pos).Length();
				Factor /= Nodes.Count();
				if(Factor)
				{
					// total error per distance
					Nodes[i].Error = (AverageScreenPos - Pos).Length();// / Factor;
					// change of negative (error per change of node coefficient) per distance
					Nodes[i].Error *= (100 * ((Pos - AverageScreenPos).Length()
						- (Pos - (AverageScreenPos + 0.01f * (Nodes[i].ScreenCenter - AverageScreenPos))).Length()));// / Factor;
					Nodes[i].Error /= Factor;
					// distribute change to nodes
					Nodes[i].Error /= Nodes.Count();
				}
				else
					Nodes[i].Error = 0;
			}
			//AvError /= Nodes.Count();
			// apply and correct affinity
			//float Eucl = 0;
			//for(i = 0; i < Nodes.Count(); i++)
			//{
			//	Nodes[i].Coeff += Nodes[i].Error;
			//	Eucl += Nodes[i].Coeff * Nodes[i].Coeff;
			//}
			//Eucl = sqrt(Eucl);
			//if(Eucl)
			//	for(i = 0; i < Nodes.Count(); i++)
			//		Nodes[i].Coeff /= Eucl;
			for(i = 0; i < Nodes.Count(); i++)
			{
				Nodes[i].Coeff += Nodes[i].Error;
				// maintain affinity
				for(int j = 0; j < Nodes.Count(); j++)
					if(j != i)
						Nodes[i].Coeff -= Nodes[j].Error / (Nodes.Count() - 1);
			}*/
			// debug
			//if(!IsInScreenBBox(Pos))
			/*{
				deb << "coeff: ";
				for(i = 0; i < Nodes.Count(); i++)
				{
					deb << Nodes[i].Coeff << ' ';
				}
				deb << std::endl << "error: ";
				for(i = 0; i < Nodes.Count(); i++)
					deb << Nodes[i].Error << ' ';
				deb << std::endl;
				deb << "pos: " << Pos.x << ", " << Pos.y << std::endl;
				deb << "scr: " << AverageScreenPos.x << ", " << AverageScreenPos.y << std::endl;
				deb << "total error: " << (AverageScreenPos - Pos).Length() << std::endl;
				deb << std::endl;
			}*/
/*		}*/
		// prepare return value
		XZPos Ret;
		Ret.Level = (IsInScreenBBox(Pos) ? Level : -1);
		// calculate affine combination
		Ret.Pos = IMATH::Vector2DXZ(0, 0);
		for(i = 0; i < Nodes.Count(); i++)
			Ret.Pos += Nodes[i].Coeff * IMATH::Vector2DXZ(Nodes[i].Center.x, Nodes[i].Center.z);
		if(Ret.Pos.x < -Tree->Specs.Size)
			Ret.Pos.x = -Tree->Specs.Size;
		if(Ret.Pos.x > Tree->Specs.Size)
			Ret.Pos.x = Tree->Specs.Size;
		if(Ret.Pos.z < -Tree->Specs.Size)
			Ret.Pos.z = -Tree->Specs.Size;
		if(Ret.Pos.z > Tree->Specs.Size)
			Ret.Pos.z = Tree->Specs.Size;
		// return interpolated position
		return Ret;
	}

/*	XZPos Ret;
	if(!(Flags & QT_FLAG_SUBDIV))
		CalculateScreenBBox(Tree->Specs.DS);
	if(ScreenBBox.IsIn(&Pos) || (ScreenBBox.Data[BB_NEAR] < 0 && ScreenBBox.Data[BB_FAR] > 0))
	{
		// pos is in bounding box
		// try to get XZPos from clients
		XZPos ChRet[4];
		int HighestLevel = 0, HighestNum = -1;
		for(int i = 0; i < 4; i++)
			if(Child[i])
			{
				ChRet[i] = Child[i]->GetXZPos(Pos);
				if(ChRet[i].Level > HighestLevel)
				{
					HighestLevel = ChRet[i].Level;
					HighestNum = i;
				}
			}
		if(HighestNum != -1 && HighestLevel > 0)
		{
			// found position
			return ChRet[HighestNum];
		}
		else if(HighestNum != -1)
		{
			// pos is in no no childs bbox, find nearest child
		}
		else
		{
			// return position
			Ret.Level = Level;
			Ret.Pos = IMATH::Vector2DXZ(Center.x, Center.z);
			return Ret;
		}
	}
	else
	{
		// pos is outside of this bounding box
		Ret.Pos = IMATH::Vector2DXZ(0, 0);
		Ret.Level = 0;
		return Ret;
	}*/
}

QTNode::~QTNode(void)
{
	for(int i = 0; i < 4; i++)
		if(Child[i])
			delete Child[i];
}

// ****** QTNodeRoot ******

QTNodeRoot::QTNodeRoot(TreeCreateStruct *TCS)
:QTNode(TCS)
{
	// we are still in the constructor, so Tree->Root has not been set yet
	Tree->Root = this;
	QTNode::CreateTerrain(); // can't call virtual functions of this object in constructor!
}

void QTNodeRoot::Visibility(D3DStruct *DS)
{
	Tree->Prf.Start(0);
	ClearSubdivisions();
	Tree->Prf.Stop(0);
	Tree->Prf.Start(1);
	QTNode::Visibility(DS);
	Tree->Prf.Stop(1);
	Tree->Prf.Start(2);
	BalanceSubTrees(DS);
	Tree->Prf.Stop(2);
	Tree->Prf.Start(3);
	CreateScene(DS);
	Tree->Prf.Stop(3);
}

// ****** QTNodeTerra ******

QTNodeTerra::QTNodeTerra(QTNode::TreeCreateStruct *TCS)
:QTNode(TCS)
{
}

// factors
#define QT_STD_COMPLEXITY 1.4f
#define QT_WATER_COMPLEXITY 1.9f
// calculate terrain complexity for detail-based LODing
void QTNodeTerra::CalculateComplexity(void)
{
	// calculate node complexity for detail-based LODing
	Vector NS = 0.5f * (TexNode->Vertices[Indices[QT_INDEX_N]].v + TexNode->Vertices[Indices[QT_INDEX_S]].v);
	Vector WE = 0.5f * (TexNode->Vertices[Indices[QT_INDEX_W]].v + TexNode->Vertices[Indices[QT_INDEX_E]].v);
	Vector NWSE = 0.5f * (TexNode->Vertices[Indices[QT_INDEX_NW]].v + TexNode->Vertices[Indices[QT_INDEX_SE]].v);
	Vector NESW = 0.5f * (TexNode->Vertices[Indices[QT_INDEX_NE]].v + TexNode->Vertices[Indices[QT_INDEX_SW]].v);
	float Diff1 = (NS - TexNode->Vertices[Indices[QT_INDEX_C]].v).Length();
	float Diff2 = (WE - TexNode->Vertices[Indices[QT_INDEX_C]].v).Length();
	float Diff3 = (NWSE - TexNode->Vertices[Indices[QT_INDEX_C]].v).Length();
	float Diff4 = (NESW - TexNode->Vertices[Indices[QT_INDEX_C]].v).Length();
	Complexity = QT_STD_COMPLEXITY * max(max(Diff1, Diff2), max(Diff3, Diff4));
	// water
	if(absf(TexNode->Vertices[Indices[QT_INDEX_C]].v.y - WATERLEVEL) < 5.0f)
		Complexity *= QT_WATER_COMPLEXITY / QT_STD_COMPLEXITY;
	Complexity /= (TexNode->Vertices[Indices[QT_INDEX_N]].v - TexNode->Vertices[Indices[QT_INDEX_S]].v).Length();
	Complexity = Complexity * Complexity;
	// subnode complexity
	if(Level != Tree->Specs.Depth)
		Complexity = (float)sqrt(2.0) * max(Complexity, max(((QTNodeTerra *)Child[QT_CHILD_SW])->Complexity,
		                             max(((QTNodeTerra *)Child[QT_CHILD_SE])->Complexity,
									 max(((QTNodeTerra *)Child[QT_CHILD_NW])->Complexity,
									     ((QTNodeTerra *)Child[QT_CHILD_NE])->Complexity))));
}

void QTNodeTerra::CalculateVisibility(D3DStruct *DS)
{
	Flags &= ~QT_FLAG_VISIBLE_MASK;
	// parent is completely visible -> this node is completely visible
	if(Level != Tree->Specs.TexDepth && (Parent->Flags & QT_FLAG_VISIBLE_COMPLETELY))
	{
		Flags |= (QT_FLAG_VISIBLE_COMPLETELY | QT_FLAG_VISIBLE_PARTLY);
		return;
	}
	// parent not visible -> this node is not visible
	if(Level != Tree->Specs.TexDepth && !(Parent->Flags & QT_FLAG_VISIBLE_PARTLY))
		return;
	// outline vectors of the visibility box
	Vector Vec[8];
	Vec[0] = Vector(VBox.Data[BB_LEFT ], VBox.Data[BB_BOTTOM], VBox.Data[BB_NEAR]);
	Vec[1] = Vector(VBox.Data[BB_RIGHT], VBox.Data[BB_BOTTOM], VBox.Data[BB_NEAR]);
	Vec[2] = Vector(VBox.Data[BB_LEFT ], VBox.Data[BB_TOP   ], VBox.Data[BB_NEAR]);
	Vec[3] = Vector(VBox.Data[BB_RIGHT], VBox.Data[BB_TOP   ], VBox.Data[BB_NEAR]);
	Vec[4] = Vector(VBox.Data[BB_LEFT ], VBox.Data[BB_BOTTOM], VBox.Data[BB_FAR ]);
	Vec[5] = Vector(VBox.Data[BB_RIGHT], VBox.Data[BB_BOTTOM], VBox.Data[BB_FAR ]);
	Vec[6] = Vector(VBox.Data[BB_LEFT ], VBox.Data[BB_TOP   ], VBox.Data[BB_FAR ]);
	Vec[7] = Vector(VBox.Data[BB_RIGHT], VBox.Data[BB_TOP   ], VBox.Data[BB_FAR ]);
	// all edges of the visibility box
	const int Edges[12][2] = {	{0, 1}, {0, 2}, {0, 4},
								{5, 1}, {5, 4}, {5, 7},
								{3, 1}, {3, 2}, {3, 7},
								{6, 2}, {6, 7}, {6, 4}};
	//
	// Code[8]:
	//
	// code representing the position of the vertex on the screen:
	//
	//           1001 |  1000  | 1010
	//          ------+--------+------
	//           0001 |  0000  | 0010
	//                | screen |
	//          ------+--------+------
	//           0101 |  0100  | 0110
	//
	// vertices with the code 0000 are inside the screen clipping area [-1.0f, 1.0f]
	// additionally (because of 3D), a fifth digit 10000 is added, if Vertex.z is < -1.0f
	//
	// if (Code[x] & Code[y]) is 0, the box is potentially partially visible
	// if (Code[x] | Code[y]) is 0, the box is completely visible
	//
	int Code[8];
	int i;
	// calculate the code
	for(i = 0; i < 8; i++)
	{
		Code[i] = 0;
		// TODO: find bug with a visible line at ca. z = 75
		Vec[i] *= DS->ProjViewMtx;
		//if(Vec[i].z <= 0.90f) Code[i] |= 16;
		//if(Vec[i].z <= 0.99f) Code[i] |= 16;
		if(Vec[i].x < -1.0f) Code[i] |= 1;
		if(Vec[i].x >  1.0f) Code[i] |= 2;
		if(Vec[i].y < -1.0f) Code[i] |= 4;
		if(Vec[i].y >  1.0f) Code[i] |= 8;
		if(Vec[i].z >  1.0f)
		{
			Code[i] |= 16; // this means z < -1 before the projection matrix is applied
/*			Vec[i].x = -Vec[i].x;
			Vec[i].y = -Vec[i].y;*/
		}
	}
	Flags |= QT_FLAG_VISIBLE_COMPLETELY;
	for(i = 0; i < 12; i++)
	{
		// test if the box is partly or completely visible
		if(!(Code[Edges[i][0]] & Code[Edges[i][1]]))
			Flags |= QT_FLAG_VISIBLE_PARTLY;
		if(Code[Edges[i][0]] | Code[Edges[i][1]])
			Flags &= ~QT_FLAG_VISIBLE_COMPLETELY;
	}
}

// minimal complexity for splitting
#define QT_COMPLEXITY_MIN		0.0001f
#define QT_COMPLEXITY_MIN_SQRT	((float)sqrt(QT_COMPLEXITY_MIN))
#define QT_COMPLEXITY_DEPTH		0.0000004f //800.0f
#define QT_COMPLEXITY_DEPTH_SQ	(QT_COMPLEXITY_DEPTH * QT_COMPLEXITY_DEPTH)

void QTNodeTerra::Visibility(D3DStruct *DS)
{
	// TODO: optimize for speed
	Flags &= ~QT_FLAG_SUBDIV;

	CalculateVisibility(DS);
	CalculateScreenBBox(DS);
	if(Flags & QT_FLAG_VISIBLE_PARTLY)
	{
/*		// calculate distance from camera
		Vector Screen = Center;
		Screen = DS->ViewMtx * Screen;
		//float z = DS->ProjMtx.m[3][3];
		Screen = DS->ProjMtx * Screen;
		//DS->ProjMtx.m[3][3] = z;
		Distance = Screen.Length();*/

		Vector ScreenCenter = Center;
		float Comp = 1.0f;
		Vector ScreenDeltaX = Center + IMATH::Vector(Comp, 0, 0);
		Vector ScreenDeltaY = Center + IMATH::Vector(0, Comp, 0);
		Vector ScreenDeltaZ = Center + IMATH::Vector(0, 0, Comp);
		ScreenCenter *= DS->ProjViewMtx;
		ScreenDeltaX *= DS->ProjViewMtx;
		ScreenDeltaY *= DS->ProjViewMtx;
		ScreenDeltaZ *= DS->ProjViewMtx;
		float DistanceX = (ScreenCenter - ScreenDeltaX).LengthSq();
		float DistanceY = (ScreenCenter - ScreenDeltaY).LengthSq();
		float DistanceZ = (ScreenCenter - ScreenDeltaZ).LengthSq();
		//Distance = 20.0f * max(DistanceX, max(DistanceY, DistanceZ)) * max(DistanceX, max(DistanceY, DistanceZ));
		Distance = Complexity * std::sqrt(max(DistanceX, max(DistanceY, DistanceZ)));
		//Distance = 0.02f * sqrt(max(DistanceX, max(DistanceY, DistanceZ)));
		//Distance *= 400.0f;
		//Distance = 40.0f;
		//Complexity = 1;

		// calculate distance factor for terrain LODing
		/*if(Distance >= 0 && Distance <  100.0f) Distance =  100.0f;
		if(Distance <  0 && Distance > -100.0f) Distance = -100.0f;*/
/*		if(Distance != 0)
			Distance = QT_COMPLEXITY_DEPTH / Distance;*/
		// subdivide
		if(/*Complexity **/ Distance >= QT_COMPLEXITY_MIN && Level != Tree->Specs.Depth)
		{
			Flags |= QT_FLAG_SUBDIV;
			QTNode::Visibility(DS);
		}
	}
}

void QTNodeTerra::BalanceSubTrees(D3DStruct *DS)
{
	if(Flags & QT_FLAG_SUBDIV)
	{
		ShowNeighbors(DS);
		QTNode::BalanceSubTrees(DS);
	}
}

void QTNodeTerra::ShowNeighbors(D3DStruct *DS)
{
	if(Level > Tree->Specs.TexDepth)
		for(int i = 0; i < 4; i++)
			if(Neighbors[i])
				((QTNodeTerra *)Neighbors[i])->Show(DS);
}

void QTNodeTerra::Show(D3DStruct *DS)
{
	if(Level > Tree->Specs.TexDepth)
		if(!(Parent->Flags & QT_FLAG_SUBDIV))
			if(Parent->Flags & QT_FLAG_VISIBLE_PARTLY)
			{
				// subdivide parent
				Parent->Flags |= QT_FLAG_SUBDIV;
				// visibility
				for(int i = 0; i < 4; i++)
					((QTNodeTerra *)Parent->Child[i])->CalculateVisibility(DS);
				// show parent and parent's neighbors
				((QTNodeTerra *)Parent)->ShowNeighbors(DS);
				((QTNodeTerra *)Parent)->Show(DS);
			}
}

void QTNodeTerra::AddManagedObject(ManagedObject *Obj)
{
	Objects[Objects.Count()] = Obj;
	if((Flags & QT_FLAG_VISIBLE_PARTLY) && !(Flags & QT_FLAG_SUBDIV))
	{
		Obj->SetVisible(true);
		Tree->AddVisibleObject(Obj);
	}
	if(Level == Tree->Specs.Depth)
		Obj->Node = this;
	for(int i = 0; i < 4; i++)
	{
		IMATH::Vector2DXZ Pos = IMATH::Vector2DXZ(Obj->GetPos().x, Obj->GetPos().z);
		if(Child[i] && Child[i]->BBox.IsIn(&Pos))
			Child[i]->AddManagedObject(Obj);
	}
}

int QTNodeTerra::DeleteManagedObject(ManagedObject *Obj)
{
	int NumDeleted = 0;
	while(Objects.Delete(Obj))
	{
		NumDeleted++;
		for(int i = 0; i < 4; i++)
		{
			if(Child[i])
				NumDeleted += Child[i]->DeleteManagedObject(Obj);
		}
	}
	return NumDeleted;
}

void QTNodeTerra::CreateScene(D3DStruct *DS)
{
	// TODO: optimize for speed
	if((Flags & QT_FLAG_VISIBLE_PARTLY) && !(Flags & QT_FLAG_SUBDIV))
	{
		// objects
		for(int i = 0; i < Objects.Count(); i++)
			Tree->AddVisibleObject(Objects[i]);
		// terrain
		const bool North = Neighbors[QT_NEIGHBOR_N] && (Neighbors[QT_NEIGHBOR_N]->Flags & QT_FLAG_SUBDIV);
		const bool South = Neighbors[QT_NEIGHBOR_S] && (Neighbors[QT_NEIGHBOR_S]->Flags & QT_FLAG_SUBDIV);
		const bool West  = Neighbors[QT_NEIGHBOR_W] && (Neighbors[QT_NEIGHBOR_W]->Flags & QT_FLAG_SUBDIV);
		const bool East  = Neighbors[QT_NEIGHBOR_E] && (Neighbors[QT_NEIGHBOR_E]->Flags & QT_FLAG_SUBDIV);
		const unsigned short NW = Indices[QT_INDEX_NW];
		const unsigned short NE = Indices[QT_INDEX_NE];
		const unsigned short SW = Indices[QT_INDEX_SW];
		const unsigned short SE = Indices[QT_INDEX_SE];
		//if(!(North || South || East || West) && /*Complexity **/ Distance < QT_COMPLEXITY_MIN)
		if(!(North || South || East || West) && Distance < QT_COMPLEXITY_MIN)
		{
			// low detail quad
			TexNode->RS.AddIndices(NW, SW, NE);
			TexNode->RS.AddIndices(NE, SW, SE);
		}
		else
		{
			const unsigned short N = Indices[QT_INDEX_N];
			const unsigned short S = Indices[QT_INDEX_S];
			const unsigned short W = Indices[QT_INDEX_W];
			const unsigned short E = Indices[QT_INDEX_E];
			const unsigned short C = Indices[QT_INDEX_C];
			// north
			if(North)
			{
				// two triangles, so no gaps occur
				TexNode->RS.AddIndices(NW, C, N);
				TexNode->RS.AddIndices(N, C, NE);
			}
			else
				TexNode->RS.AddIndices(NW, C, NE);
			// east
			if(East)
			{
				// two triangles, so no gaps occur
				TexNode->RS.AddIndices(NE, C, E);
				TexNode->RS.AddIndices(E, C, SE);
			}
			else
				TexNode->RS.AddIndices(NE, C, SE);
			// south
			if(South)
			{
				// two triangles, so no gaps occur
				TexNode->RS.AddIndices(SE, C, S);
				TexNode->RS.AddIndices(S, C, SW);
			}
			else
				TexNode->RS.AddIndices(SE, C, SW);
			// west
			if(West)
			{
				// two triangles, so no gaps occur
				TexNode->RS.AddIndices(SW, C, W);
				TexNode->RS.AddIndices(W, C, NW);
			}
			else
				TexNode->RS.AddIndices(SW, C, NW);
		}
	}
	else
		QTNode::CreateScene(DS);
}

void QTNodeTerra::CreateTerrain(void)
{
	// get QTNodeTerraTex
	QTNodeTerra *Srch = this;
	while(Srch->Level > Tree->Specs.TexDepth)
		Srch = (QTNodeTerra *)Srch->Parent;
	TexNode = (QTNodeTerraTex *)Srch;
	if(Level > Tree->Specs.TexDepth)
	{
		QTNodeTerra *Parent = (QTNodeTerra *)QTNodeTerra::Parent;
		switch(Flags & QT_CHILD_MASK)
		{
		case QT_CHILD_NW:
			Indices[QT_INDEX_SW] = Parent->Indices[QT_INDEX_W];
			Indices[QT_INDEX_SE] = Parent->Indices[QT_INDEX_C];
			Indices[QT_INDEX_NW] = Parent->Indices[QT_INDEX_NW];
			Indices[QT_INDEX_NE] = Parent->Indices[QT_INDEX_N];
			break;
		case QT_CHILD_NE:
			Indices[QT_INDEX_SW] = Parent->Indices[QT_INDEX_C];
			Indices[QT_INDEX_SE] = Parent->Indices[QT_INDEX_E];
			Indices[QT_INDEX_NW] = Parent->Indices[QT_INDEX_N];
			Indices[QT_INDEX_NE] = Parent->Indices[QT_INDEX_NE];
			break;
		case QT_CHILD_SW:
			Indices[QT_INDEX_SW] = Parent->Indices[QT_INDEX_SW];
			Indices[QT_INDEX_SE] = Parent->Indices[QT_INDEX_S];
			Indices[QT_INDEX_NW] = Parent->Indices[QT_INDEX_W];
			Indices[QT_INDEX_NE] = Parent->Indices[QT_INDEX_C];
			break;
		case QT_CHILD_SE:
			Indices[QT_INDEX_SW] = Parent->Indices[QT_INDEX_S];
			Indices[QT_INDEX_SE] = Parent->Indices[QT_INDEX_SE];
			Indices[QT_INDEX_NW] = Parent->Indices[QT_INDEX_C];
			Indices[QT_INDEX_NE] = Parent->Indices[QT_INDEX_E];
			break;
		}
		Indices[QT_INDEX_S] = (Indices[QT_INDEX_SW] + Indices[QT_INDEX_SE]) / 2;
		Indices[QT_INDEX_W] = (Indices[QT_INDEX_NW] + Indices[QT_INDEX_SW]) / 2;
		Indices[QT_INDEX_E] = (Indices[QT_INDEX_NE] + Indices[QT_INDEX_SE]) / 2;
		Indices[QT_INDEX_N] = (Indices[QT_INDEX_NW] + Indices[QT_INDEX_NE]) / 2;
		Indices[QT_INDEX_C] = (Indices[QT_INDEX_N ] + Indices[QT_INDEX_S ]) / 2;
	}

	QTNode::CreateTerrain();

	// BBox, VBox
	if(Level == Tree->Specs.Depth)
	{
		BBox.Data[BB_BOTTOM] = TexNode->Vertices[Indices[QT_INDEX_SW]].y;
		BBox.Data[BB_TOP] = TexNode->Vertices[Indices[QT_INDEX_SW]].y;
		for(int i = 0; i < 9; i++)
		{
			if(TexNode->Vertices[Indices[i]].y < BBox.Data[BB_BOTTOM])
				BBox.Data[BB_BOTTOM] = TexNode->Vertices[Indices[i]].y;
			if(TexNode->Vertices[Indices[i]].y > BBox.Data[BB_TOP])
				BBox.Data[BB_TOP] = TexNode->Vertices[Indices[i]].y;
		}
		VBox.Data[BB_LEFT] = BBox.Data[BB_LEFT] - 8.0f;
		VBox.Data[BB_RIGHT] = BBox.Data[BB_RIGHT] + 8.0f;
		VBox.Data[BB_NEAR] = BBox.Data[BB_NEAR] - 8.0f;
		VBox.Data[BB_FAR] = BBox.Data[BB_FAR] + 8.0f;
		VBox.Data[BB_BOTTOM] = BBox.Data[BB_BOTTOM] - 8.0f;
		VBox.Data[BB_TOP] = BBox.Data[BB_TOP] + 8.0f;
		// set center
		Center = Vector(0.5f * (BBox.Data[BB_LEFT] + BBox.Data[BB_RIGHT]),
						0.5f * (BBox.Data[BB_BOTTOM] + BBox.Data[BB_TOP]),
						0.5f * (BBox.Data[BB_NEAR] + BBox.Data[BB_FAR]));
	}
	// Complexity
	CalculateComplexity();
}

// ****** QTNodeTerraTex ******

QTNodeTerraTex::QTNodeTerraTex(QTNode::TreeCreateStruct *TCS)
:QTNodeTerra(TCS), Texture(0)
{
}

void QTNodeTerraTex::CreateVertices(QuadTreeSpecs *QTSpecs)
{
	RS.ClearVertex();
//	for(int j = VtxSize - 1; j >= 0; j--) // z
	for(int j = 0; j < VtxSize; j++) // z
		for(int i = 0; i < VtxSize; i++) // x
		{
			float XFac = (float)i / (VtxSize - 1);
			float ZFac = (float)j / (VtxSize - 1);
			float x = (1 - XFac) * BBox.Data[BB_LEFT] + XFac * BBox.Data[BB_RIGHT];
			float z = (1 - ZFac) * BBox.Data[BB_NEAR] + ZFac * BBox.Data[BB_FAR];
//			float z = (1 - ZFac) * BBox.Data[BB_FAR] + ZFac * BBox.Data[BB_NEAR];
			Vertices[j * VtxSize + i] = VertexTex2(Vector(x, QTSpecs->Surface->GetY(x, z), z), QTSpecs->Surface->GetNormal(x, z), Vector2D(XFac, 1 - ZFac), Vector2D(16 * XFac, 16 * (1 - ZFac)));
			RS.AddVertex(Vertices[j * VtxSize + i]);
		}
	RS.CreateVBuf(QTSpecs->DS->D3DDevice);

	Texture = Tree->Specs.Surface->GetTexture(QTSpecs->DS, BBox.Data[BB_LEFT], BBox.Data[BB_NEAR], BBox.Data[BB_RIGHT], BBox.Data[BB_FAR]);
}

void QTNodeTerraTex::CreateTextures(D3DStruct *DS)
{
	RELEASE(Texture);
	Texture = Tree->Specs.Surface->GetTexture(DS, BBox.Data[BB_LEFT], BBox.Data[BB_NEAR], BBox.Data[BB_RIGHT], BBox.Data[BB_FAR]);
	QTNode::CreateTextures(DS);
}

void QTNodeTerraTex::ReconstructMipMaps(D3DStruct *DS)
{
	I3D::TextureManager::CreateMipMaps(Texture, 0);
}

void QTNodeTerraTex::Init3D(D3DStruct *DS, bool recurse)
{
	Texture = Tree->Specs.Surface->GetTexture(DS, BBox.Data[BB_LEFT], BBox.Data[BB_NEAR], BBox.Data[BB_RIGHT], BBox.Data[BB_FAR]);
	if(recurse)
		Water.Init3D(DS, true);
	RS.CreateVBuf(DS->D3DDevice);
	RS.CreateIBuf(DS->D3DDevice);
	if(recurse)
		QTNode::Init3D(DS, true);
}

void QTNodeTerraTex::CreateView(D3DStruct *DS)
{
}

void QTNodeTerraTex::RenderWater(D3DStruct *DS)
{
	if(Water.GetVisible())
		Water.Render(DS);
}

void QTNodeTerraTex::Render(D3DStruct *DS)
{
	if(Texture)
		DS->D3DDevice->SetTexture(0, Texture);
	DS->D3DDevice->SetSamplerState(0,D3DSAMP_ADDRESSU,D3DTADDRESS_CLAMP); 
	DS->D3DDevice->SetSamplerState(0,D3DSAMP_ADDRESSV,D3DTADDRESS_CLAMP); 
	//DS->D3DDevice->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
	//DS->D3DDevice->SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_CLAMP);
	RenderableIndexTex2::Render(DS);
	DS->D3DDevice->SetSamplerState(0,D3DSAMP_ADDRESSU,D3DTADDRESS_WRAP); 
	DS->D3DDevice->SetSamplerState(0,D3DSAMP_ADDRESSV,D3DTADDRESS_WRAP); 
	//DS->D3DDevice->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP);
	//DS->D3DDevice->SetTextureStageState(0, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP);
}

void QTNodeTerraTex::CreateScene(D3DStruct *DS)
{
	Tree->AddRenderableGlobal(this);
	RS.ClearIndexElements();

	QTNodeTerra::CreateScene(DS);

	RS.CreateIBuf(Tree->Specs.DS->D3DDevice);
	// Water
	Water.SetVisible((Flags & QT_FLAG_VISIBLE_PARTLY) != 0 && BBox.Data[BB_BOTTOM] < WATERLEVEL);
}

void QTNodeTerraTex::CreateTerrain(void)
{
	// create vertices
	VtxSize = (1 << (Tree->Specs.Depth - Tree->Specs.TexDepth + 1)) + 1;
	Vertices = new VertexTex2[VtxSize * VtxSize];
	CreateVertices(&Tree->Specs);
	// create indices
	RS.ClearIndex();
	int Max = VtxSize - 1;
	int Mid = Max / 2;
	Indices[QT_INDEX_SW] = 0;
	Indices[QT_INDEX_S ] = Mid;
	Indices[QT_INDEX_SE] = Max;
	Indices[QT_INDEX_W ] = Mid * VtxSize;
	Indices[QT_INDEX_C ] = Mid * VtxSize + Mid;
	Indices[QT_INDEX_E ] = Mid * VtxSize + Max;
	Indices[QT_INDEX_NW] = Max * VtxSize;
	Indices[QT_INDEX_N ] = Max * VtxSize + Mid;
	Indices[QT_INDEX_NE] = Max * VtxSize + Max;

	Water.Create(Tree->Specs.DS, &BBox, -19);

	QTNodeTerra::CreateTerrain();
}

QTNodeTerraTex::~QTNodeTerraTex()
{
	delete[] Vertices;
}

}
