/***********************************************************/
/*                                                         */
/* QTNode.h                                                */
/*                                                         */
/* created 11.08.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef QTNODE_H
#define QTNODE_H

#include "D3DStructures.h"
#include "DataStructures.h"
#include "Renderable.h"
#include "RenderableModel.h"
#include "DynamicArray.h"
#include "WaterPlane.h"
#include "QuadTree.h"

// www.infinitecode.com 3d library
namespace I3D
{

// flag and index values
enum
{
	// index indices
	QT_INDEX_SW = 0,
	QT_INDEX_S  = 1,
	QT_INDEX_SE = 2,
	QT_INDEX_W  = 3,
	QT_INDEX_C  = 4,
	QT_INDEX_E  = 5,
	QT_INDEX_NW = 6,
	QT_INDEX_N  = 7,
	QT_INDEX_NE = 8,
	// neighbor direction indices
	QT_NEIGHBOR_N = 0,
	QT_NEIGHBOR_S = 1,
	QT_NEIGHBOR_W = 2,
	QT_NEIGHBOR_E = 3,
	// child indices (used in TreeCreateStruct::Flags and QTNode::Flags)
	QT_CHILD_SW = 0,
	QT_CHILD_SE = 1,
	QT_CHILD_NW = 2,
	QT_CHILD_NE = 3,
	QT_CHILD_MASK = 3,
	// QTNode flags
	QT_FLAG_SUBDIV =				4,
	QT_FLAG_VISIBLE_PARTLY =		8,
	QT_FLAG_VISIBLE_COMPLETELY =	16,
	QT_FLAG_VISIBLE_MASK =			16 + 8
};

// in QuadTree.h:
class QuadTree;
struct QuadTreeSpecs;
class XZPos;

class QTNode
{
	class NodeInfo
	{
	public:
		IMATH::Vector2D ScreenCenter;
		IMATH::Vector Center;
		float Coeff, Error;

		NodeInfo() {}

		NodeInfo(const IMATH::Vector2D &SC, const IMATH::Vector &C)
			: ScreenCenter(SC), Center(C) {}
	};

public:
	struct TreeCreateStruct
	{
		int Level;
		QTNode *Parent;
		int Flags;
		QuadTree *Tree;
	};

	int Flags;
	int Level;
	QTNode *Parent;
	QTNode *Child[4];
	QTNode *Neighbors[4];
	QuadTree *Tree;
	BoundingBox BBox, VBox, ScreenBBox; // Bounding box, outer bbox, and screen bbox for node visibility calculation
	Vector Center;
	Vector2D ScreenCenter;

	QTNode(TreeCreateStruct *TCS);
	virtual ~QTNode(void);

	// create terrain vertices and indices
	virtual void CreateTerrain(void);
	// clear subdivision flags
	void ClearSubdivisions(void);
	// calculate terrain visibility for CreateScene
	virtual void Visibility(D3DStruct *DS);
	// balance subtrees for CreateScene
	virtual void BalanceSubTrees(D3DStruct *DS);
	// add visible indices and objects to the quadtree
	virtual void CreateScene(D3DStruct *DS);
	// add a model to the node
	virtual void AddManagedObject(ManagedObject *Obj);
	virtual int DeleteManagedObject(ManagedObject *Obj);
	// (re)create terraintextures
	virtual void CreateTextures(D3DStruct *DS);
	virtual void ReconstructMipMaps(D3DStruct *DS);
	virtual void RenderWater(D3DStruct *DS);
	virtual void Init3D(D3DStruct *DS, bool recurse);

	// get the node at a given position in a given level
	QTNode *GetNode(const IMATH::Vector2DXZ *Pos, int Level);
	void CalculateScreenBBox(D3DStruct *DS);
	bool IsInScreenBBox(const IMATH::Vector2D &Pos);
	IMATH::Vector2D GetScreenBBoxCenter();
	BoundingBox GetScreenBBox();
	IMATH::Vector GetBBoxCenter();
	IMATH::Vector2D GetScreenCenter();
	//IMATH::Vector2D GetScreenCorner(int id);
	IMATH::Vector2D ToScreenCoords(const IMATH::Vector &V);
	QTNode *GetNode(const IMATH::Vector2D *Pos);
	XZPos GetXZPos(const IMATH::Vector2D &Pos);

	IMATH::Vector2D GetScreenCorner(int id)
	{
		IMATH::Vector Corner;
		switch(id)
		{
		case QT_INDEX_NW:
			Corner = IMATH::Vector(BBox.Data[BB_LEFT], Tree->Specs.Surface->GetY(BBox.Data[BB_LEFT], BBox.Data[BB_FAR]), BBox.Data[BB_FAR]);
			break;
		case QT_INDEX_SW:
			Corner = IMATH::Vector(BBox.Data[BB_LEFT], Tree->Specs.Surface->GetY(BBox.Data[BB_LEFT], BBox.Data[BB_NEAR]), BBox.Data[BB_NEAR]);
			break;
		case QT_INDEX_NE:
			Corner = IMATH::Vector(BBox.Data[BB_RIGHT], Tree->Specs.Surface->GetY(BBox.Data[BB_RIGHT], BBox.Data[BB_FAR]), BBox.Data[BB_FAR]);
			break;
		case QT_INDEX_SE:
			Corner = IMATH::Vector(BBox.Data[BB_RIGHT], Tree->Specs.Surface->GetY(BBox.Data[BB_RIGHT], BBox.Data[BB_NEAR]), BBox.Data[BB_NEAR]);
			break;
		}
		IMATH::Vector ScreenCorner3D = Tree->Specs.DS->ProjViewMtx * Corner;
		if(ScreenCorner3D.z < 0 || ScreenCorner3D.z > 1)
		{
			ScreenCorner3D.x = -ScreenCorner3D.x;
			ScreenCorner3D.y = -ScreenCorner3D.y;
		}
		return IMATH::Vector2D(ScreenCorner3D.x, ScreenCorner3D.y);
	}
};

// ****** QTNodeRoot ******

class QTNodeRoot: public QTNode
{
public:
	QTNodeRoot(TreeCreateStruct *TCS);
	//virtual ~QTNodeRoot() {}

	// calculate terrain visibility
	virtual void Visibility(D3DStruct *DS);
};

// ****** QTNodeTerra ******

class QTNodeTerraTex;

class QTNodeTerra: public QTNode
{
	// terrain
	QTNodeTerraTex *TexNode;
	float Complexity, Distance;
	// subobjects
	ISL::DynamicArray<ManagedObject *> Objects;

	void ShowNeighbors(D3DStruct *DS);
	void Show(D3DStruct *DS);

protected:
	unsigned short Indices[9];

public:
	QTNodeTerra(TreeCreateStruct *TCS);

	// create terrain vertices and indices
	virtual void CreateTerrain(void);
	// calculate terrain visibility for CreateScene
	virtual void Visibility(D3DStruct *DS);
	// balance subtrees for CreateScene
	virtual void BalanceSubTrees(D3DStruct *DS);
	// add visible indices and objects to the quadtree
	virtual void CreateScene(D3DStruct *DS);
	// add a model to the node
	virtual void AddManagedObject(ManagedObject *Obj);
	virtual int DeleteManagedObject(ManagedObject *Obj);
	// calculate terrain complexity for detailbased LODing
	void CalculateComplexity(void);
	// calculate visibility for culling
	void CalculateVisibility(D3DStruct *DS);
};

// ****** QTNodeTerraTex ******

class QTNodeTerraTex: public QTNodeTerra, RenderableIndexTex2
{
	friend QTNodeTerra;

	LPDIRECT3DTEXTURE9 Texture;
	// terrain vertices
	VertexTex2 *Vertices;
	int VtxSize;
	WaterPlane Water;

	void CreateVertices(QuadTreeSpecs *QTSpecs);

protected:
	virtual void CreateView(D3DStruct *DS);
	virtual void CreateTextures(D3DStruct *DS);
	virtual void ReconstructMipMaps(D3DStruct *DS);
	virtual void Init3D(D3DStruct *DS, bool recurse);

public:

	LPDIRECT3DTEXTURE9 GetTerrainTexture(void)
	{
		return Texture;
	}

	QTNodeTerraTex(TreeCreateStruct *TCS);
	~QTNodeTerraTex(void);

	virtual void Render(D3DStruct *DS);

	// create terrain vertices and indices
	virtual void CreateTerrain(void);
	// add visible indices and objects to the quadtree
	virtual void CreateScene(D3DStruct *DS);
	virtual void RenderWater(D3DStruct *DS);
};

}

#endif
