/***********************************************************/
/*                                                         */
/* QuadTree.cpp                                            */
/*                                                         */
/* created 16.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

// roam quadtree

#include "QuadTree.h"
#include "QTNode.h"
#include "TextureManager.h"

// www.infinitecode.com 3d library
namespace I3D
{

QuadTree::QuadTree(void)
:Root(0)
{
}

QuadTree::~QuadTree(void)
{
	if(Root)
	{
		delete Root;
		Root = 0;
	}
}

void QuadTree::Create(QuadTreeSpecs *Specs)
{
	ISpaceMgr.Tree = this;
	QuadTree::Specs = *Specs;
	Prf.Create(QuadTree::Specs.DS);
	Prf.SetName(0, "ClearSubdivisions");
	Prf.SetName(1, "Visibility");
	Prf.SetName(2, "BalaceSubTrees");
	Prf.SetName(3, "CreateScene");
	Prf.SetName(4, "ObjectsRnd");
	Prf.SetName(5, "Managed Objects");
	Prf.SetName(6, "Water");
	QTNode::TreeCreateStruct TCS;
	TCS.Level = 0;
	TCS.Parent = NULL;
	TCS.Flags = 0;
	TCS.Tree = this;
	if(Root)
	{
		delete Root;
		Root = 0;
	}
	try
	{
		Root = new QTNodeRoot(&TCS);
	}
	catch(char *s)
	{
		Root = 0;
		throw(s);
	}
	catch(...)
	{
		Root = 0;
		throw("QuadTree::Create#new QTNodeRoot");
	}
	//Root->CreateTerrain();
	// create vertices
	//RS.ClearVertex();
	//CreateVertices(Root);
	//RS.CreateVBuf(Specs->DS->D3DDevice);
	// load texture
	//Texture = TextureManager::Load(Specs->DS->D3DDevice, "Textures\\Terrain.bmp", NULL, 0);
	Texture = TextureManager::GetTexture(Specs->DS, "Terrain");
	DetailTexture = TextureManager::GetTexture(Specs->DS, "Textures\\TerrainDetail.bmp", 0, TM_FLAG_GRAYMIPMAPS);
	// create view
	OnViewChange();
}

void QuadTree::CreateTextures(D3DStruct *DS)
{
	Root->CreateTextures(DS);
}

void QuadTree::ReconstructMipMaps()
{
	Root->ReconstructMipMaps(Specs.DS);
}

void QuadTree::AddTexture(const IMATH::Vector2DXZ &Pos, const IMATH::Vector2DXZ &Offset, const IMATH::Vector2DXZ &Size, LPDIRECT3DTEXTURE9 Texture, bool ReconstructMipMaps)
{
	int State = 0;
	try
	{
		State = 1;
		// lock
		D3DLOCKED_RECT LRect;
		memset(&LRect, 0, sizeof(LRect));
		State = 2;
		if(Texture->LockRect(0, &LRect, NULL, 0) != D3D_OK)
			throw("QuadTree::AddTexture can't lock texture (1)");
		State = 3;
		// buffers
		unsigned int *Tex=(unsigned int*)LRect.pBits;
		// calculate step
		D3DSURFACE_DESC TDesc, Desc;
		State = 4;
		Texture->GetLevelDesc(0, &Desc);
		State = 5;
		LPDIRECT3DTEXTURE9 Text = GetTerrainTexture(IMATH::Vector(Pos.x, 0, Pos.z));
		if(!Text)
		{
			if(Texture->UnlockRect(0) != D3D_OK)
				throw("QuadTree::AddTexture can't unlock texture (1)");
			return;
		}
		State = 6;
		Text->GetLevelDesc(0, &TDesc);
		State = 7;
		if(Desc.Width == 0 || TDesc.Width == 0)
			throw("QuadTree::AddTexture: texture size is 0");
		int TexRes = 1;
		for(int i = 0; i < Specs.TexDepth; i++)
			TexRes *= 2;
		State = 8;
		float StepX = (float)Desc.Width / (float)TDesc.Width;
		float StepZ = (float)Desc.Height / (float)TDesc.Height;
		State = 9;
		//float Step = 1.0f / ((float)TexRes);
		// blit
		LPDIRECT3DTEXTURE9 TTex = 0;
		D3DSURFACE_DESC TTexDesc;
		ISL::DynamicArray<LPDIRECT3DTEXTURE9> ChangedTextures;
		unsigned int *TTexBits = 0;
		for(float x = 0; x <= Size.x - 1; x += StepX)
			for(float z = 0; z <= Size.z - 1; z += StepZ)
			{
				float tx = x;//(x * Specs.Size) / (Desc.Width * TexRes);
				float tz = z;//(z * Specs.Size) / (Desc.Height * TexRes);
				float ux = max(-0.5f * Specs.Size, min(Pos.x + tx, 0.5f * Specs.Size));
				float uz = max(-0.5f * Specs.Size, min(Pos.z + tz, 0.5f * Specs.Size));
				State = 10;
				LPDIRECT3DTEXTURE9 NTex = GetTerrainTexture(IMATH::Vector(ux, 0, uz));
				if(NTex != TTex)
				{
					if(TTex && TTex->UnlockRect(0) != D3D_OK)
						throw("QuadTree::AddTexture can't unlock texture (4)");
					TTex = NTex;
					ChangedTextures.Add(TTex);
					D3DLOCKED_RECT LRect;
					memset(&LRect, 0, sizeof(LRect));
					if(TTex->LockRect(0, &LRect, NULL, 0) != D3D_OK)
						throw("QuadTree::AddTexture can't lock texture (2)");
					TTex->GetLevelDesc(0, &TTexDesc);
					// buffers
					TTexBits = (unsigned int*)LRect.pBits;
				}
				State = 11;
				//I3D::TextureManager::SetPixel(TTex, GetTexturePosX(ux), GetTexturePosZ(uz), Tex[((int)(z + Offset.z)) * Desc.Height + (int)(x + Offset.x)]);
				unsigned int Pixel = Tex[((int)(z + Offset.z)) * Desc.Height + (int)(x + Offset.x)];
				unsigned char Alpha = Pixel >> 24;
				unsigned char Red = (Pixel >> 16) & 0xff;
				unsigned char Green = (Pixel >> 8) & 0xff;
				unsigned char Blue = Pixel & 0xff;
				unsigned int OldPixel = TTexBits[GetTexturePosZ(uz) * TTexDesc.Height + GetTexturePosX(ux)];
				unsigned char OldAlpha = OldPixel >> 24;
				unsigned char OldRed = (OldPixel >> 16) & 0xff;
				unsigned char OldGreen = (OldPixel >> 8) & 0xff;
				unsigned char OldBlue = OldPixel & 0xff;
				TTexBits[GetTexturePosZ(uz) * TTexDesc.Height + GetTexturePosX(ux)] =
					  (((Alpha * Alpha + (0xff - Alpha) * OldAlpha) / 0xff) << 24)
					+ (((Alpha * Red + (0xff - Alpha) * OldRed) / 0xff) << 16)
					+ (((Alpha * Green + (0xff - Alpha) * OldGreen) / 0xff) << 8)
					+ ((Alpha * Blue + (0xff - Alpha) * OldBlue) / 0xff);
				State = 12;
				//I3D::TextureManager::Brightness(TTex, GetTexturePosX(tx), GetTexturePosZ(tz), -0x80);
			}
		State = 13;
		// unlock
		if(TTex && TTex->UnlockRect(0) != D3D_OK)
			throw("QuadTree::AddTexture can't unlock texture (2)");
		if(Texture->UnlockRect(0) != D3D_OK)
			throw("QuadTree::AddTexture can't unlock texture (3)");
		if(ReconstructMipMaps)
			for(int i = 0; i < ChangedTextures.Count(); i++)
				TextureManager::CreateMipMaps(ChangedTextures[i], 0);
		State = 14;
	}
	catch(char *s)
	{
		throw s;
	}
	catch(...)
	{
		char Str[1024];
		sprintf(Str, "QuadTree::AddTexture#%d (%f, %f)", State, Pos.x, Pos.z);
		throw Str;
	}
}

int QuadTree::GetTexturePosX(float x)
{
	int TexLevel = Specs.TexDepth;
	int Level = 2;
	while(--TexLevel)
		Level *= 2;
	return (int)((x / Specs.Size + 0.5f) * 256 * Level - 1) % 256;
}

int QuadTree::GetTexturePosZ(float z)
{
	int TexLevel = Specs.TexDepth;
	int Level = 2;
	while(--TexLevel) // TODO: faster with exp()?
		Level *= 2;
	return 255 - (int)((z / Specs.Size + 0.5f) * 256 * Level - 1) % 256;
}

// call this, when the heightmap is modified
void QuadTree::UpdateSurface(void)
{
	Root->CreateTerrain();
	OnViewChange();
}

LPDIRECT3DTEXTURE9 QuadTree::GetTerrainTexture(const IMATH::Vector &Pos)
{
	int TexLevel = Specs.TexDepth;
	int Level = 2;
	while(--TexLevel)
		Level *= 2;
	int x = (int)((Pos.x / Specs.Size + 0.5f) * 256 * Level - 1) / 256;
	int z = (int)((Pos.z / Specs.Size + 0.5f) * 256 * Level - 1) / 256;
	if(x < 0) x = 0;
	if(x > Level - 1) x = Level - 1;
	if(z < 0) z = 0;
	if(z > Level - 1) z = Level - 1;
	IMATH::Vector Pos2 = IMATH::Vector((float)(128 + x * 256) - Specs.Size / 2, 0, (float)(128 + z * 256) - Specs.Size / 2);

	IMATH::Vector2DXZ PosXZ(Pos2.x, Pos2.z);
	QTNode *Node = GetNode(&PosXZ, Specs.TexDepth);
	if(Node)
		return ((QTNodeTerraTex *)Node)->GetTerrainTexture();
	else 
		return 0;
}

void QuadTree::CreateView(D3DStruct *DS)
{
	ObjectsRnd.Clear();
	VisibleObjects.Clear();
	Root->Visibility(DS);
	for(int i = 0; i < ObjectsRnd.Count(); i++)
		ObjectsRnd[i]->OnViewChange();
}

void QuadTree::CheckVisibleObjects(int ID)
{
	try
	{
		// for textures
		for(int i = 0; i < VisibleObjects.CountKeys(); i++)
		{
			/*LPDIRECT3DTEXTURE9 Txt;
			Txt = VisibleObjects.GetKey(i);
			DS->D3DDevice->SetTexture(0, Txt);*/
			// for models
			for(int j = 0; j < VisibleObjects.GetValuePtr(i)->CountKeys(); j++)
			{
				NodeModel *Ptr;
				Model *RM;
				NodeModel::Entry::HTEValues *PVP;
				Ptr = VisibleObjects.GetValuePtr(i); // texture
				RM = Ptr->GetKey(j); // texture->model
				/*RM->PreRender(DS);
				RM->RenderSetup(DS);*/
				for(int k = 0; k < Ptr->GetValuesPtr(j)->Count(); k++)
				{
					//(*VisibleObjects.GetValuePtr(i)->GetValuesPtr(j))[k]->Render(DS);
					PVP = Ptr->GetValuesPtr(j); // texture->model->object<RenderableModel>
					if((*PVP)[k]->UnLinked)
					{
						char Str[1024];
						sprintf(Str, "QTree::CheckVisibleObjects()#unlinked object found#1#%d", ID);
						throw(Str);
					}
					/*DS->D3DDevice->SetTransform(D3DTS_WORLD, (*PVP)[k]->Mtx);
					DS->D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, RM->RS.GetVertexNumber(), 0, RM->RS.GetIndexNumber() / 3);*/
				}
				//RM->PostRender(DS);
			}
		}
	}
	catch(char *s)
	{
		throw(s);
	}
	catch(...)
	{
		char Str[1024];
		sprintf(Str, "QTree::CheckVisibleObjects()#unlinked object found#2#%d", ID);
		throw(Str);
	}
}

void QuadTree::Init3D(D3DStruct *DS, bool recurse)
{
	Texture = TextureManager::GetTexture(DS, "Terrain");
	DetailTexture = TextureManager::GetTexture(DS, "Textures\\TerrainDetail.bmp", 0, TM_FLAG_GRAYMIPMAPS);
	if(recurse)
	{
		Root->Init3D(DS, true);
		Prf.Init3D(DS, true);
	}
	OnViewChange();
}

void QuadTree::Render(D3DStruct *DS)
{
	int State = 0;
	Model *RM;
	try
	{
		State = 1;
		if(!(RenderFlags & RENDERABLE_FLAG_VIEWCHANGE))
		{
			Prf.SetToZero(0);
			Prf.SetToZero(1);
			Prf.SetToZero(2);
			Prf.SetToZero(3);
		}
		I3D::Matrix M;
		M.SetIdentityMatrix();
		DS->D3DDevice->SetTransform(D3DTS_WORLD, &M);
		Renderable::Render(DS); // calls CreateView when necessary
		// render terrain
		DS->D3DDevice->SetTexture(0, Texture);
		DS->D3DDevice->SetTexture(1, DetailTexture);
		int i;
		DS->D3DDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_ADDSIGNED);
		Prf.Start(4);
		// render terrain and other renderables
		for(i = 0; i < ObjectsRnd.Count(); i++)
			ObjectsRnd[i]->Render(DS);
		Prf.Stop(4);
		DS->D3DDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_DISABLE);
		// render managed objects
		Prf.Start(5);
		int Textures = VisibleObjects.CountKeys();
		State = 2;
		// for textures
		for(i = 0; i < Textures; i++)
		{
			NodeModel *Ptr = VisibleObjects.GetValuePtr(i); // texture
			LPDIRECT3DTEXTURE9 Txt = VisibleObjects.GetKey(i);
			State = 20;
			DS->D3DDevice->SetTexture(0, Txt);
			State = 21;
			int ModelNum = VisibleObjects.GetValuePtr(i)->CountKeys();
			// for models
			for(int j = 0; j < ModelNum; j++)
			{
				RM = Ptr->GetKey(j); // texture->model
				NodeModel::Entry::HTEValues *PVP = Ptr->GetValuesPtr(j); // texture->model->object<RenderableModel>
				State = 22;
				RM->PreRender(DS);
				State = 23;
				RM->RenderSetup(DS);
				State = 24;
				int ObjectNum = Ptr->GetValuesPtr(j)->Count();
				int VNumber = RM->RS.GetVertexNumber();
				int INumber3 = RM->RS.GetIndexNumber() / 3;
				// for objects
				/*Prf.Stop(5);
				Prf.Start(6);*/
				for(int k = 0; k < ObjectNum; k++)
				{
					/*Vector ScreenCenter = (*PVP)[k]->GetPos();
					Vector ScreenDeltaX = (*PVP)[k]->GetPos() + IMATH::Vector(1, 0, 0);
					Vector ScreenDeltaY = (*PVP)[k]->GetPos() + IMATH::Vector(0, 1, 0);
					Vector ScreenDeltaZ = (*PVP)[k]->GetPos() + IMATH::Vector(0, 0, 1);
					ScreenCenter = DS->ProjViewMtx * ScreenCenter;
					ScreenDeltaX = DS->ProjViewMtx * ScreenDeltaX;
					ScreenDeltaY = DS->ProjViewMtx * ScreenDeltaY;
					ScreenDeltaZ = DS->ProjViewMtx * ScreenDeltaZ;
					float DistanceX = (ScreenCenter - ScreenDeltaX).Length();
					float DistanceY = (ScreenCenter - ScreenDeltaY).Length();
					float DistanceZ = (ScreenCenter - ScreenDeltaZ).Length();
					float Dist = max(DistanceX, max(DistanceY, DistanceZ));*/
/*					IMATH::Vector e1(1, 0, 0);
					e1 = *(*PVP)[k]->Mtx * e1;
					e1 = DS->ProjViewMtx * e1;
					IMATH::Vector e2(0, 1, 0);
					e2 = *(*PVP)[k]->Mtx * e2;
					e2 = DS->ProjViewMtx * e2;
					IMATH::Vector e3(0, 0, 1);
					e3 = *(*PVP)[k]->Mtx * e3;
					e3 = DS->ProjViewMtx * e3;
					float Dist = e1.Length() + e2.Length() + e3.Length();
*/
					DS->D3DDevice->SetTransform(D3DTS_WORLD, (*PVP)[k]->Mtx);
					DS->D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, VNumber, 0, INumber3);

					/*if(Dist > 0.003f)
						DS->D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, VNumber, 0, INumber3);
					else if(Dist > 0.001f)
						DS->D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, VNumber, 0, min(INumber3, 100));
					else
						DS->D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, VNumber, 0, min(INumber3, 50));*/
				}
				/*Prf.Stop(6);
				Prf.Start(5);*/
				State = 25;
				RM->PostRender(DS);
			}
		}
		State = 3;
		Prf.Stop(5);
		// render water
		Prf.Start(6);
		Root->RenderWater(DS);
		Prf.Stop(6);
		Prf.Render(DS);
		State = 4;
	}
	catch(char *s)
	{
		throw(s);
	}
	catch(...)
	{
		char Str[1024];
		sprintf(Str, "QTree::Render()#%i %s", State, RM->GetStandardTexture());
		throw(Str);
	}
}

void QuadTree::AddRenderableGlobal(Renderable *Obj)
{
	ObjectsRnd[ObjectsRnd.Count()] = Obj;
}

void QuadTree::CISpaceMgr::AddManagedObject(ManagedObject *Obj)
{
	if(!Obj)
		throw("QuadTree::CISpaceMgr::AddManagedObject added NULL Object");
	Tree->Root->AddManagedObject(Obj);
}

void QuadTree::CISpaceMgr::DeleteManagedObject(ManagedObject *Obj)
{
	Tree->DeleteVisibleObject(Obj);
	int NumDeleted = Tree->Root->DeleteManagedObject(Obj);
}

void QuadTree::CISpaceMgr::MoveManagedObject(ManagedObject *Obj)
{
	// TODO: fix bug with actor visibility
	IMATH::Vector2DXZ Point = IMATH::Vector2DXZ(Obj->GetPos().x, Obj->GetPos().z);
	if((!Obj->Node) || (!((QTNode *)Obj->Node)->BBox.IsIn(&Point)))
	{
		//bool OldVisible = Obj->GetVisible();
		DeleteManagedObject(Obj);
		AddManagedObject(Obj);
		//QTNode *NewNode = ((ObjectInformation *)Obj->GetData())->DeepestNode;
		/*if(!Obj->GetVisible() && OldVisible)
		{
			//DeleteVisibleObject(Obj);
		}
		if(Obj->GetVisible() && !OldVisible)
		{
			//AddVisibleObject(Obj);
		}*/
	}
}

void QuadTree::AddVisibleObject(ManagedObject *Obj)
{
	if(!Obj)
		throw("QuadTree::AddVisibleObject added NULL Object");
	RenderableModel *RM;
	if(RM = dynamic_cast<RenderableModel *>(Obj))
	{
		if(!RM->Mdl)
			throw("QuadTree::AddVisibleObject: RM->Mdl is NULL");
		if(VisibleObjects.GetValuePtr(RM->Texture))
			VisibleObjects.GetValuePtr(RM->Texture)->AddToSet(RM->Mdl, RM);
		else
			VisibleObjects.Add(RM->Texture, NodeModel(RM));
	}
	ParticleEffect *PFX;
	if(PFX = dynamic_cast<ParticleEffect *>(Obj))
	{
		VisiblePFX[VisiblePFX.Count()] = PFX;
	}
}

void QuadTree::DeleteVisibleObject(ManagedObject *Obj)
{
	RenderableModel *RM;
	if(RM = dynamic_cast<RenderableModel *>(Obj))
	{
		bool Deleted = false;
		int Num = 0;
		do
		{
			NodeModel *NM = VisibleObjects.GetValuePtr(RM->Texture);
			if(NM)
			{
				NodeModel::Entry::HTEValues *HV = NM->GetValuesPtr(RM->Mdl);
				if(HV)
				{
					Deleted = HV->Delete(RM);
					if(Deleted)
						Num++;
				}
			}
		} while(Deleted);
		if(Num > 1)
			throw("QuadTree::DeleteVisibleObject#multiple instances of one object found in VisibleObject");
	}
	ParticleEffect *PFX;
	if(PFX = dynamic_cast<ParticleEffect *>(Obj))
	{
		VisiblePFX.Delete(PFX);
	}
}

QTNode *QuadTree::GetNode(IMATH::Vector2DXZ *Pos, int Level)
{
/*	if(Pos->x < Root->BBox.Data[BB_LEFT ]) Pos->x = Root->BBox.Data[BB_LEFT ];
	if(Pos->x > Root->BBox.Data[BB_RIGHT]) Pos->x = Root->BBox.Data[BB_RIGHT];
	if(Pos->z < Root->BBox.Data[BB_FAR  ]) Pos->z = Root->BBox.Data[BB_FAR  ];
	if(Pos->z > Root->BBox.Data[BB_NEAR ]) Pos->z = Root->BBox.Data[BB_NEAR ];*/
	return Root->GetNode(Pos, Level);
}

XZPos QuadTree::GetXZPos(const IMATH::Vector2D &Pos)
{
	return Root->GetNode(&Pos)->GetXZPos(Pos);
}

}
