/***********************************************************/
/*                                                         */
/* QuadTree.h                                              */
/*                                                         */
/* created 16.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

// roam quadtree based on a NURBS surface

#ifndef QUADTREE_H
#define QUADTREE_H

#include "DynamicArray.h"
#include "HashTable.h"
#include "MultiHashTable.h"
#include "D3DStructures.h"
#include "RenderStruct.h"
#include "DataStructures.h"
#include "Terrain.h"
#include "Renderable.h"
#include "RenderableModel.h"
#include "Model.h"
#include "ManagedObject.h"
#include "SpaceManagerInterface.h"
#include "ParticleEffect.h"
#include "Profiler.h"

// www.infinitecode.com 3d library
namespace I3D
{

// in QTNode.h:
class QTNode;
class QTNodeRoot;
class QTNodeTerra;
class QTNodeTerraTex;

// model node
/*class QTRNodeModel
{
	Model *Mdl;
	ISL::DynamicArray<RenderableModel *, 0, 1> Obj;
};

// texture node
class QTRNodeTexture
{
	LPDIRECT3DTEXTURE9 Texture;
	ISL::DynamicArray<QTRNodeModel, 0, 2> Models;
};

// RenderableModel Objects
class QTRObjectArray
{
	ISL::DynamicArray<QTRNodeTexture, 0, 8> Objects;
};*/

struct QuadTreeSpecs
{
	int TexDepth, Depth;
	float Size, Waterlevel; // size of the HeightMap and Render Size
	Terrain *Surface;
	D3DStruct *DS;
};

class XZPos
{
public:
	IMATH::Vector2DXZ Pos;
	int Level;

	XZPos() {}
	XZPos(const IMATH::Vector2DXZ &Position, int Lvl = 0)
		:Pos(Position), Level(Lvl) {}
};

class QuadTree: public Renderable
{
	friend QTNode;
	friend QTNodeRoot;
	friend QTNodeTerra;
	friend QTNodeTerraTex;

	int Flags;
	// root of the tree structure
	QTNode *Root;

	// TODO: replace by mission textures
	LPDIRECT3DTEXTURE9 Texture, DetailTexture;

	// renderable objects
	ISL::DynamicArray<Renderable *, 0, 8> ObjectsRnd;

	// renderable objects with a model (array of visible objects)
	class NodeModel: public ISL::MultiHashTable<I3D::Model *, I3D::RenderableModel *>
	{
	public:
		NodeModel(void) {};

		NodeModel(I3D::RenderableModel *Obj)
		{
			Add(Obj->Mdl, Obj);
		}
	};

	class NodeTexture: public ISL::HashTable<LPDIRECT3DTEXTURE9, NodeModel> {};

	NodeTexture VisibleObjects;

	ISL::DynamicArray<ParticleEffect *, 0, 4> VisiblePFX;
	//QTRObjectArray Objects;
	ITL::Profiler Prf;

protected:
	virtual void CreateView(D3DStruct *DS);

public:

	class CISpaceMgr: public SpaceManagerInterface
	{
		friend QuadTree;

		QuadTree *Tree;

	public:
		virtual void AddManagedObject(ManagedObject *Obj);
		virtual void DeleteManagedObject(ManagedObject *Obj);
		virtual void MoveManagedObject(ManagedObject *Obj);
	} ISpaceMgr;

	friend CISpaceMgr;

	// the properties of this quadtree
	QuadTreeSpecs Specs;

	void SetProfShow(bool Value)
	{
		Prf.SetShow(Value);
	}

	bool GetProfShow(void)
	{
		return Prf.GetShow();
	}

	SpaceManagerInterface *GetSpaceManagerInterface(void)
	{
		return &ISpaceMgr;
	}

	QuadTree(void);
	~QuadTree(void);

	virtual void Render(D3DStruct *DS);
	void Init3D(D3DStruct *DS, bool recurse);
	void Create(QuadTreeSpecs *Specs);
	// add an object to tha tree
	void AddRenderableGlobal(Renderable *Obj);
	// add an object to the object array (array of visible objects)
	void AddVisibleObject(ManagedObject *Obj);
	void DeleteVisibleObject(ManagedObject *Obj);
	void CheckVisibleObjects(int ID);
	QTNode *GetNode(IMATH::Vector2DXZ *Pos, int Level);
	XZPos GetXZPos(const IMATH::Vector2D &Pos);
	void CreateTextures(D3DStruct *DS);
	void AddTexture(const IMATH::Vector2DXZ &Pos, const IMATH::Vector2DXZ &Offset, const IMATH::Vector2DXZ &Size, LPDIRECT3DTEXTURE9 Texture, bool ReconstructMipMaps = true);
	void ReconstructMipMaps();
	LPDIRECT3DTEXTURE9 GetTerrainTexture(const IMATH::Vector &Pos);
	int GetTexturePosX(float x);
	int GetTexturePosZ(float z);
	// call this, when the heightmap is modified
	void UpdateSurface(void);
};

}

#endif
