/***********************************************************/
/*                                                         */
/* RenderStruct.h                                          */
/*                                                         */
/* created 20.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef RENDERSTRUCT_H
#define RENDERSTRUCT_H

#include "DynamicArray.h"
#include "D3DStructures.h"

#include <d3d9.h>
#include <cassert>

// www.infinitecode.com 3d library
namespace I3D
{

template<class VtxType, int FVF, int StartSize = 256, int Increment = 256>
class RenderStruct
{
	LPDIRECT3DVERTEXBUFFER9 VBuf;
	LPDIRECT3DINDEXBUFFER9 IBuf;

public:
	ISL::DynamicArray<VtxType, StartSize, Increment> RenderList;
	ISL::DynamicArray<unsigned short, StartSize, Increment> IndexList;

	RenderStruct(void)
	{
		VBuf = NULL;
		IBuf = NULL;
	}

	~RenderStruct(void)
	{
		RELEASE(VBuf);
		RELEASE(IBuf);
	}

	/*RenderStruct& operator = (RenderStruct& RS)
	{
		VBuf = 0;
		IBuf = 0;
		RenderList = RS.RenderList;
		IndexList = RS.IndexList;
		return *this;
	}*/

	void AddVertex(const VtxType &Vtx)
	{
		RenderList.Add(Vtx);
	}

	const VtxType &GetVertex(int Index)
	{
		return RenderList[Index];
	}

	const VtxType &GetIndex(int Index)
	{
		return IndexList[Index];
	}

	void AddIndex(unsigned short Index)
	{
		IndexList.Add(Index);
	}

	void AddIndices(unsigned short Index1, unsigned short Index2)
	{
		int Count = IndexList.Count();
		IndexList.ReSize(Count + 1);;
		IndexList[Count] = Index1;
		IndexList[Count + 1] = Index2;
	}

	void AddIndices(unsigned short Index1, unsigned short Index2, unsigned short Index3)
	{
		int Count = IndexList.Count();
		IndexList.ReSize(Count + 2);
		IndexList[Count] = Index1;
		IndexList[Count + 1] = Index2;
		IndexList[Count + 2] = Index3;
	}

	void AddIndices(unsigned short Index1, unsigned short Index2, unsigned short Index3, unsigned short Index4)
	{
		int Count = IndexList.Count();
		IndexList.ReSize(Count + 3);
		IndexList[Count] = Index1;
		IndexList[Count + 1] = Index2;
		IndexList[Count + 2] = Index3;
		IndexList[Count + 3] = Index4;
	}

	void ClearVertex(void)
	{
		RenderList.Clear();
	}

	void ClearVertexElements(void)
	{
		RenderList.ClearElements();
	}

	void ClearIndex(void)
	{
		IndexList.Clear();
	}

	void ClearIndexElements(void)
	{
		IndexList.ClearElements();
	}

	int GetVertexNumber(void)
	{
		return RenderList.Count();
	}

	int GetIndexNumber(void)
	{
		return IndexList.Count();
	}

	void CreateVBuf(LPDIRECT3DDEVICE9 D3DDevice)
	{
		if(VBuf) RELEASE(VBuf);
		VBuf = NULL;
		if(RenderList.Count())
		{
			// create vertexbuffer
			D3DDevice->CreateVertexBuffer(RenderList.Count() * 3 * sizeof(VtxType), D3DUSAGE_WRITEONLY, FVF, D3DPOOL_MANAGED, &VBuf, NULL);
			// fill vertexbuffer
			VtxType *Buf = NULL;
			VBuf->Lock(0, 0, (VOID**)&Buf, 0 /*D3DLOCK_NOSYSLOCK*/);
			memcpy(Buf, RenderList.GetPtr(), RenderList.Count() * sizeof(VtxType));
			VBuf->Unlock();
		}
	}

	void CreateIBuf(LPDIRECT3DDEVICE9 D3DDevice)
	{
		if(IBuf) RELEASE(IBuf);
		IBuf = NULL;
		if(IndexList.Count())
		{
			// create vertexbuffer
			D3DDevice->CreateIndexBuffer(IndexList.Count() * 3 * sizeof(unsigned short), D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_MANAGED, &IBuf, NULL);
			// fill indexbuffer
			unsigned short *Buf = NULL;
			IBuf->Lock(0, 0, (VOID**)&Buf, 0 /*D3DLOCK_NOSYSLOCK*/);
			memcpy(Buf, IndexList.GetPtr(), IndexList.Count() * sizeof(unsigned short));
			IBuf->Unlock();
		}
	}

	void SetupVertex(LPDIRECT3DDEVICE9 D3DDevice)
	{
		if(VBuf)
		{
			D3DDevice->SetFVF(FVF);
			D3DDevice->SetStreamSource(0, VBuf, 0, sizeof(VtxType));
		}
	}

	void RenderVertex(LPDIRECT3DDEVICE9 D3DDevice)
	{
		if(VBuf)
		{
			D3DDevice->SetFVF(FVF);
			D3DDevice->SetStreamSource(0, VBuf, 0, sizeof(VtxType));
			D3DDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, GetVertexNumber() / 3);
		}
	}

	void RenderVertex(LPDIRECT3DDEVICE9 D3DDevice, int Offset, int Number)
	{
		if(VBuf)
		{
			D3DDevice->SetFVF(FVF);
			D3DDevice->SetStreamSource(0, VBuf, 0, sizeof(VtxType));
			D3DDevice->DrawPrimitive(D3DPT_TRIANGLELIST, Offset, Number);
		}
	}

	void RenderLineVertex(LPDIRECT3DDEVICE9 D3DDevice)
	{
		if(VBuf)
		{
			D3DDevice->SetFVF(FVF);
			D3DDevice->SetStreamSource(0, VBuf, 0, sizeof(VtxType));
			D3DDevice->DrawPrimitive(D3DPT_LINELIST, 0, GetVertexNumber() / 2);
		}
	}

	void RenderIndex(LPDIRECT3DDEVICE9 D3DDevice)
	{
		if(VBuf && IBuf)
		{
			D3DDevice->SetFVF(FVF);
			D3DDevice->SetStreamSource(0, VBuf, 0, sizeof(VtxType));
			D3DDevice->SetIndices(IBuf);
			D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, GetVertexNumber(), 0, GetIndexNumber() / 3);
		}
	}

	void RenderLineIndex(LPDIRECT3DDEVICE9 D3DDevice)
	{
		if(VBuf && IBuf)
		{
			D3DDevice->SetFVF(FVF);
			D3DDevice->SetStreamSource(0, VBuf, 0, sizeof(VtxType));
			D3DDevice->SetIndices(IBuf);
			D3DDevice->DrawIndexedPrimitive(D3DPT_LINELIST, 0, GetVertexNumber(), 0, GetIndexNumber() / 2);
		}
	}

	void RenderIndex(LPDIRECT3DDEVICE9 D3DDevice, int VertexOffset, int VertexNumber, int IndexOffset, int IndexNumber)
	{
		if(VBuf && IBuf)
		{
			HRESULT Result[4];
			Result[0] = D3DDevice->SetFVF(FVF);
			Result[1] = D3DDevice->SetStreamSource(0, VBuf, 0, sizeof(VtxType));
			Result[2] = D3DDevice->SetIndices(IBuf);
			Result[3] = D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, VertexOffset, 0, VertexNumber, IndexOffset, IndexNumber / 3);
#ifdef DEBUG
			assert(Result[0] == D3D_OK);
			assert(Result[1] == D3D_OK);
			assert(Result[2] == D3D_OK);
			assert(Result[3] == D3D_OK);
#endif
		}
	}

	// just prepare for rendering
	bool SetupIndex(LPDIRECT3DDEVICE9 D3DDevice)
	{
		if(VBuf && IBuf)
		{
			D3DDevice->SetFVF(FVF);
			D3DDevice->SetStreamSource(0, VBuf, 0, sizeof(VtxType));
			D3DDevice->SetIndices(IBuf);
			return true;
		}
		else
			return false;
	}

};

}

#endif
