/***********************************************************/
/*                                                         */
/* Tank.h                                                  */
/*                                                         */
/* created 27.08.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef TANK_H
#define TANK_H

#include "MathStructures.h"
#include "TankLib.h"
#include "ProjectileLib.h"
#include "D3DStruct.h"
#include "SpaceManagerInterface.h"
#include "QuadTree.h"
#include "SpawnInterface.h"
#include "SoundManager.h"
#include "Path.h"
#include "Minimap.h"

// www.infinitecode.com game library
namespace IGAME
{

class Tank: public Minimap::ObjectInterface
{
public:

	class TankSpecs
	{
	public:
		TankSpecs(void): SndMgr(0), Commandable(true) {}

		I3D::D3DStruct *DS;
		IMATH::Vector Pos;
		I3D::SpaceManagerInterface *SM;
		I3D::SpaceManagerInterface *PosMgr;
		I3D::QuadTree *QT;
		SpawnInterface *ISpawn;
		float Angle, MaxSpeed;
		int Owner;
		int ID;
		TankLib *TLib;
		ProjectileLib *PLib;
		char Name[1024];
		IAU::SoundManager *SndMgr;
		bool Commandable, Highlight;
	};

	enum
	{
		TNK_AI_IDLE = 0,
		TNK_AI_MOVE = 1,
		TNK_AI_MOVE_PATH = 2,
		TNK_AI_ATTACK = 3,
		TNK_AI_LAND = 4,
		TNK_AI_UNLOAD = 5,
		TNK_AI_START = 6
	};

private:

	static int TypeID;

	enum
	{
		FLAG_TURRET = 1,
		FLAG_VEHICLE = 2
	};

	#define TNK_MAXSPEED 30.0f
	// vehicle
	#define TNK_MAXDANGLE (PI / 2.0f)
	// turret
	#define TNK_MAXDTANGLE (PI)
	#define TNK_MAXDTPITCH (PI)
	#define TNK_MAXTPITCH (PI / 4.0f)
	#define TNK_MINTPITCH (-PI / 8.0f)
	// shot speed [m/s]
	#define TNK_SHOTSPEED 40.0f

	class CIPosMgr: public I3D::ManagedObject
	{
	public:
		Tank *Base;

		void Create(Tank *Base);

		virtual void *GetBase(void) const;
		virtual void *GetType(void) const;

		virtual void SetPos(const IMATH::Vector &Value)
		{
			Base->RM.SetPos(Value);
		}

		virtual const IMATH::Vector &GetPos(void) const
		{
			return Base->RM.GetPos();
		}
	} IPosMgr;

	friend CIPosMgr;

	I3D::RenderableModel RM, RM2;
	I3D::RenderStruct<I3D::LVertex, I3D_FVFLVERTEX> OverlayRS;
	I3D::RenderStruct<I3D::LVertex, I3D_FVFLVERTEX> HealthRS;
	IMATH::Vector ScreenPos;
	LPDIRECT3DTEXTURE9 SelectionTexture;
	bool UnLinked;
	int Flags;
	bool Moved;
	TankLib::Tank *Lib;
	ProjectileLib::Projectile *LibPrj;
	float LastTrack, LastDust, LastWay, HitQuality, LastSec;
	TankSpecs Specs;
	// AI
	IMATH::Vector TargetPos, TargetPosFinal;
	Tank *Target;
	Tank *TurretTarget;
	// pathfinding
	I3D::Path AIPath;
	int AINextWaypoint;
	IMATH::Vector HomePos;
	int BurstSequence;

	int AIOrder;
	float dAngle, TurretAngle, dTurretAngle, TurretPitch, dTurretPitch;
	float LastShot;
	float Speed;
	float HitPoints;
	// remove from QuadTree
	void UnLink(void);
	void CollisionTest(void);
	void DrawTracks(void);
	void Dust(void);
	void MoveTo(const IMATH::Vector &Target); // use SetTargetPos for orders
	void SetWay(const IMATH::Vector2DXZ &TargetPos);

	void CreateHealthRS(void);

public:
	Tank(void):Target(0), UnLinked(false), Flags(0) {}

	~Tank(void);

	int GetAI() const
	{
		return AIOrder;
	}

	bool GetCommandable() const
	{
		return Specs.Commandable;
	}

	virtual bool GetHighlight() const
	{
		return Specs.Highlight;
	}

	void SetCommandable(bool Value)
	{
		Specs.Commandable = Value;
	}

	void SetHighlight(bool Value)
	{
		Specs.Highlight = Value;
	}

	IMATH::Vector2DXZ GetAITarget() const
	{
		return IMATH::Vector2DXZ(TargetPosFinal.x, TargetPosFinal.z);
	}

	IMATH::Vector2DXZ GetDirectAITarget() const
	{
		return IMATH::Vector2DXZ(TargetPos.x, TargetPos.z);
	}

	const char *GetName(void) const
	{
		return Specs.Name;
	}

	virtual int GetOwner(void) const
	{
		return Specs.Owner;
	}

	void SetOwner(int Value)
	{
		Specs.Owner = Value;
		unsigned int Color;
		switch(Tank::Specs.Owner)
		{
		case 0: Color = 0xFFFFFFFF; break;
		case 1: Color = 0xFFFFFF00; break;
		case 2: Color = 0xFF0000FF; break;
		case 3: Color = 0xFFFF0000; break;
		}
		OverlayRS.AddVertex(I3D::LVertex(IMATH::Vector(-1, -1, 0), Color, IMATH::Vector2D(0, 0)));
		OverlayRS.AddVertex(I3D::LVertex(IMATH::Vector( 1, -1, 0), Color, IMATH::Vector2D(1, 0)));
		OverlayRS.AddVertex(I3D::LVertex(IMATH::Vector( 1,  1, 0), Color, IMATH::Vector2D(1, 1)));
		OverlayRS.AddVertex(I3D::LVertex(IMATH::Vector(-1,  1, 0), Color, IMATH::Vector2D(0, 1)));
		OverlayRS.CreateVBuf(Specs.DS->D3DDevice);
	}

	int GetID(void) const
	{
		return Specs.ID;
	}

	void *GetType(void) const
	{
		return &TypeID;
	}

	static void *GetClassType(void)
	{
		return &TypeID;
	}

	float GetSize() const
	{
		return Lib->Size;
	}

	bool IsVehicle(void) const
	{
		if(UnLinked) throw("Actor::GetPos#Access to unlinked Actor");
		return (Flags & FLAG_VEHICLE) != 0; // != 0 to avoid Visual C++ Warning
	}

	virtual int GetMinimapSize() const
	{
		return IsVehicle() ? 1 : 2;
	}

	void SetPos(const IMATH::Vector &Value)
	{
		Moved = true;
		RM.SetPos(Value);
	}

	float GetAngle()
	{
		return Specs.Angle;
	}

	void SetAngle(float Value)
	{
		Specs.Angle = Value;
	}

	const IMATH::Vector &GetPos(void) const
	{
		if(UnLinked)
			//throw("Actor::GetPos#Access to unlinked Actor");
			throw 1;
		return RM.GetPos();
	}

	virtual IMATH::Vector2DXZ GetXZPos(void) const
	{
		if(UnLinked)
			throw("Actor::GetPos#Access to unlinked Actor");
		return IMATH::Vector2DXZ(RM.GetPos().x, RM.GetPos().z);
	}

	void SetHomePos(const IMATH::Vector &Home)
	{
		HomePos = Home;
	}

	I3D::ManagedObject *GetManagedObjectIf()
	{
		return &IPosMgr;
	}

	virtual void Create(TankSpecs *Specs);
	virtual void RelinkQTree();
	virtual void Physics(I3D::D3DStruct *DS);
	void Init3D(I3D::D3DStruct *DS, bool recurse);
	void UpdateMatrices();
	// AI
	virtual void SetTargetPos(const IMATH::Vector2DXZ &TargetPos);
	virtual void Land(const IMATH::Vector2DXZ &TargetPos);
	virtual void Attack(Tank *Target);
	void NearAttack(Tank *T);
	virtual void Idle(void);
	virtual void ClearTank(Tank *Target);
	virtual void Hit(Tank *Attacker, float Amount);
	void RenderSelection(I3D::D3DStruct *DS);
	bool Select(I3D::BoundingBox2D *BBox);
	virtual bool Select(const I3D::XZPos &Pos);
	bool Select(const IMATH::Vector2D &Pos);
	float SelectDist(const IMATH::Vector2D &Pos);
	void ProjectileHit(bool Hit);
};

}

#endif
