/***********************************************************/
/*                                                         */
/* NURBS.cpp                                               */
/*                                                         */
/* created 21.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

// non uniform rational b-spline

#include "Terrain.h"
#include "TextureManager.h"
#include "Color.h"
#include "IDTerrain.h"
#include "IMath.h"
#include "WaterLevel.h"

// TODO: delete if not needed anymore
#include <stdio.h>

#include <fstream>
using namespace std;

// www.infinitecode.com 3d library
namespace I3D
{

Terrain::Terrain(void)
	: HeightMap(0), VarianceMap(0), AccMap(0)
{
}

Terrain::~Terrain(void)
{
	if(HeightMap)
		delete HeightMap;
	if(HeightDisplacementMap)
		delete HeightDisplacementMap;
	if(VarianceMap)
		delete VarianceMap;
	if(AccMap)
		delete AccMap;
}

float Terrain::GetAcc(float x, float z) const
{
	// TODO: implement b-spline
	// TODO: optimize calculations
	x /= Size * 0.5f;
	z /= Size * 0.5f;
	x += 1.0f;
	z += 1.0f;
	x *= XResAcc * 0.5f;
	z *= ZResAcc * 0.5f;
	if(x < 0.0f) x = 0.0f;
	if(x > XResAcc - 2) x = (float)XResAcc - 2;
	if(z < 0.0f) z = 0.0f;
	if(z > ZResAcc - 2) z = (float)ZResAcc - 2;
	float XFac = x - int(x);
	float ZFac = z - int(z);
	float XInt1 = (1 - XFac) * AccMap[(int)x + (int)z * XResAcc] + XFac * AccMap[(int)x + 1 + (int)z * XResAcc];
	float XInt2 = (1 - XFac) * AccMap[(int)x + ((int)z + 1) * XResAcc] + XFac * AccMap[(int)x + 1 + ((int)z + 1) * XResAcc];
	return ((1 - ZFac) * XInt1 + ZFac * XInt2);
}

float Terrain::GetAccNormed(float x, float z) const
{
	if(GetY(x, z) < WATERLEVEL)
		return 1.0f;
	float Acc = absf(0.04f * GetAcc(x, z));
	if(Acc >= 1.0f || Acc < 0)
		Acc = 1.0f;
	return Acc;
}

bool Terrain::GetTerrainDefined(float x, float z)
{
	for(int i = 0; i < TerrainDscs.Count(); i++)
	{
		//IMATH::Vector2DXZ PosH[3];
		// transform to heightmap coordinates
		IMATH::Vector2DXZ *Pos = TerrainDscs[i].Specs.Pos;
//		TriangleToHMap(Pos, PosH);
		if(TerrainDscs[i].Specs.ID != IDT_TERRAIN_LOW && PointInTriangle((float)x, (float)z, Pos))
			return true;
	}
	return false;
}

float Terrain::GetAccWater(float x, float z) const
{
	if(GetAcc(x, z) == -2.0f)
		return 0.0f;
	else return 1.0f;
}

Vector Terrain::GetNormal(float x, float z) const
{
	float Delta = 3.0f;
	Vector V1 = Vector(x + Delta, GetY(x + Delta, z), z) - Vector(x - Delta, GetY(x - Delta, z), z);
	Vector V2 = Vector(x, GetY(x, z + Delta), z + Delta) - Vector(x, GetY(x, z - Delta), z - Delta);
	Vector CP = -Vector::CrossProduct(V1, V2);
	Vector NCP = Vector::Normalize(CP);
	return NCP;
}

void Terrain::Init(int XRes, int ZRes, int XResAcc, int ZResAcc, float Size)
{
	Terrain::Size = Size;
	OOSize = 1.0f / Size;
	Terrain::XRes = XRes;
	Terrain::ZRes = ZRes;
	Terrain::XResAcc = XResAcc;
	Terrain::ZResAcc = ZResAcc;
	delete HeightMap;
	delete VarianceMap;
	delete AccMap;
	TerrainDscs.Clear();
	Ways.Clear();
	HeightMap = (float *)malloc(XRes * ZRes * sizeof(float));
	HeightDisplacementMap = (float *)malloc(XRes * ZRes * sizeof(float));
	VarianceMap = (float *)malloc(XRes * ZRes * sizeof(float));
	AccMap = (float *)malloc(XResAcc * ZResAcc * sizeof(float));
}

void Terrain::Create(int XRes, int ZRes, int XResAcc, int ZResAcc, float Size)
{
	Init(XRes, ZRes, XResAcc, ZResAcc, Size);
	Terrain::Seed = 0;
	CreateHMap(0);
}

void Terrain::Create(int XRes, int ZRes, int XResAcc, int ZResAcc, float Size, Way **Ways, int WNum, TerrainDsc **NewTerrainDscs, int TNum, int Seed)
{
	Init(XRes, ZRes, XResAcc, ZResAcc, Size);
	for(int i = 0; i < TNum; i++)
		TerrainDscs.Add(*NewTerrainDscs[i]);
	for(int i = 0; i < WNum; i++)
		AddWayNoHMap(Ways[i]);
	Terrain::Seed = Seed;
	CreateHMap(Seed);
}

void Terrain::Create(int XRes, int ZRes, float Size)
{
	Create(XRes, ZRes, XRes / 4, ZRes / 4, Size);
	// load heightmap from file
	/*char fname[1024];
	sprintf(fname, "HeightMap%dx%d.bmp", XRes, ZRes);
	FILE *f = fopen(fname, "rb");
	if(!f)
		throw("Terrain::Create#open heightmap file");
	fseek(f, -XRes*ZRes*3, SEEK_END);
//	for(int j = ZRes - 1; j >= 0; j--)
	for(int j = 0; j < ZRes; j++)
		for(int i = 0; i < XRes; i++)
			HeightMap[i + j * XRes] = (fgetc(f)/256.0f + fgetc(f)/256.0f + fgetc(f)/256.0f)/3.0f * 0.3f * (Size * 0.5f);
	fclose(f);*/
}

void Terrain::CreateHMap(int Seed)
{
	memset(HeightMap, 0, XRes * ZRes * sizeof(float));
	int i;
	const float StdVariance = 0.12f;
	for(i = 0; i < XRes; i++)
		for(int j = 0; j < ZRes; j++)
			VarianceMap[j * ZRes + i] = StdVariance;
	for(i = 0; i < TerrainDscs.Count(); i++)
	{
		switch(TerrainDscs[i].Specs.ID)
		{
		case IDT_TERRAIN_MOUNTAIN:
			SetHeight(TerrainDscs[i].Specs.Pos, 10);
			SetVariance(TerrainDscs[i].Specs.Pos, 0.3f);
			break;
		case IDT_TERRAIN_HIGHMOUNTAIN:
			SetHeight(TerrainDscs[i].Specs.Pos, 25);
			SetVariance(TerrainDscs[i].Specs.Pos, 0.5f);
			break;
		case IDT_TERRAIN_LOW:
			SetHeight(TerrainDscs[i].Specs.Pos, -10);
			SetVariance(TerrainDscs[i].Specs.Pos, StdVariance);
			break;
		case IDT_TERRAIN_ISLANDS:
			SetHeight(TerrainDscs[i].Specs.Pos, WATERLEVEL);
			SetVariance(TerrainDscs[i].Specs.Pos, 0.3f);
			break;
		case IDT_TERRAIN_SMALLISLANDS:
			SetHeight(TerrainDscs[i].Specs.Pos, -30);
			SetVariance(TerrainDscs[i].Specs.Pos, 0.3f);
			break;
		case IDT_TERRAIN_WATER:
			SetHeight(TerrainDscs[i].Specs.Pos, -40);
			SetVariance(TerrainDscs[i].Specs.Pos, 0.25f);
			break;
		case IDT_TERRAIN_DEEPWATER:
			SetHeight(TerrainDscs[i].Specs.Pos, -60);
			SetVariance(TerrainDscs[i].Specs.Pos, 0.25f);
			break;
		}
	}
	// perlin noise
	const int LayerMin = 3;
	const int HDispLayerMin = 5;
	const int LayerMax = 8;
	srand(Seed);
	// maps
	float *Random = new float[XRes * ZRes];
//	float *Layer = new float[XRes * ZRes];
	/*int Div = 1; // number of segments
	for(i = 0; i < LayerMin; i++)
		Div *= 2;
	float RndX = (float)XRes / (float)Div; // size of segments
	float RndZ = (float)ZRes / (float)Div;
	// create first random map
	for(i = 0; i < Div; i++)
		for(int j = 0; j < Div; j++)
			Random[i + j * XRes] = HeightMap[(int)(i * RndX) + (int)(j * RndZ) * XRes];
	// set layer
	for(i = 0; i < XRes; i++)
		for(int j = 0; j < ZRes; j++)
		{
			// interpolate
			float XOfs = (i / RndX) - (int)(i / RndX);
			float ZOfs = (j / RndZ) - (int)(j / RndZ);
			int x1 = (int)(i / RndX);
			int x2 = min((int)(i / RndX) + 1, Div - 1);
			int z1 = (int)(j / RndX);
			int z2 = min((int)(j / RndZ) + 1, Div - 1);
			float Rnd = (1 - ZOfs) * ((1 - XOfs) * Random[x1 + z1 * XRes] + XOfs * Random[x2 + z1 * XRes])
				+ ZOfs * ((1 - XOfs) * Random[x1 + z2 * XRes] + XOfs * Random[x2 + z2 * XRes]);
			// add
			HeightMap[i + j * XRes] = Rnd;
		}*/
	// smooth terraindscs
	for(int m = 0; m < 4; m++)
		for(i = 0; i < XRes; i++)
			for(int j = 0; j < ZRes; j++)
			{
				float Av = 0;
				for(int k = -2; k <= 2; k++)
					for(int l = -2; l <= 2; l++)
						Av += HeightMap[max(min(i + k, XRes - 1), 0) + max(min(j + l, ZRes - 1), 0) * XRes];
				HeightMap[i + j * XRes] = Av / 25.0f;
				Av = 0;
				for(int k = -2; k <= 2; k++)
					for(int l = -2; l <= 2; l++)
						Av += VarianceMap[max(min(i + k, XRes - 1), 0) + max(min(j + l, ZRes - 1), 0) * XRes];
				VarianceMap[i + j * XRes] = Av / 25.0f;
			}
	memset(HeightDisplacementMap, 0, XRes * ZRes * sizeof(HeightDisplacementMap[0]));
	// add perlin noise
	float Factor = 1.0f;
	for(int k = LayerMin + 1; k <= LayerMax; k++)
	{
		// sizes
		int Div = 1; // number of segments
		for(i = 0; i < k; i++)
			Div *= 2;
		float RndX = (float)XRes / Div; // size of segments
		float RndZ = (float)ZRes / Div;
		// create random map
		for(i = 0; i < Div; i++)
			for(int j = 0; j < Div; j++)
				Random[i + XRes * j] = (float)((rand() % 128) - 64);
		// add layer
		for(i = 0; i < XRes; i++)
			for(int j = 0; j < ZRes; j++)
			{
				// interpolate
				float XOfs = (i / RndX) - (int)(i / RndX);
				float ZOfs = (j / RndZ) - (int)(j / RndZ);
				int x1 = (int)(i / RndX);
				int x2 = min((int)(i / RndX) + 1, Div - 1);
				int z1 = (int)(j / RndX);
				int z2 = min((int)(j / RndZ) + 1, Div - 1);
				float Rnd = (1 - ZOfs) * ((1 - XOfs) * Random[x1 + z1 * XRes] + XOfs * Random[x2 + z1 * XRes])
					+ ZOfs * ((1 - XOfs) * Random[x1 + z2 * XRes] + XOfs * Random[x2 + z2 * XRes]);
				// add
				HeightMap[i + j * XRes] += VarianceMap[i + j * XRes] * Factor * Rnd;
				if(k >= HDispLayerMin + 1)
					HeightDisplacementMap[i + j * XRes] += Factor * Rnd;
			}
		Factor *= 0.5f;
	}
//	delete[] Layer;
	delete[] Random;
	// flatten ways
	for(i = 0; i < Ways.Count(); i++)
	{
		IMATH::Vector2DXZ Delta = Ways[i].Specs.Pos2 - Ways[i].Specs.Pos1;
		IMATH::Vector2DXZ NDelta = Delta.Normalize();
		for(float j = 0; j <= Delta.Length(); j += 5)
			AddPlain(Ways[i].Specs.Pos1 + IMATH::Vector2DXZ(j * NDelta.x, j * NDelta.z), 2);
	}
	OnHeightMapChange();
}

// also creates WaterMap
void Terrain::OnHeightMapChange(void)
{
	const float impassable = -1;
	const float water = -2;
	//ofstream o("_debug.txt", ios::app);
	for(int x = 0; x < XResAcc; x++)
		for(int z = 0; z < ZResAcc; z++)
		{
			AccMap[x + XResAcc * z] = 0.0f;
			// accessibility
			for(int i = 0; i < TerrainDscs.Count(); i++)
			{
				switch(TerrainDscs[i].Specs.ID)
				{
				//case IDT_TERRAIN_MOUNTAIN:
				case IDT_TERRAIN_HIGHMOUNTAIN:
					if(PointInTriangle(AMapToWorldX((float)x), AMapToWorldZ((float)z), TerrainDscs[i].GetTriangle()))
					{
						AccMap[x + XResAcc * z] = impassable;
						//o << x << ',' << z << endl;
					}
					break;
				case IDT_TERRAIN_WATER:
					if(PointInTriangle(AMapToWorldX((float)x), AMapToWorldZ((float)z), TerrainDscs[i].GetTriangle()))
					{
						AccMap[x + XResAcc * z] = water;
						//o << x << ',' << z << endl;
					}
					break;
				}
			}
			if(AccMap[x + XResAcc * z] >= 0)
			{
				// gradient
				float y = GetY(((float)x * 2 / XResAcc - 1) * 0.5f * Size, ((float)z * 2 / ZResAcc - 1) * 0.5f * Size);
				float ydx = GetY((((float)x + 1) * 2 / XResAcc - 1) * 0.5f * Size, ((float)z * 2 / ZResAcc - 1) * 0.5f * Size);
				float ydz = GetY(((float)x * 2 / XResAcc - 1) * 0.5f * Size, (((float)z + 1) * 2 / ZResAcc - 1) * 0.5f * Size);
				AccMap[x + XResAcc * z] = (absf(ydx - y) + absf(ydz - y));
			}
		}
	// make ways passable
	int i;
	for(i = 0; i < Ways.Count(); i++)
	{
		IMATH::Vector2DXZ DeltaWorld = Ways[i].Specs.Pos2 - Ways[i].Specs.Pos1;
		IMATH::Vector2DXZ Delta = IMATH::Vector2DXZ(WorldToAMapX(DeltaWorld.x), WorldToAMapZ(DeltaWorld.z));
		IMATH::Vector2DXZ NDelta = Delta.Normalize();
		for(float j = 0; j <= Delta.Length(); j += 0.5f)
		{
			int x = (int)(WorldToAMapX(Ways[i].Specs.Pos1.x) + j * NDelta.x);
			int z = (int)(WorldToAMapZ(Ways[i].Specs.Pos1.z) + j * NDelta.z);
			if(x < 0) x = 0;
			if(x >= XResAcc) x = XResAcc - 1;
			if(z < 0) z = 0;
			if(z >= ZResAcc) z = ZResAcc - 1;
			AccMap[x + XResAcc * z] = 0;
		}
	}

	for(i = 0; i < 8; i++)
		for(int j = 0; j < 8; j++)
			WaterMap[i][j] = false;
	for(i = 0; i < XRes; i++)
		for(int j = 0; j < ZRes; j++)
			if(HeightMap[i + XRes * j] <= WATERLEVEL)
				WaterMap[i * 8 / XRes][j * 8 / ZRes] = true;
}

float Terrain::TreePropability(const IMATH::Vector2DXZ &Pos)
{
	float Factor = 0.5f;
	if(GetY(Pos.x, Pos.z) < WATERLEVEL)
		Factor = 0.0f;
	else
		for(int i = 0; i < TerrainDscs.Count(); i++)
		{
			if(PointInTriangle(Pos.x, Pos.z, TerrainDscs[i].GetTriangle()))
			{
				switch(TerrainDscs[i].Specs.ID)
				{
				case IDT_TERRAIN_HIGHMOUNTAIN:
				case IDT_TERRAIN_WATER:
					Factor = 0.0f;
					break;
				case IDT_TERRAIN_MOUNTAIN:
					Factor = 0.2f;
					break;
				case IDT_TERRAIN_LOW:
					Factor = 1.0f;
					break;
				}
			}
		}
	return Factor;
}

bool Terrain::GetWater(int x, int z)
{
	return WaterMap[x][z];
}

void Terrain::AddPlain(const IMATH::Vector2DXZ &Pos, float Size)
{
	float x = Pos.x;
	float z = Pos.z;
	x /= Terrain::Size * 0.5f;
	z /= Terrain::Size * 0.5f;
	x += 1.0f;
	z += 1.0f;
	x *= XRes * 0.5f;
	z *= ZRes * 0.5f;
	for(int dx = (int)-Size; dx <= (int)Size; ++dx)
		for(int dz = (int)-Size; dz <= (int)Size; ++dz)
			if((int)x + dx >= 0 && (int)x + dx < XRes && (int)z + dz >= 0 && (int)z + dz < ZRes)
			{
				float Av = 0;
				const int AvRadius = 3;
				for(int i = -AvRadius; i <= AvRadius; ++i)
					for(int j = -AvRadius; j <= AvRadius; ++j)
					{
						int hx = (int)x + dx + i;
						int hz = (int)z + dz + j;
						if(hx < 0)
							hx = 0;
						if(hx > XRes - 1)
							hx = XRes - 1;
						if(hz < 0)
							hz = 0;
						if(hz > ZRes - 1)
							hz = ZRes - 1;
						Av += HeightMap[hx + hz * XRes];
					}
				HeightMap[(int)x + dx + ((int)z + dz) * XRes] = Av / ((1 + 2 * AvRadius) * (1 + 2 * AvRadius));
			}
}

void Terrain::AddWayNoHMap(const Way *NewWay)
{
	Ways.Add(*NewWay);
}

void Terrain::AddWay(const Way *NewWay)
{
	AddWayNoHMap(NewWay);
	CreateHMap(Seed);
	//OnHeightMapChange();
}

Terrain::WayDistanceInfo Terrain::GetWayDistance(const Way &W, const IMATH::Vector2DXZ &Pos)
{
	WayDistanceInfo Info;
	IMATH::Vector2DXZ A = W.GetPos(0);
	IMATH::Vector2DXZ B = W.GetPos(1);
	IMATH::Vector2DXZ C = Pos;
	IMATH::Vector2DXZ BmA = B - A;
	IMATH::Vector2DXZ BmAn = BmA.Normalize();
	IMATH::Vector2DXZ O(-BmAn.z, BmAn.x);
	O = O.Normalize();
	//float t = ((A.z + (C.x - A.x) / BmAn.x * BmAn.z - C.z) / O.z) / (1 - O.x / BmAn.x * BmAn.z / O.z);
	float t = ((A.z - C.z) * BmAn.x + (C.x - A.x) * BmAn.z) / (O.z * BmAn.x - O.x * BmAn.z);
	Info.SideDistance = t;
	IMATH::Vector2DXZ D = C + t * O.Normalize();
	float s = (D - A).Length();
	float s2 = (C.x + t * O.x - A.x) / BmA.x;
	// dirty hack
	if((s < 0 || s > 1) && (s2 >= 0 && s2 <= 1))
		s = s2;
	Info.LengthCoefficient = s;
	Info.W = &W;
	return Info;
}

void Terrain::SplitWay(const IMATH::Vector2DXZ &Pos)
{
	for(int i = 0; i < Ways.Count(); i++)
	{
		WayDistanceInfo Info = GetWayDistance(Ways[i], Pos);
		Way::WaySpecs WS;
		WS.ID = Ways[i].GetID();
		strcpy(WS.Name, Ways[i].GetName());
		WS.Pos1 = Ways[i].GetPos(0) + Info.LengthCoefficient * (Ways[i].GetPos(1) - Ways[i].GetPos(0));
		WS.Pos2 = Ways[i].GetPos(1);
		Way *W = new Way(&WS);
		AddWay(W);
		Ways[i].SetPos(1, WS.Pos1);
	}
}

float Terrain::WorldToHMapX(float Pos)
{
	return max(0, min(((Pos * OOSize) + 0.5f) * XRes, XRes - 1));
}

float Terrain::WorldToHMapZ(float Pos)
{
	return max(0, min(((Pos * OOSize) + 0.5f) * ZRes, ZRes - 1));
}

float Terrain::HMapToWorldX(float Pos)
{
	return (Pos / XRes - 0.5f) * Size;
}

float Terrain::HMapToWorldZ(float Pos)
{
	return (Pos / ZRes - 0.5f) * Size;
}

float Terrain::AMapToWorldX(float Pos)
{
	return (Pos / XResAcc - 0.5f) * Size;
}

float Terrain::AMapToWorldZ(float Pos)
{
	return (Pos / ZResAcc - 0.5f) * Size;
}

void Terrain::TriangleToHMap(const IMATH::Vector2DXZ *Pos, IMATH::Vector2DXZ *PosH)
{
	for(int i = 0; i < 3; i++)
	{
		PosH[i].x = max(WorldToHMapX(Pos[i].x), 0);
		PosH[i].z = max(WorldToHMapZ(Pos[i].z), 0);
	}
}

void Terrain::TriangleToBBox(const IMATH::Vector2DXZ *PosH, I3D::BoundingBox2D *BB)
{
	BB->Left = BB->Right = PosH->x;
	BB->Near = BB->Far = PosH->z;
	for(int i = 0; i < 3; i++)
	{
		if(PosH[i].x < BB->Left)
			BB->Left = PosH[i].x;
		if(PosH[i].x > BB->Right)
			BB->Right = PosH[i].x;
		if(PosH[i].z < BB->Near)
			BB->Near = PosH[i].z;
		if(PosH[i].z > BB->Far)
			BB->Far = PosH[i].z;
	}
}

bool Terrain::PointInTriangle(float x, float z, const IMATH::Vector2DXZ *PosH)
{
	/*int Crossed = 0;
	// point in polygon test
	// get orientation
	IMATH::Vector2DXZ V1 = PosH[1] - PosH[0];
	IMATH::Vector2DXZ V2 = PosH[2] - PosH[0];
	IMATH::Vector V1_3D = IMATH::Vector(V1.x, 0, V1.z);
	IMATH::Vector V2_3D = IMATH::Vector(V2.x, 0, V2.z);
	//bool Orientation = IMATH::Vector::CrossProduct(V1_3D, V2_3D).y > 0;
	for(int i = 0; i < 3; i++)
	{
		//float dx1 = (PosH[i].x - PosH[(i + 1) % 3].x) * (PosH[(i + 1) % 3].z - z) / (PosH[(i + 1) % 3].z - PosH[i].z);
		float ratio = (z - PosH[(i + 1) % 3].z) / (PosH[i].z - PosH[(i + 1) % 3].z);
		float dx1 = (PosH[i].x - PosH[(i + 1) % 3].x) * ratio;
		dx1 += PosH[(i + 1) % 3].x - x;
		if(dx1 <= 0)
			Crossed++;
	}
	return (Crossed % 2 == 0);//(Orientation?1:0));*/
	float ax = PosH[1].x - PosH[2].x;
	float az = PosH[1].z - PosH[2].z;
	float bx = PosH[0].x - PosH[2].x;
	float bz = PosH[0].z - PosH[2].z;
/*	float px = x - PosH[2].x;
	float pz = z - PosH[2].z;*/
	float co = bx * az - bz * ax;
	//float co = (PosH[0].x - PosH[2].x) * (PosH[1].z - PosH[2].z) - (PosH[0].z - PosH[2].z) * (PosH[1].x - PosH[2].x);
	if(!co)
		return false;
	float px = x - PosH[2].x;
	float pz = z - PosH[2].z;
	float ca = (pz * bx - px * bz) / co;
	float cb = (px * az - pz * ax) / co;
	return ca >= 0 && cb >= 0 && ca + cb <= 1;
}

void Terrain::SetHeight(const IMATH::Vector2DXZ *Pos, float Value)
{
	IMATH::Vector2DXZ PosH[3];
	// transform to heightmap coordinates
	TriangleToHMap(Pos, PosH);
	// get bounding box
	I3D::BoundingBox2D BB;
 	TriangleToBBox(PosH, &BB);
	for(int x = (int)BB.Left; x <= (int)BB.Right; x++)
		for(int z = (int)BB.Near; z <= (int)BB.Far; z++)
		{
			if(PointInTriangle((float)x, (float)z, PosH))
			{
				// point is inside the polygon
				if(x >= 0 && x < XRes && z >= 0 && z < ZRes)
				{
					HeightMap[x + z * XRes] = Value;
				}
			}
		}
}

void Terrain::SetVariance(const IMATH::Vector2DXZ *Pos, float Value)
{
	IMATH::Vector2DXZ PosH[3];
	// transform to heightmap coordinates
	TriangleToHMap(Pos, PosH);
	// get bounding box
	I3D::BoundingBox2D BB;
 	TriangleToBBox(PosH, &BB);
	for(int x = (int)BB.Left; x <= (int)BB.Right; x++)
		for(int z = (int)BB.Near; z <= (int)BB.Far; z++)
		{
			if(PointInTriangle((float)x, (float)z, PosH))
			{
				// point is inside the polygon
				if(x >= 0 && x < XRes && z >= 0 && z < ZRes)
				{
					VarianceMap[x + z * XRes] = Value;
				}
			}
		}
}

void Terrain::AddTerrainDsc(TerrainDsc *NewTerrainDsc)
{
	TerrainDscs.Add(*NewTerrainDsc);
	CreateHMap(Seed);
}

void Terrain::AddTerrainDscNoHMap(TerrainDsc *NewTerrainDsc)
{
	TerrainDscs.Add(*NewTerrainDsc);
}

void Terrain::AddTerrainDscs(TerrainDsc **NewTerrainDscs, int Num)
{
	for(int i = 0; i < Num; i++)
		TerrainDscs.Add(*NewTerrainDscs[i]);
	CreateHMap(Seed);
}

ISL::DynamicArray<TerrainDsc> *Terrain::GetTerrainDscs()
{
	return &TerrainDscs;
}

ISL::DynamicArray<Way> *Terrain::GetWays()
{
	return &Ways;
}

float Terrain::Limit01(float x)
{
	if(x < 0)
		return 0;
	if(x > 1)
		return 1;
	return x;
}

LPDIRECT3DTEXTURE9 Terrain::GetTexture(I3D::D3DStruct *DS, float Left, float Near, float Right, float Far)
{
	int State = 0;
	try
	{
		State = 1;
		LPDIRECT3DTEXTURE9 Texture;
		if(DS->D3DDevice->CreateTexture(256, 256, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &Texture, NULL) != D3D_OK)
			throw("Terrain::GetTexture: Couldn't create texture");
		State = 2;
		// lock texture
		D3DLOCKED_RECT LRect;
		memset(&LRect, 0, sizeof(LRect));
		if(Texture->LockRect(0, &LRect, NULL, 0) != D3D_OK)
			throw("Terrain::GetTexture can't lock texture 1");
		// buffers
		unsigned int *Buf = (unsigned int *)LRect.pBits;
		unsigned int *Map = (unsigned int *)malloc(256 * 256 * 4);
		// parts
		enum
		{
			PART_NUM = 5,

			WATER = 0,
			DESERT = 1,
			BEACH = 2,
			ROCK = 3,
			GRASS = 4
		};
		State = 3;
		// ground texture
		LPDIRECT3DTEXTURE9 GroundTex[PART_NUM];
		GroundTex[0] = TextureManager::GetTexture(DS, "Textures\\WaterGround.bmp");
		GroundTex[1] = TextureManager::GetTexture(DS, "Textures\\Desert.bmp");
		GroundTex[2] = TextureManager::GetTexture(DS, "Textures\\Beach.bmp");
		GroundTex[3] = TextureManager::GetTexture(DS, "Textures\\Rock.bmp");
		GroundTex[4] = TextureManager::GetTexture(DS, "Textures\\Grass.bmp");
		State = 4;
		D3DSURFACE_DESC GDesc[PART_NUM];
		D3DLOCKED_RECT GRect[PART_NUM];
		unsigned int *GroundMap[PART_NUM];
		int i;
		for(i = 0; i < PART_NUM; ++i)
		{
			if(GroundTex[i]->GetLevelDesc(0, &GDesc[i]) != D3D_OK)
				throw("Terrain::GetTexture: Can't get description of Texture level 0");
			memset(&GRect[i], 0, sizeof(GRect[i]));
			if(GroundTex[i]->LockRect(0, &GRect[i], NULL, 0) != D3D_OK)
				throw("Terrain::GetTexture can't lock texture 2");
			GroundMap[i] = (unsigned int *)GRect[i].pBits;
		}
		// paint
		srand(0);
		State = 5;
		/*const int FOURIER_DEPTH = 200;
		float CoeffX[FOURIER_DEPTH], CoeffZ[FOURIER_DEPTH];
		for(i = 0; i < FOURIER_DEPTH; ++i)
		{
			CoeffX[i] = 0.01f * (rand() % 101 - 50) / (i + 1);
			CoeffZ[i] = 0.01f * (rand() % 101 - 50) / (i + 1);
		}*/
		// lock texture
		for(i = 0; i < 256; i++)
			for(int j = 0; j < 256; j++)
			{
				int k;
				struct 
				{
					float P; // percentual part
					int r, g, b, a;
				} Parts[PART_NUM];

				State = 6;
				float TerrX = Left + (Right - Left) * i / 256.0f;
				float TerrZ = Far + (Near - Far) * j / 256.0f;
				float Height = GetY(TerrX, TerrZ);
				/*for(k = 0; k < FOURIER_DEPTH; ++k)
				{
					Height += 50 * CoeffX[k] * sin((k + 1) * TerrX * 2.0f * PI / Size);
					Height += 50 * CoeffZ[k] * sin((k + 1) * TerrZ * 2.0f * PI / Size);
				}*/
				int x = min((int)WorldToHMapX(TerrX), XRes - 2);
				int z = min((int)WorldToHMapZ(TerrZ), ZRes - 2);
				float fx = WorldToHMapX(TerrX) - (int)WorldToHMapX(TerrX);
				float fz = WorldToHMapZ(TerrZ) - (int)WorldToHMapZ(TerrZ);
				Height += (rand() % 101 - 50) / 20.0f +  0.2f * 
					((1 - fz) * ((1 - fx) * HeightDisplacementMap[z * XRes + x]
					+ fx * HeightDisplacementMap[z * XRes + x + 1])
					+ fz * ((1 - fx) * HeightDisplacementMap[(z + 1) * XRes + x]
					+ fx * HeightDisplacementMap[(z + 1) * XRes + x + 1]));
				int Bright = rand() % 0x40;
				Parts[WATER].P = Limit01(((WATERLEVEL - 10) - Height) / 25);
				/*Parts[WATER].r = 0x60;
				Parts[WATER].g = 0x80 + Bright/3;
				Parts[WATER].b = 0xA0 + Bright/2;
				Parts[WATER].a = 0xff;*/
				Parts[WATER].r = (GroundMap[0][((i % GDesc[0].Width) + GDesc[0].Width * (j % GDesc[0].Height))] >> 16) & 0xff;
				Parts[WATER].g = (GroundMap[0][((i % GDesc[0].Width) + GDesc[0].Width * (j % GDesc[0].Height))] >> 8) & 0xff;
				Parts[WATER].b = GroundMap[0][((i % GDesc[0].Width) + GDesc[0].Width * (j % GDesc[0].Height))] & 0xff;
				Parts[WATER].a = (GroundMap[0][((i % GDesc[0].Width) + GDesc[0].Width * (j % GDesc[0].Height))] >> 24) & 0xff;
				Parts[DESERT].P = Limit01((Height - (WATERLEVEL + 10)) / 20) * Limit01(((WATERLEVEL + 65) - Height) / 30);
				/*Parts[DESERT].r = 0xA0 + Bright;
				Parts[DESERT].g = (0xA0 + Bright) * 7 / 8;
				Parts[DESERT].b = 0x20;
				Parts[DESERT].a = 0xff;*/
				Parts[DESERT].r = (GroundMap[1][((i % GDesc[1].Width) + GDesc[1].Width * (j % GDesc[1].Height))] >> 16) & 0xff;
				Parts[DESERT].g = (GroundMap[1][((i % GDesc[1].Width) + GDesc[1].Width * (j % GDesc[1].Height))] >> 8) & 0xff;
				Parts[DESERT].b = GroundMap[1][((i % GDesc[1].Width) + GDesc[1].Width * (j % GDesc[1].Height))] & 0xff;
				Parts[DESERT].a = (GroundMap[1][((i % GDesc[1].Width) + GDesc[1].Width * (j % GDesc[1].Height))] >> 24) & 0xff;
				Parts[BEACH].P = Limit01((Height - (WATERLEVEL - 25)) / 25) * Limit01(((WATERLEVEL + 30) - Height) / 20);//Limit01(1 - absf((Height - (WATERLEVEL + 10)) / 20));
				/*Parts[BEACH].r = 0xC0 + Bright / 2;
				Parts[BEACH].g = 0xC0 + Bright / 2;
				Parts[BEACH].b = 0xA0;
				Parts[BEACH].a = 0xff;*/
				Parts[BEACH].r = (GroundMap[2][((i % GDesc[2].Width) + GDesc[2].Width * (j % GDesc[2].Height))] >> 16) & 0xff;
				Parts[BEACH].g = (GroundMap[2][((i % GDesc[2].Width) + GDesc[2].Width * (j % GDesc[2].Height))] >> 8) & 0xff;
				Parts[BEACH].b = GroundMap[2][((i % GDesc[2].Width) + GDesc[2].Width * (j % GDesc[2].Height))] & 0xff;
				Parts[BEACH].a = (GroundMap[2][((i % GDesc[2].Width) + GDesc[2].Width * (j % GDesc[2].Height))] >> 24) & 0xff;
				Parts[ROCK].P = Limit01((Height - (WATERLEVEL + 30)) / 40);
				/*Parts[ROCK].r = 0xB0 + Bright;
				Parts[ROCK].g = 0xB0 + Bright;
				Parts[ROCK].b = 0xB0 + Bright;
				Parts[ROCK].a = 0xff;*/
				Parts[ROCK].r = (GroundMap[3][((i % GDesc[3].Width) + GDesc[3].Width * (j % GDesc[3].Height))] >> 16) & 0xff;
				Parts[ROCK].g = (GroundMap[3][((i % GDesc[3].Width) + GDesc[3].Width * (j % GDesc[3].Height))] >> 8) & 0xff;
				Parts[ROCK].b = GroundMap[3][((i % GDesc[3].Width) + GDesc[3].Width * (j % GDesc[3].Height))] & 0xff;
				Parts[ROCK].a = (GroundMap[3][((i % GDesc[3].Width) + GDesc[3].Width * (j % GDesc[3].Height))] >> 24) & 0xff;
				Parts[GRASS].P = Limit01((Height - (WATERLEVEL + 3)) / 15) * Limit01(((WATERLEVEL + 15) - Height) / 20);
				Parts[GRASS].r = (GroundMap[4][((i % GDesc[4].Width) + GDesc[4].Width * (j % GDesc[4].Height))] >> 16) & 0xff;
				Parts[GRASS].g = (GroundMap[4][((i % GDesc[4].Width) + GDesc[4].Width * (j % GDesc[4].Height))] >> 8) & 0xff;
				Parts[GRASS].b = GroundMap[4][((i % GDesc[4].Width) + GDesc[4].Width * (j % GDesc[4].Height))] & 0xff;
				Parts[GRASS].a = (GroundMap[4][((i % GDesc[4].Width) + GDesc[4].Width * (j % GDesc[4].Height))] >> 24) & 0xff;
				float PTotal = 0;
				for(k = 0; k < PART_NUM; k++)
					PTotal += Parts[k].P;
				float a = 0, r = 0, g = 0, b = 0;
				for(k = 0; k < PART_NUM; k++)
				{
					a += Parts[k].a * Parts[k].P / PTotal;
					r += Parts[k].r * Parts[k].P / PTotal;
					g += Parts[k].g * Parts[k].P / PTotal;
					b += Parts[k].b * Parts[k].P / PTotal;
				}
				int ia = (int)a;
				int ir = (int)r;
				int ig = (int)g;
				int ib = (int)b;
				unsigned int Color = (ia << 24) + (ir << 16) + (ig << 8) + ib;
				Map[i + j * 256] = Color;
				State = 7;
			}
		State = 8;
		for(i = 0; i < Ways.Count(); i++)
		{
			State = 9;
			//IMATH::Vector2DXZ Delta = Ways[i].Specs.Pos2 - Ways[i].Specs.Pos1;
			//IMATH::Vector2DXZ NDelta = Delta.Normalize();
			IMATH::Vector2DXZ Point1(256.0f * (Ways[i].Specs.Pos1.x - Left) / (Right - Left), 256.0f * (Ways[i].Specs.Pos1.z - Near) / (Far - Near));
			IMATH::Vector2DXZ Point2(256.0f * (Ways[i].Specs.Pos2.x - Left) / (Right - Left), 256.0f * (Ways[i].Specs.Pos2.z - Near) / (Far - Near));
			State = 10;
			int Size = max((int)(Ways[i].GetSize() * (256.0f / (Right - Left))), 1);
			State = 100;
			State = 1000000 + i;
			Way *W = &Ways[i];
			State = 105;
			int ID = W->GetID();
			State = 106;
			if(ID == 0)
			{
				State = 101;
				for(int j = 0; j < 2 * Size; j++)
				{
					State = 11;
					int dx = rand() % Size - Size / 2;
					int dz = rand() % Size - Size / 2;
					State = 200;
					int dx2 = rand() % max(Size / 2, 1) - (Size / 4);
					State = 201;
					int dz2 = rand() % max(Size / 2, 1) - (Size / 4);
					int dx3 = rand() % max(Size / 2, 1) - (Size / 4);
					int dz3 = rand() % max(Size / 2, 1) - (Size / 4);
					State = 202;
					IMATH::Vector2DXZ Point1Disp = Point1 + IMATH::Vector2DXZ((float)dx2, (float)dz2);
					IMATH::Vector2DXZ Point2Disp = Point2 + IMATH::Vector2DXZ((float)dx3, (float)dz3);
					IMATH::Vector2DXZ Delta = Point2Disp - Point1Disp;
					IMATH::Vector2DXZ NDelta = Delta.Normalize();
					IMATH::Vector2DXZ Spot = Point1Disp + IMATH::Vector2DXZ((float)dx, (float)dz);
					const float Increment = 0.25f;
					State = 12;
					for(float l = 0; l <= Delta.Length(); l += Increment)
					{
						Spot += Increment * IMATH::Vector2DXZ(NDelta.x, NDelta.z);
						float z = 255 - Spot.z;
						if(Spot.x >= 0 && Spot.x < 256 && z >= 0 && z < 256)
							Map[(int)Spot.x + (int)z * 256] = Color::Brightness(Map[(int)Spot.x + (int)z * 256], -4);
					}
					State = 13;
				}
			}
			else
			{
				State = 15;
				float Step = 0.7f;
				IMATH::Vector2DXZ Delta = Point2 - Point1;
				IMATH::Vector2DXZ NDelta = Delta.Normalize();
				for(float j = (float)-Size; j <= Size; j += Step)
				{
					State = 16;
					IMATH::Vector2DXZ Disp = IMATH::Vector2DXZ(Delta.Normalize().z, -Delta.Normalize().x) * j;
					IMATH::Vector2DXZ Spot = Point1 + Disp - (float)Size * NDelta;
					const float Increment = 0.7f;
					State = 17;
					for(float l = (float)-Size; l <= Delta.Length() + Size; l += Increment)
					{
						Spot += Increment * NDelta;
						float Blend2;
						if(l >= 0 && l <= Delta.Length())
							Blend2 = 1.0f - min(0.5f * min(absf(j - -Size), absf(j - Size)), 1.0f);
						else
							Blend2 = 1.0f - max(min(0.5f * max((Size - (Point1 - Spot).Length()), (Size - (Point2 - Spot).Length())), 1.0f), 0.0f);
						float z = 255 - Spot.z;
						if(Spot.x >= 0 && Spot.x < 256 && z >= 0 && z < 256)
							Map[(int)Spot.x + (int)z * 256] = Color::Blend(Map[(int)Spot.x + (int)z * 256], 0xAA808080, Blend2);
					}
				}
				const float Increment = 0.7f;
				float StripeWidth = 1.0f;
				for(float j = -StripeWidth; j <= StripeWidth; j += Step)
				{
					float Blend = 1.0f - min(0.5f * min(absf(j - -StripeWidth), absf(j - StripeWidth)), 1.0f);
					IMATH::Vector2DXZ Disp = IMATH::Vector2DXZ(Delta.Normalize().z, -Delta.Normalize().x) * j;
					IMATH::Vector2DXZ Spot = Point1 + Disp + (float)Size * NDelta;
					State = 18;
					if(Ways[i].GetID() == IDT_WAY_STREET)
						for(float l = (float)Size; l <= Delta.Length() - Size; l += Increment)
						{
							Spot += Increment * NDelta;
							float z = 255 - Spot.z;
							if((((int)l / 10) % 3 == 0) && Spot.x >= 0 && Spot.x < 256 && z >= 0 && z < 256)
								Map[(int)Spot.x + (int)z * 256] = Color::Blend(Map[(int)Spot.x + (int)z * 256], 0xAAFFFFFF, Blend);
						}
					State = 19;
					if(Ways[i].GetID() == IDT_WAY_HIGHWAY)
					{
						// line
						IMATH::Vector2DXZ Spot = Point1 + Disp + (float)Size * NDelta;
						for(float l = (float)Size; l <= Delta.Length() - Size; l += Increment)
						{
							Spot += Increment * NDelta;
							float z = 255 - Spot.z;
							if(Spot.x >= 0 && Spot.x < 256 && z >= 0 && z < 256)
								Map[(int)Spot.x + (int)z * 256] = Color::Blend(Map[(int)Spot.x + (int)z * 256], 0xAAFFFFFF, Blend);
						}
						// stripes
						Spot = Point1 + Disp + (float)Size * NDelta + ((float)Size / 2) * IMATH::Vector2DXZ(Delta.Normalize().z, -Delta.Normalize().x);
						for(float l = (float)Size; l <= Delta.Length() - Size; l += Increment)
						{
							Spot += Increment * NDelta;
							float z = 255 - Spot.z;
							if((((int)l / 10) % 3 == 0) && Spot.x >= 0 && Spot.x < 256 && z >= 0 && z < 256)
								Map[(int)Spot.x + (int)z * 256] = Color::Blend(Map[(int)Spot.x + (int)z * 256], 0xAAFFFFFF, Blend);
						}
						Spot = Point1 + Disp + (float)Size * NDelta - ((float)Size / 2) * IMATH::Vector2DXZ(Delta.Normalize().z, -Delta.Normalize().x);
						for(float l = (float)Size; l <= Delta.Length() - Size; l += Increment)
						{
							Spot += Increment * NDelta;
							float z = 255 - Spot.z;
							if((((int)l / 10) % 3 == 0) && Spot.x >= 0 && Spot.x < 256 && z >= 0 && z < 256)
								Map[(int)Spot.x + (int)z * 256] = Color::Blend(Map[(int)Spot.x + (int)z * 256], 0xAAFFFFFF, Blend);
						}
					}
					State = 20;
				}
				State = 21;
			}
			State = 22;
			//IMATH::Vector2DXZ Pos = Ways[i].Specs.Pos1 + IMATH::Vector2DXZ(l * NDelta.x, l * NDelta.z);
		}
		State = 23;
		// copy
		memcpy(Buf, Map, 256 * 256 * 4);
		delete Map;
		State = 24;
		// unlock
		for(i = 0; i < PART_NUM; ++i)
			if(GroundTex[i]->UnlockRect(0) != D3D_OK)
				throw("Terrain::GetTexture can't unlock texture");
		if(Texture->UnlockRect(0) != D3D_OK)
			throw("Terrain::GetTexture can't unlock texture");
		State = 25;
		// mipmaps
		TextureManager::CreateMipMaps(Texture, 0);
		State = 26;
		return Texture;
	}
	catch(char *s)
	{
		throw(s);
	}
	catch(...)
	{
		char Str[1024];
		sprintf(Str, "Terrain::GetTexture#%i", State);
		throw(Str);
	}
}

void Terrain::WorldToAMap(IMATH::Vector2DXZ *Dst, const IMATH::Vector2DXZ *Src)
{
	Dst->x = ((Src->x * OOSize) + 0.5f) * XResAcc;
	Dst->z = ((Src->z * OOSize) + 0.5f) * ZResAcc;
}

float Terrain::WorldToAMapX(float x)
{
	return ((x * OOSize) + 0.5f) * XResAcc;
}

float Terrain::WorldToAMapZ(float z)
{
	return ((z * OOSize) + 0.5f) * ZResAcc;
}

// pathfinding classes
class PrePath
{
public:
	class Waypoint
	{
	public:
		enum Direction
		{
			NORTH = 1,
			NORTHEAST = 2,
			EAST = 3,
			SOUTHEAST = 4,
			SOUTH = 5,
			SOUTHWEST = 6,
			WEST = 7,
			NORTHWEST = 8,
			ALL = 9
		};
		int x, z, Dir;

		void Set(int x, int z, int Dir)
		{
			Waypoint::x = x;
			Waypoint::z = z;
			Waypoint::Dir = Dir;
		}
	};

private:
	Waypoint *Way;

	int WaypointNum;
	float DistanceStraight, DistanceEdge, DistanceAcc;
	int TargetX, TargetZ;
	float TargetXf, TargetZf, StartXf, StartZf;
	Terrain *Terr;

public:

	PrePath(Terrain *Terr, int Size)
	{
		PrePath::Terr = Terr;
		WaypointNum = Size;
		DistanceStraight = 0;
		DistanceEdge = 0;
		DistanceAcc = 0;
		Way = new Waypoint[WaypointNum];
	}

	PrePath(PrePath *P, int x, int z, int Dir)
	{
		Terr = P->Terr;
		TargetX = P->TargetX;
		TargetZ = P->TargetZ;
		TargetXf = P->TargetXf;
		TargetZf = P->TargetZf;
		StartXf = P->StartXf;
		StartZf = P->StartZf;
		WaypointNum = P->WaypointNum + 1;
		Way = new Waypoint[WaypointNum];
		for(int i = 0; i < WaypointNum - 1; i++)
			Way[i] = P->Way[i];
		Way[WaypointNum - 1].x = x;
		Way[WaypointNum - 1].z = z;
		Way[WaypointNum - 1].Dir = Dir;
		// fitness
		DistanceStraight = P->DistanceStraight;
		DistanceEdge = P->DistanceEdge;
		if(WaypointNum >= 2 && !Finished())
		{
			if(P->Way[P->WaypointNum - 1].Dir == Dir)
				DistanceStraight += ((Dir % 2)?1:(float)sqrt(2.0));
			else
				DistanceEdge += ((Dir % 2)?1:(float)sqrt(2.0));
		}
		DistanceAcc = P->DistanceAcc + Terr->GetAccNormed(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z)));
	}

	~PrePath(void)
	{
		delete[] Way;
	}

/*	void Set(int i, int x, int z, int Dir)
	{
		Way[i].x = x;
		Way[i].z = z;
		Way[i].Dir = Dir;
	}*/

	float Pyth(float x1, float z1, float x2, float z2)
	{
		return (float)sqrt((x2 - x1) * (x2 - x1) + (z2 - z1) * (z2 - z1));
	}

	void SetStart(float x, float z)
	{
		Way[0].x = (int)x;
		Way[0].z = (int)z;
		Way[0].Dir = Waypoint::ALL;
		StartXf = x;
		StartZf = z;
	}

	void SetTarget(float x, float z)
	{
		TargetX = (int)x;
		TargetZ = (int)z;
		TargetXf = x;
		TargetZf = z;
	}

	float GetFitness(void)
	{
		float Fitness = 0;
		//Fitness += 0.9f * DistanceStraight + DistanceEdge + 2 * DistanceAcc;
		// DistanceEdge: changed direction
		// DistanceStraight: same direction
		// DistanceAcc: terrain gradient (additive)
		Fitness += DistanceStraight + 1.1f * DistanceEdge + 0.5f * DistanceAcc;
		if(WaypointNum > 1)
		{
			Fitness += Pyth(StartXf, StartZf, (float)Way[1].x, (float)Way[1].z);
			//Fitness += 10 * Pyth((float)Way[WaypointNum - 2].x, (float)Way[WaypointNum - 2].z, TargetXf, TargetZf);
			Fitness += 2.0f * Pyth((float)Way[WaypointNum - 2].x, (float)Way[WaypointNum - 2].z, TargetXf, TargetZf);
		}
		return Fitness;
	}

	bool Finished(void)
	{
		return Way[WaypointNum - 1].x == TargetX && Way[WaypointNum - 1].z == TargetZ;
	}

	void GetPath(Path *P)
	{
		for(int i = 1; i < WaypointNum - 1; i++)
			P->Add(new IMATH::Vector2DXZ(Terr->AMapToWorldX((float)Way[i].x), Terr->AMapToWorldZ((float)Way[i].z)));
	}

	int GetNext(Waypoint **NextWPs, bool Water)
	{
		int it = 0;
		float (Terrain::*GetAcc)(float, float) const = Water ? &Terrain::GetAccWater : &Terrain::GetAccNormed;
		switch(Way[WaypointNum - 1].Dir)
		{
		case Waypoint::NORTH:
			(*NextWPs) = new Waypoint[3];
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z + 1, Waypoint::NORTHWEST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x    )), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x    , Way[WaypointNum - 1].z + 1, Waypoint::NORTH);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z + 1, Waypoint::NORTHEAST);
			return it;
		case Waypoint::NORTHEAST:
			(*NextWPs) = new Waypoint[3];
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x    )), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x    , Way[WaypointNum - 1].z + 1, Waypoint::NORTH);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z + 1, Waypoint::NORTHEAST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z   ))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z    , Waypoint::EAST);
			return it;
		case Waypoint::EAST:
			(*NextWPs) = new Waypoint[3];
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z + 1, Waypoint::NORTHEAST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z    ))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z    , Waypoint::EAST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z - 1, Waypoint::SOUTHEAST);
			return it;
		case Waypoint::SOUTHEAST:
			(*NextWPs) = new Waypoint[3];
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z    ))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z    , Waypoint::EAST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z - 1, Waypoint::SOUTHEAST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x    )), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x    , Way[WaypointNum - 1].z - 1, Waypoint::SOUTH);
			return it;
		case Waypoint::SOUTH:
			(*NextWPs) = new Waypoint[3];
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z - 1, Waypoint::SOUTHEAST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x   )), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x    , Way[WaypointNum - 1].z - 1, Waypoint::SOUTH);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z - 1, Waypoint::SOUTHWEST);
			return it;
		case Waypoint::SOUTHWEST:
			(*NextWPs) = new Waypoint[3];
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x    )), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x    , Way[WaypointNum - 1].z - 1, Waypoint::SOUTH);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z - 1, Waypoint::SOUTHWEST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z   ))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z    , Waypoint::WEST);
			return it;
		case Waypoint::WEST:
			(*NextWPs) = new Waypoint[3];
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z - 1, Waypoint::SOUTHWEST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z    ))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z    , Waypoint::WEST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z + 1, Waypoint::NORTHWEST);
			return it;
		case Waypoint::NORTHWEST:
			(*NextWPs) = new Waypoint[3];
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z    ))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z    , Waypoint::WEST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z + 1, Waypoint::NORTHWEST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x    )), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x    , Way[WaypointNum - 1].z + 1, Waypoint::NORTH);
			return it;
		//case Waypoint::ALL:
		default:
			*NextWPs = new Waypoint[8];
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x    )), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x    , Way[WaypointNum - 1].z + 1, Waypoint::NORTH);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z + 1, Waypoint::NORTHEAST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z    ))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z    , Waypoint::EAST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x + 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x + 1, Way[WaypointNum - 1].z - 1, Waypoint::SOUTHEAST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x    )), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x    , Way[WaypointNum - 1].z - 1, Waypoint::SOUTH);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z - 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z - 1, Waypoint::SOUTHWEST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z    ))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z    , Waypoint::WEST);
			if((Terr->*GetAcc)(Terr->AMapToWorldX((float)(Way[WaypointNum - 1].x - 1)), Terr->AMapToWorldZ((float)(Way[WaypointNum - 1].z + 1))) < 1)
				(*NextWPs)[it++].Set(Way[WaypointNum - 1].x - 1, Way[WaypointNum - 1].z + 1, Waypoint::NORTHWEST);
			return it;
		}
	}
};

void Terrain::GetPath(const IMATH::Vector2DXZ *From, const IMATH::Vector2DXZ *To, Path *P, bool Water)
{
	// From, To
	IMATH::Vector2DXZ FromA;
	IMATH::Vector2DXZ ToA;
	WorldToAMap(&FromA, From);
	WorldToAMap(&ToA, To);
	// init Path
	P->Clear();
	P->Add(new IMATH::Vector2DXZ(*From));
	P->Add(new IMATH::Vector2DXZ(*To));
	return;
	// init PrePaths
	ISL::DynamicArray<PrePath *> PrePaths;
	ISL::DynamicArray<PrePath *> CompletePrePaths;
	PrePath *StartPath = new PrePath(this, 1);
	StartPath->SetStart(FromA.x, FromA.z);
	StartPath->SetTarget(ToA.x, ToA.z);
	PrePaths.Add(StartPath);
	bool Finished = PrePaths.Last()->Finished();
	//ofstream path("Path.txt");
	//path << "From: " << FromA.x << ',' << FromA.z << endl;
	//path << "To  : " << ToA.x << ',' << ToA.z << endl;
	int Paths = 0;
	if(!PrePaths[0]->Finished())
		while(Paths < 5)
		{
			int i;
			// get best waypoint
			PrePath::Waypoint *NextWPs;
			PrePath *Next = PrePaths[0];
			float Fitness = Next->GetFitness();
			float BestNewFitness = -1;
			for(i = 0; i < PrePaths.Count(); i++)
				if(PrePaths[i]->GetFitness() < Fitness)
				{
					Next = PrePaths[i];
					Fitness = Next->GetFitness();
				}
			//path << "best: " << Fitness << std::endl;
			// iterate, add new waypoints
			int Num = Next->GetNext(&NextWPs, Water);
			if(!Num)
			{
				//path << "no way found" << std::endl;
				//break;
			}
			for(i = 0; i < Num; i++)
			{
				PrePath *NewPath = new PrePath(Next, NextWPs[i].x, NextWPs[i].z, NextWPs[i].Dir);
				//Finished = Finished || NewPath->Finished();
				if(NewPath->Finished())
				{
					Paths++;
					CompletePrePaths.Add(NewPath);
					//path << "added successful prepath" << std::endl;
					Finished = true;
				}
				else
				{
					PrePaths.Add(NewPath);
				}
				if(NewPath->GetFitness() < BestNewFitness || BestNewFitness == -1)
					BestNewFitness = NewPath->GetFitness();
				//path << NextWPs[i].x << ", " << NextWPs[i].x << ", " << NewPath->GetFitness() << std::endl;
			}
			//path.flush();
			delete[] NextWPs;
			//if(BestNewFitness < Fitness && BestNewFitness != -1)
			PrePaths.DeleteKeepOrder(Next);
			if(!PrePaths.Count())
			{
				//path << "all prepaths deleted" << std::endl;
				break;
			}
			if(PrePaths.Count() > 200)
			{
				//path << "> 200 prepaths" << std::endl;
				break;
			}
		}
	else
	{
		CompletePrePaths.Add(PrePaths[0]);
		PrePaths.Delete(0);
	}
	//path << "Way:" << endl;
	if(Finished)
	{
		//PrePaths.Last()->GetPath(P);
		float BestFitness = CompletePrePaths[0]->GetFitness();
		PrePath *BestPath = CompletePrePaths[0];
		for(int i = 0; i < CompletePrePaths.Count(); ++i)
		{
			if((CompletePrePaths[i]->GetFitness() < BestFitness) && CompletePrePaths[i]->Finished())
			{
				BestFitness = CompletePrePaths[i]->GetFitness();
				BestPath = CompletePrePaths[i];
			}
		}
		BestPath->GetPath(P);
		// add target point
		if((!Water && GetAccNormed(To->x, To->z) < 1) || (Water && GetAccWater(To->x, To->z) < 1))
			P->Add(new IMATH::Vector2DXZ(*To));
	}
	else
	{
		// get best (incomplete) path so far
		PrePath *Sel = PrePaths[0];
		float Fitness = Sel->GetFitness();
		for(int i = 0; i < PrePaths.Count(); i++)
			if(PrePaths[i]->GetFitness() < Fitness)
			{
				Sel = PrePaths[i];
				Fitness = Sel->GetFitness();
			}
		Sel->GetPath(P);
	}
	/*for(int i = 0; i < P->CountWaypoints(); i++)
		path << P->GetWaypoint(i).x << ',' << P->GetWaypoint(i).z << endl;*/
	PrePaths.ClearPtr();
	CompletePrePaths.ClearPtr();
}

int Terrain::GetResX(void)
{
	return XRes;
}

int Terrain::GetResZ(void)
{
	return ZRes;
}

int Terrain::GetAccResX(void)
{
	return XResAcc;
}

int Terrain::GetAccResZ(void)
{
	return ZResAcc;
}

float Terrain::GetSize(void)
{
	return Size;
}

}
