/***********************************************************/
/*                                                         */
/* Terrain.h                                               */
/*                                                         */
/* created 21.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef TERRAIN_H
#define TERRAIN_H

#include "D3DStructures.h"
#include "Way.h"
#include "TerrainDsc.h"
#include "D3DStruct.h"
#include "DynamicArray.h"
#include "TerrainInterface.h"
#include "DataStructures.h"
#include "Path.h"

// www.infinitecode.com 3d library
namespace I3D
{

class Terrain: public TerrainInterface
{
public:
	struct WayDistanceInfo
	{
		float SideDistance, LengthCoefficient;
		const Way *W;
	};

private:
	int XRes, ZRes, XResAcc, ZResAcc, Seed;
	float Size, OOSize;
	// TODO: implement dHeightMap
	float *HeightMap, *HeightDisplacementMap, *VarianceMap, *AccMap;
	bool WaterMap[8][8];

	ISL::DynamicArray<Way> Ways;
	ISL::DynamicArray<TerrainDsc> TerrainDscs;

	void WorldToAMap(IMATH::Vector2DXZ *Dst, const IMATH::Vector2DXZ *Src);
	void TriangleToHMap(const IMATH::Vector2DXZ *Pos, IMATH::Vector2DXZ *PosH);
	void TriangleToBBox(const IMATH::Vector2DXZ *PosH, I3D::BoundingBox2D *BB);
	bool PointInTriangle(float x, float z, const IMATH::Vector2DXZ *PosH);
	void SetHeight(const IMATH::Vector2DXZ *Pos, float Value);
	void SetVariance(const IMATH::Vector2DXZ *Pos, float Value);
	void OnHeightMapChange(void); // also creates WaterMap
	float Limit01(float x);
	void Init(int XRes, int ZRes, int XResAcc, int ZResAcc, float Size);

public:
	Terrain(void);
	~Terrain(void);

	// TerrainInterface

	virtual float GetY(float x, float z) const
	{
		// TODO: implement b-spline
		// TODO: optimize calculations
		/*x *= OOSize;
		z *= OOSize;
		x += 0.5f;
		z += 0.5f;
		x *= XRes;
		z *= ZRes;*/
		x = (x * OOSize + 0.5f) * XRes;
		z = (z * OOSize + 0.5f) * ZRes;
		if(x < 0.0f) x = 0.0f;
		if(x > XRes - 2) x = (float)(XRes - 2);
		if(z < 0.0f) z = 0.0f;
		if(z > ZRes - 2) z = (float)(ZRes - 2);
		const int ix = (int)x;
		const int iz = (int)z;
		const float XFac = x - ix;
		const float ZFac = z - iz;
		const float OMXFac = 1.0f - XFac;
		const float *HeightMapRel = HeightMap + ix + iz * XRes;
		return
			(
				(1.0f - ZFac)
				* (
					OMXFac * *HeightMapRel
					+ XFac * HeightMapRel[1]
				)
				+ ZFac
				* (
					OMXFac * HeightMapRel[XRes]
					+ XFac * HeightMapRel[1 + XRes]
				)
			);
	}

	virtual float GetAcc(float x, float z) const;
	virtual float GetAccNormed(float x, float z) const;
	virtual bool GetTerrainDefined(float x, float z);
	virtual float GetAccWater(float x, float z) const;
	virtual Vector GetNormal(float x, float z) const;
	virtual int GetResX(void);
	virtual int GetResZ(void);
	virtual int GetAccResX(void);
	virtual int GetAccResZ(void);
	virtual float GetSize(void);
	virtual bool GetWater(int x, int z);

	float WorldToHMapX(float Pos);
	float WorldToHMapZ(float Pos);
	float WorldToAMapX(float x);
	float WorldToAMapZ(float z);
	float HMapToWorldX(float Pos);
	float HMapToWorldZ(float Pos);
	float AMapToWorldX(float Pos);
	float AMapToWorldZ(float Pos);

	void Create(int XRes, int ZRes, float Size);
	void Create(int XRes, int ZRes, int XResAcc, int ZResAcc, float Size);
	void Create(int XRes, int ZRes, int XResAcc, int ZResAcc, float Size, Way **Ways, int WNum, TerrainDsc **NewTerrainDscs, int TNum, int Seed = 0);
	void AddWay(const Way *NewWay);
	void AddWayNoHMap(const Way *NewWay);
	void AddTerrainDsc(TerrainDsc *NewTerrainDsc);
	void AddTerrainDscNoHMap(TerrainDsc *NewTerrainDsc);
	void AddTerrainDscs(TerrainDsc **NewTerrainDscs, int Num);
	void AddPlain(const IMATH::Vector2DXZ &Pos, float Size);
	LPDIRECT3DTEXTURE9 GetTexture(I3D::D3DStruct *DS, float Left, float Near, float Right, float Far);
	void GetPath(const IMATH::Vector2DXZ *From, const IMATH::Vector2DXZ *To, Path *P, bool Water);
	ISL::DynamicArray<TerrainDsc> *GetTerrainDscs();
	ISL::DynamicArray<Way> *GetWays();
	float TreePropability(const IMATH::Vector2DXZ &Pos);
	void CreateHMap(int Seed);
	WayDistanceInfo GetWayDistance(const Way &W, const IMATH::Vector2DXZ &Pos);
	void SplitWay(const IMATH::Vector2DXZ &Pos);
};

}

#endif
