/***********************************************************/
/*                                                         */
/* TextureLoader.cpp                                       */
/*                                                         */
/* created 21.08.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#include "TextureManager.h"

// www.infinitecode.com 3d library
namespace I3D
{

ITL::IniLoader TextureManager::Ini = ITL::IniLoader("Textures\\Textures.ini");
ISL::DynamicArray<TextureManager::LoadedTexture, 0, 1> TextureManager::LoadedTextures;

void TextureManager::BilinearOptimization(unsigned int *Map, int Width, int Height)
{
	bool Change;
	int iterations = 0;
	do
	{
		Change = false;
		// for all pixels
		for(int i = 0; i < Width; i++)
			for(int j = 0; j < Height; j++)
				// if the pixel is colorkeyed (alpha == 0) and black
				if(Map[j * Width + i] >> 24 == 0 && Map[j * Width + i] & 0xFFFFFF)
				{
					Change = true;
					iterations++;
					// initialise color and number of averaged pixels
					int r = 0, b = 0, g = 0;
					int num = 0;
					// for all neighbors
					for(int k = -1; k <= 1; k++)
						for(int l = -1; l <= 1; l++)
							// clipping
							if(i + k >= 0 && i + k < Width && j + l >= 0 && j + l < Height)
							{
								// if not colorkeyed
								if(Map[(j + l) * Width + i + k] >> 24 != 0)
								{
									// store the color
									num++;
									r += (Map[(j + l) * Width + i + k] >> 16) & 0xFF;
									g += (Map[(j + l) * Width + i + k] >> 8) & 0xFF;
									b += Map[(j + l) * Width + i + k] & 0xFF;
								}
							}
					// if a color has been found
					if(num)
					{
						// calculate average
						r /= num;
						g /= num;
						b /= num;
						// paint the pixel
						Map[j * Width + i] = 0x10000 * r + 0x100 * g + b;
					}
				}
	} while(Change && iterations < 10000000);
}

void TextureManager::CreateMipMaps(LPDIRECT3DTEXTURE9 Texture, int Flags)
{
	int i;
	// get Texture width and height
	D3DSURFACE_DESC LDesc;
	if(Texture->GetLevelDesc(0, &LDesc) != D3D_OK)
		throw TextureManagerException("CreateMipMaps: Can't get description of Texture level 0");
	// create temporary Texture map
	unsigned int *Map = (unsigned int *)malloc(LDesc.Width * LDesc.Height * sizeof(unsigned int));
	// get the Texture data
	D3DLOCKED_RECT LRect;
	memset(&LRect, 0, sizeof(LRect));
	if(Texture->LockRect( 0, &LRect, NULL, 0 )!=D3D_OK)
		throw TextureManagerException("CreateMipMaps: Can't lock Texture");
	unsigned int *TexMap = (unsigned int*)LRect.pBits;
	if(LRect.pBits == NULL)
		throw TextureManagerException("CreateMipMaps: Texture lock returned nullpointer");
	// copy Texture data into temporary map
	for(i = 0; i < (int)LDesc.Width; i++)
	{
		for(int j = 0; j < (int)LDesc.Height; j++)
		{
			Map[((LDesc.Height - 1) - j) * LDesc.Width + i] = TexMap[((LDesc.Height - 1) - j) * LDesc.Width + i];
		}
	}
	if(Texture->UnlockRect(0) != D3D_OK)
		throw TextureManagerException("CreateMipMaps: Can't unlock Texture");
	// create mipmap levels
	int Res = LDesc.Width;
	for(i = 1; i < (int)Texture->GetLevelCount(); i++)
	{
		Res /= 2;
		memset(&LRect, 0, sizeof(LRect));
		if(Texture->LockRect(i, &LRect, NULL, 0) != D3D_OK)
			throw TextureManagerException("CreateMipMaps: Can't lock texture mipmap level");
		unsigned int *TexMap = (unsigned int*)LRect.pBits;
		if(LRect.pBits == NULL)
			throw TextureManagerException("CreateMipMaps: Textureture mipmap level lock returned nullpointer");
		int j, k;
		// for all pixels
		for(j = 0; j < Res; j++)
		{
			for(k = 0; k < Res; k++)
			{
				int num = 0;
				int a = 0, r = 0, g = 0, b = 0;
				// for the four pixels of the lower mipmap
				for(int l = 0; l < 2; l++)
					for(int m = 0; m < 2; m++)
					{
						int MapAlpha = Map[(2 * j + m) * LDesc.Width + 2 * k + l] >> 24;
						if(!(Flags & TM_FLAG_MIPALPHAIGN) || MapAlpha)
						{
							// add pixel color
							num++;
							a += MapAlpha;
							r += (Map[(2 * j + m) * LDesc.Width + 2 * k + l] >> 16) & 0xFF;
							g += (Map[(2 * j + m) * LDesc.Width + 2 * k + l] >> 8) & 0xFF;
							b += Map[(2 * j + m) * LDesc.Width + 2 * k + l] & 0xFF;
						}
					}
				if(num)
				{
					// calculate average
					a /= num;
					r /= num;
					g /= num;
					b /= num;
					if(Flags & TM_FLAG_MIPALPHA60 && a && a < 0x60)
						a = 0x60;
					if(Flags & TM_FLAG_GRAYMIPMAPS && i > 3)
					{
						r = (int)((r - 128.0f) * 0.5f + 128.0f);
						g = (int)((g - 128.0f) * 0.5f + 128.0f);
						b = (int)((b - 128.0f) * 0.5f + 128.0f);
					}
				}
				Map[j * LDesc.Width + k] = a * 0x1000000 + r * 0x10000 + g * 0x100 + b;
			}
		}
		// copy the mipmap
		for(j = 0; j < Res; j++)
		{
			for(k = 0; k < Res; k++)
			{
				TexMap[j * Res + k] = Map[j * LDesc.Width + k];
			}
		}
		if(Texture->UnlockRect(i)!=D3D_OK)
			throw TextureManagerException("CreateMipMaps: Can't unlock texture mipmap level");
	}
	delete Map;
}

LPDIRECT3DTEXTURE9 TextureManager::Load(LPDIRECT3DDEVICE9 D3DDevice, const char *FileName, const char *AlphaFileName, int Flags)
{
/*		char FName[1024];
	strcpy(FName, "Textures\\");
	strcat(FName, FileName);*/
	if(strstr(FileName, ".bmp") != FileName + strlen(FileName) - 4)
	{
		char Str[1024];
		sprintf(Str, "TextureManager::Load: Unsupported file format: %s", FileName);
		throw(Str);
	}
	char Str[1024];
	LPDIRECT3DTEXTURE9 Texture = NULL;
	FILE *f = fopen(FileName, "rb");
	if(!f)
	{
		char Str[1024];
		sprintf(Str, "TextureManager::Load: Couldn't open texture file: %s", FileName);
		throw(Str);
	}
	fseek(f, 18, SEEK_SET);
	int Width, Height, Bit;
	fread(&Width, 4, 1, f);
	fread(&Height, 4, 1, f);
	fseek(f, 28, SEEK_SET);
	fread(&Bit, 4, 1, f);
	FILE *g;
	int ABit;
	bool Alpha = AlphaFileName && *AlphaFileName;
	// autoload alpha map
	char *AutoAlpha = 0;
	if(!Alpha)
	{
		AutoAlpha = new char[strlen(FileName) + 1 + sizeof("_alpha")];
		strcpy(AutoAlpha, FileName);
		char *it = AutoAlpha;
		while(*it != '.' && *it)
			++it;
		strcpy(it, "_alpha.bmp");
		AlphaFileName = AutoAlpha;
		Alpha = true;
	}
	if(Alpha)
	{
		// alpha map
		g = fopen(AlphaFileName, "rb");
		if(!g)
		{
			if(AutoAlpha)
				Alpha = false;
			else
			{
				char Str[1024];
				sprintf(Str, "TextureManager::Load: Couldn't open alpha texture file: %s", AlphaFileName);
				delete[] AutoAlpha;
				throw(Str);
			}
		}
		else
		{
			/*if(AutoAlpha)
			{
				int foo = 0;
			}*/
			fseek(g, 18, SEEK_SET);
			int AWidth, AHeight;
			fread(&AWidth, 4, 1, g);
			fread(&AHeight, 4, 1, g);
			fseek(g, 28, SEEK_SET);
			fread(&ABit, 4, 1, g);
			if(AWidth != Width || AHeight != Height || (ABit != 24 && ABit != 8))
			{
				char Str[1024];
				sprintf(Str, "TextureManager::Load: Alpha map has different height/width/bpp: %s", AlphaFileName);
				delete[] AutoAlpha;
				throw(Str);
			}
		}
	}

	D3DDevice->CreateTexture(Width, Height, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &Texture, NULL);
	if(!Texture)
	{
		sprintf(Str, "TextureLoader::Load(\"%s\") D3DDevice->CreateTexture returned NULL. Wrong file format or not enough graphics memory? Width:%d, Height:%d", FileName, Width, Height);
		throw TextureManagerException(Str);
	}
	D3DLOCKED_RECT LRect;
	memset(&LRect, 0, sizeof(LRect));
	if(Texture->LockRect( 0, &LRect, NULL, 0 )!=D3D_OK)
	{
		sprintf(Str, "TextureLoader::Load(\"%s\") can't lock texture", FileName);
		delete[] AutoAlpha;
		throw TextureManagerException(Str);
	}
	unsigned int *Buf = (unsigned int*)LRect.pBits;
	unsigned int *Map = (unsigned int *)malloc(Width * Height * sizeof(unsigned int));
	
	if(!Buf)
	{
		delete[] AutoAlpha;
		throw TextureManagerException("TextureManager::Load: texture lock returned nullpointer");
	}
	switch(Bit)
	{
	case 24:
		{
			fseek(f, 0xa, SEEK_SET);
			int Offset;
			fread(&Offset, 1, 4, f);
			fseek(f, Offset, SEEK_SET);
			if(Alpha)
			{
				fseek(g, 0xa, SEEK_SET);
				fread(&Offset, 1, 4, g);
				fseek(g, Offset, SEEK_SET);
			}
			// bmp's are BGR and reverse-y, textures are RGB, so we sort everything here
			for(int j = Height - 1; j >= 0; j--)
				for(int i = 0; i < Width; i++)
				{
					int a = 0xFF;
					if((Flags & TM_FLAG_MASK) == TM_FLAG_ALPHA00)
						a = 0x00;
					if((Flags & TM_FLAG_MASK) == TM_FLAG_ALPHA20)
						a = 0x20;
					if((Flags & TM_FLAG_MASK) == TM_FLAG_ALPHA40)
						a = 0x40;
					if((Flags & TM_FLAG_MASK) == TM_FLAG_ALPHA60)
						a = 0x60;
					if((Flags & TM_FLAG_MASK) == TM_FLAG_ALPHA80)
						a = 0x80;
					if((Flags & TM_FLAG_MASK) == TM_FLAG_ALPHAC0)
						a = 0xC0;
					int b = fgetc(f);
					int gr = fgetc(f);
					int r = fgetc(f);
					if(Flags & TM_FLAG_ALPHACC0 && r + b + gr == 0)
						a = 0;
					if(Flags & TM_FLAG_ALPHACCFF00FF && r == 0xFF && gr == 0 && b == 0xFF)
						a = 0;
					if(Alpha)
					{
						switch(ABit)
						{
						case 8:
							a = fgetc(g);
							break;
						case 24:
							int b = fgetc(g);
							int gr = fgetc(g);
							int r = fgetc(g);
							a = (r + b + gr) / 3;
							break;
						}
					}
					Map[j * Width + i] = 0x1000000 * a + 0x10000 * r + 0x100 * gr + b;
				}
			break;
		}
	case 8:
		// TODO: implement this
		break;
	default:
		char Str[1024];
		sprintf(Str, "TextureLoader::Load(\"%s\") invalid bitrate: %d", FileName, Bit);
		RELEASE(Texture);
		delete[] AutoAlpha;
		throw TextureManagerException(Str);
		exit(-1);
		break;
	}
	if(Flags & TM_FLAG_ACCBILOPT) BilinearOptimization(Map, Width, Height);
	memcpy(Buf, Map, Width * Height * 4);
	delete Map;
	if(Texture->UnlockRect(0) != D3D_OK)
	{
		sprintf(Str, "TextureLoader::Load(\"%s\") can't unlock texture", FileName);
		throw TextureManagerException(Str);
	}
	if(Alpha)
		fclose(g);
	fclose(f);
	CreateMipMaps(Texture, Flags);
	delete[] AutoAlpha;
	return Texture;
}

LPDIRECT3DTEXTURE9 TextureManager::GetTexture(D3DStruct *DS, const char *Name, const char *AlphaName, int Flags)
{
	// if already loaded, return texture handle
	for(int i = 0; i < LoadedTextures.Count(); i++)
		if(!strcmpi(Name, LoadedTextures[i].Name))
			return LoadedTextures[i].Texture;
	char FileName[1024], AlphaFileName[1024];
	// filename
	const char *IniFileName = Ini.GetValue(Name, "File");
	if(IniFileName)
		strcpy(FileName, IniFileName);
	else
		strcpy(FileName, Name);
	// alpha map filename
	const char *IniAlphaFileName = Ini.GetValue(Name, "AlphaFile");
	if(IniAlphaFileName)
	{
		strcpy(AlphaFileName, "Textures\\");
		strcat(FileName, IniAlphaFileName);
	}
	else if(AlphaName && *AlphaName)
		strcpy(AlphaFileName, AlphaName);
	else
		*AlphaFileName = 0;
	// flags
	// alpha color key 0
	const char *AlphaColorKeyZero = Ini.GetValue(Name, "AlphaColorKeyZero");
	if(AlphaColorKeyZero && !strcmpi(AlphaColorKeyZero, "true"))
		Flags |= TM_FLAG_ALPHACC0;
	// bilinear optimizations for alpha color key
	const char *AlphaCCBilinearOptimize = Ini.GetValue(Name, "AlphaCCBilinearOptimize");
	if(AlphaCCBilinearOptimize && !strcmpi(AlphaCCBilinearOptimize, "true"))
		Flags |= TM_FLAG_ACCBILOPT;
	// alpha = 0 ignore for mipmap creation
	const char *MipMapIgnoreAlpha0 = Ini.GetValue(Name, "MipMapIgnoreAlpha0");
	if(MipMapIgnoreAlpha0 && !strcmpi(MipMapIgnoreAlpha0, "true"))
		Flags |= TM_FLAG_MIPALPHAIGN;
	// keep nonzero alpha >= 60
	const char *MipMapAlpha60 = Ini.GetValue(Name, "MipMapAlpha60");
	if(MipMapAlpha60 && !strcmpi(MipMapAlpha60, "true"))
		Flags |= TM_FLAG_MIPALPHA60;
	const char *GrayMipMaps = Ini.GetValue(Name, "GrayMipMaps");
	if(GrayMipMaps && !strcmpi(GrayMipMaps, "true"))
		Flags |= TM_FLAG_GRAYMIPMAPS;
	LPDIRECT3DTEXTURE9 Texture;
	if(*AlphaFileName)
		Texture = Load(DS->D3DDevice, FileName, AlphaFileName, Flags);
	else
		Texture = Load(DS->D3DDevice, FileName, NULL, Flags);
	LoadedTextures[LoadedTextures.Count()] = LoadedTexture(Name, Texture);
	return Texture;
}

void TextureManager::UpdateMipmaps(LPDIRECT3DTEXTURE9 Texture, int x, int z, I3D::Color::Color_int4 Blend)
{
	D3DSURFACE_DESC LDesc;
	if(Texture->GetLevelDesc(0, &LDesc) != D3D_OK)
		throw("TextureManager::UpdateMipmaps: Can't get description of Texture level 0");
	int Levels = (int)Texture->GetLevelCount();
	for(int i = 1; i < Levels; i++)
	{
		// recalculate
		x /= 2;
		z /= 2;
		LDesc.Width /= 2;
		LDesc.Height /= 2;
		// lock
		D3DLOCKED_RECT LRect;
		memset(&LRect, 0, sizeof(LRect));
		if(Texture->LockRect(i, &LRect, NULL, 0 ) != D3D_OK)
			throw("Tank::DrawTracks can't lock texture");
		// buffer
		unsigned int *Buf=(unsigned int*)LRect.pBits;
		// draw
		Buf[x + LDesc.Width * z] = I3D::Color::ToColor_uint(Blend);
		if(i != Levels - 1)
		{
			// Blend
			I3D::Color::Color_int4 c1 = I3D::Color::ToColor_int4(Buf[(x - x % 2    ) + LDesc.Width * (z - z % 2    )]);
			I3D::Color::Color_int4 c2 = I3D::Color::ToColor_int4(Buf[(x - x % 2 + 1) + LDesc.Width * (z - z % 2    )]);
			I3D::Color::Color_int4 c3 = I3D::Color::ToColor_int4(Buf[(x - x % 2 + 1) + LDesc.Width * (z - z % 2 + 1)]);
			I3D::Color::Color_int4 c4 = I3D::Color::ToColor_int4(Buf[(x - x % 2    ) + LDesc.Width * (z - z % 2 + 1)]);
			Blend = I3D::Color::Color_int4(
				(int)((c1.Data[I3D::Color::ALPHA] + c2.Data[I3D::Color::ALPHA] + c3.Data[I3D::Color::ALPHA] + c4.Data[I3D::Color::ALPHA]) / 4.0f),
				(int)((c1.Data[I3D::Color::RED  ] + c2.Data[I3D::Color::RED  ] + c3.Data[I3D::Color::RED  ] + c4.Data[I3D::Color::RED  ]) / 4.0f),
				(int)((c1.Data[I3D::Color::GREEN] + c2.Data[I3D::Color::GREEN] + c3.Data[I3D::Color::GREEN] + c4.Data[I3D::Color::GREEN]) / 4.0f),
				(int)((c1.Data[I3D::Color::BLUE ] + c2.Data[I3D::Color::BLUE ] + c3.Data[I3D::Color::BLUE ] + c4.Data[I3D::Color::BLUE ]) / 4.0f));
		}
		// unlock
		if(Texture->UnlockRect(i) != D3D_OK)
			throw("TextureManager::Brightness can't unlock texture");
	}
}

void TextureManager::SetPixel(LPDIRECT3DTEXTURE9 Texture, int x, int z, unsigned int Color)
{
	// get Texture width and height
	D3DSURFACE_DESC LDesc;
	if(Texture->GetLevelDesc(0, &LDesc) != D3D_OK)
		throw("TextureManager::SetPixel: Can't get description of Texture level 0");
	D3DLOCKED_RECT LRect;
	memset(&LRect, 0, sizeof(LRect));
	if(Texture->LockRect( 0, &LRect, NULL, 0 ) != D3D_OK)
		throw("TextureManager::SetPixel can't lock texture");
	// get buffer
	unsigned int *Buf=(unsigned int*)LRect.pBits;
	bool GenMips = false;
	if(Texture->GetLevelCount() != 1)
		GenMips = true;
	// set pixel
	I3D::Color::Color_int4 Blend;
	if(x >= 0 && x < (int)LDesc.Width && z >= 0 && z < (int)LDesc.Height)
	{
		unsigned int Alpha = Color >> 24;
		unsigned int Red = (Color >> 16) & 0xFF;
		unsigned int Green = (Color >> 8) & 0xFF;
		unsigned int Blue = (Color) & 0xFF;
		unsigned int OldRed = (Buf[x + LDesc.Width * z] >> 16) & 0xFF;;
		unsigned int OldGreen = (Buf[x + LDesc.Width * z] >> 8) & 0xFF;
		unsigned int OldBlue = Buf[x + LDesc.Width * z] & 0xFF;
		// Alpha
		Buf[x + LDesc.Width * z] = 0xFF << 24;
		// Red
		Buf[x + LDesc.Width * z] += (Red * Alpha + OldRed * (0xFF - Alpha)) / 0xFF << 16;
		// Green
		Buf[x + LDesc.Width * z] += (Green * Alpha + OldGreen * (0xFF - Alpha)) / 0xFF << 8;
		// Blue
		Buf[x + LDesc.Width * z] += (Blue * Alpha + OldBlue * (0xFF - Alpha)) / 0xFF;
		if(GenMips)
		{
			I3D::Color::Color_int4 c1 = I3D::Color::ToColor_int4(Buf[(x - x % 2    ) + LDesc.Width * (z - z % 2    )]);
			I3D::Color::Color_int4 c2 = I3D::Color::ToColor_int4(Buf[(x - x % 2 + 1) + LDesc.Width * (z - z % 2    )]);
			I3D::Color::Color_int4 c3 = I3D::Color::ToColor_int4(Buf[(x - x % 2 + 1) + LDesc.Width * (z - z % 2 + 1)]);
			I3D::Color::Color_int4 c4 = I3D::Color::ToColor_int4(Buf[(x - x % 2    ) + LDesc.Width * (z - z % 2 + 1)]);
			Blend = I3D::Color::Color_int4(
				(int)((c1.Data[I3D::Color::ALPHA] + c2.Data[I3D::Color::ALPHA] + c3.Data[I3D::Color::ALPHA] + c4.Data[I3D::Color::ALPHA]) / 4.0f),
				(int)((c1.Data[I3D::Color::RED  ] + c2.Data[I3D::Color::RED  ] + c3.Data[I3D::Color::RED  ] + c4.Data[I3D::Color::RED  ]) / 4.0f),
				(int)((c1.Data[I3D::Color::GREEN] + c2.Data[I3D::Color::GREEN] + c3.Data[I3D::Color::GREEN] + c4.Data[I3D::Color::GREEN]) / 4.0f),
				(int)((c1.Data[I3D::Color::BLUE ] + c2.Data[I3D::Color::BLUE ] + c3.Data[I3D::Color::BLUE ] + c4.Data[I3D::Color::BLUE ]) / 4.0f));
		}
	}
	// unlock
	if(Texture->UnlockRect(0) != D3D_OK)
		throw("TextureManager::SetPixel can't unlock texture");
	// mipmaps
	if(GenMips)
		UpdateMipmaps(Texture, x, z, Blend);
}

void TextureManager::Brightness(LPDIRECT3DTEXTURE9 Texture, int x, int z, int Brightness)
{
	// get Texture width and height
	D3DSURFACE_DESC LDesc;
	if(Texture->GetLevelDesc(0, &LDesc) != D3D_OK)
		throw("TextureManager::Brightness: Can't get description of Texture level 0");
	D3DLOCKED_RECT LRect;
	memset(&LRect, 0, sizeof(LRect));
	if(Texture->LockRect( 0, &LRect, NULL, 0 ) != D3D_OK)
		throw("TextureManager::Brightness can't lock texture");
	// buffers
	unsigned int *Buf=(unsigned int*)LRect.pBits;
	bool GenMips = false;
	if(Texture->GetLevelCount() != 1)
		GenMips = true;
	I3D::Color::Color_int4 Blend;
	if(x >= 0 && x < (int)LDesc.Width && z >= 0 && z < (int)LDesc.Height)
	{
		Buf[x + LDesc.Width * z] = I3D::Color::Brightness(Buf[x + LDesc.Width * z], Brightness);
		if(GenMips)
		{
			I3D::Color::Color_int4 c1 = I3D::Color::ToColor_int4(Buf[(x - x % 2    ) + LDesc.Width * (z - z % 2    )]);
			I3D::Color::Color_int4 c2 = I3D::Color::ToColor_int4(Buf[(x - x % 2 + 1) + LDesc.Width * (z - z % 2    )]);
			I3D::Color::Color_int4 c3 = I3D::Color::ToColor_int4(Buf[(x - x % 2 + 1) + LDesc.Width * (z - z % 2 + 1)]);
			I3D::Color::Color_int4 c4 = I3D::Color::ToColor_int4(Buf[(x - x % 2    ) + LDesc.Width * (z - z % 2 + 1)]);
			Blend = I3D::Color::Color_int4(
				(int)((c1.Data[I3D::Color::ALPHA] + c2.Data[I3D::Color::ALPHA] + c3.Data[I3D::Color::ALPHA] + c4.Data[I3D::Color::ALPHA]) / 4.0f),
				(int)((c1.Data[I3D::Color::RED  ] + c2.Data[I3D::Color::RED  ] + c3.Data[I3D::Color::RED  ] + c4.Data[I3D::Color::RED  ]) / 4.0f),
				(int)((c1.Data[I3D::Color::GREEN] + c2.Data[I3D::Color::GREEN] + c3.Data[I3D::Color::GREEN] + c4.Data[I3D::Color::GREEN]) / 4.0f),
				(int)((c1.Data[I3D::Color::BLUE ] + c2.Data[I3D::Color::BLUE ] + c3.Data[I3D::Color::BLUE ] + c4.Data[I3D::Color::BLUE ]) / 4.0f));
		}
	}
	// unlock
	if(Texture->UnlockRect(0) != D3D_OK)
		throw("TextureManager::Brightness can't unlock texture");
	// mipmaps
	if(GenMips)
		UpdateMipmaps(Texture, x, z, Blend);
}

void TextureManager::Spot(LPDIRECT3DTEXTURE9 Texture, int x, int z, int Size, int BrightnessValue)
{
	for(int i = -Size; i <= Size; i++)
		for(int j = -Size; j <= Size; j++)
		{
			//int Bright = (Size - abs(i)) * (Size - abs(j)) * BrightnessValue / (Size * Size);
			//if(Bright * BrightnessValue > 0) // if Bright and Brightness have the same sign
				Brightness(Texture, x + i, z + j, BrightnessValue);
		}
}

}
