/***********************************************************/
/*                                                         */
/* TextureLoader.h                                         */
/*                                                         */
/* created 24.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef TEXTUREMANAGER_H
#define TEXTUREMANAGER_H

#include "D3DException.h"
#include "D3DStructures.h"
#include "IniLoader.h"
#include "DynamicArray.h"
#include "D3DStruct.h"
#include "Color.h"

#include <d3d9.h>
#include <stdio.h>

// www.infinitecode.com 3d library
namespace I3D
{

enum
{
	// create alphamap with colorkey 0x000000
	TM_FLAG_ALPHACC0 = 1,
	// optimize alpha color keyed textures for bilinear filtering
	TM_FLAG_ACCBILOPT = 2,
	// ignore texels with alpha = 0 when calculation mipmaps
	TM_FLAG_MIPALPHAIGN = 4,
	// don't allow nonzero alpha < 0x60 in high-level mipmaps
	TM_FLAG_MIPALPHA60 = 8,
	// load flags from Textures\\Textures.ini
	// TODO: implement
	TM_FLAG_FLAGSFROMINI = 16,
	// make high mipmaps gray (used for detail texturing)
	TM_FLAG_GRAYMIPMAPS = 32,
	// color key FF00FF
	TM_FLAG_ALPHACCFF00FF = 64,
	// set alpha to constant value
	TM_FLAG_MASK = 512 + 256 + 128,
	TM_FLAG_ALPHA00 = 128,
	TM_FLAG_ALPHA20 = 256,
	TM_FLAG_ALPHA40 = 256 + 128,
	TM_FLAG_ALPHA60 = 512,
	TM_FLAG_ALPHA80 = 512 + 128,
	TM_FLAG_ALPHAC0 = 512 + 256,
	TM_FLAG_ALPHAFF = 0 // default
};

class TextureManager
{
	class LoadedTexture
	{
	public:
		LoadedTexture(void)
			:Name(0)
		{
		}

		LoadedTexture(const char *Name, LPDIRECT3DTEXTURE9 Texture)
		{
			LoadedTexture::Name = new char[1024];
			strcpy(LoadedTexture::Name, Name);
			LoadedTexture::Texture = Texture;
		}

		~LoadedTexture(void)
		{
			delete[] Name;
		}

		LoadedTexture& operator = (LoadedTexture &LT)
		{
			if(!Name)
				Name = new char[1024];
			strcpy(Name, LT.Name);
			Texture = LT.Texture;
			return *this;
		}

		LoadedTexture(LoadedTexture& LT)
		{
			*this = LT;
		}

		char *Name;
		LPDIRECT3DTEXTURE9 Texture;

	};

	// static class, so no instances are allowed
	TextureManager(void) {}

	// fills black colorkeyed pixels with neighboring colors
	static void BilinearOptimization(unsigned int *Map, int Width, int Height);

public:
	class TextureManagerException: public D3DException
	{
	public:
		TextureManagerException(char *Description)
			:D3DException(Description)
		{
			AddPrefix("Texture Manager: ");
		}
	};

	static ITL::IniLoader Ini;

	static ISL::DynamicArray<LoadedTexture, 0, 1> LoadedTextures;

	static void Clear(void)
	{
		for(int i = 0; i < LoadedTextures.Count(); ++i)
			RELEASE(LoadedTextures[i].Texture);
		LoadedTextures.Clear();
	}

	// creates all mipmap levels for the texture
	static void CreateMipMaps(LPDIRECT3DTEXTURE9 Texture, int Flags);
	// loads a texture from a bmp file (Direct3D X functions are evil ;-) )
	static LPDIRECT3DTEXTURE9 Load(LPDIRECT3DDEVICE9 D3DDevice, const char *FileName, const char *AlphaFileName, int Flags);
	// loads a texture with the parameters in Textures\\Textures.ini
	static LPDIRECT3DTEXTURE9 GetTexture(D3DStruct *DS, const char *Name, const char *AlphaName = 0, int Flags = 0);
	static void UpdateMipmaps(LPDIRECT3DTEXTURE9 Texture, int x, int z, I3D::Color::Color_int4 Blend);
	static void SetPixel(LPDIRECT3DTEXTURE9 Texture, int x, int z, unsigned int Color);
	static void Brightness(LPDIRECT3DTEXTURE9 Texture, int x, int z, int Brightness);
	static void Spot(LPDIRECT3DTEXTURE9 Texture, int x, int z, int Size, int BrightnessValue);


};

}

#endif
