/***********************************************************/
/*                                                         */
/* World.h                                                 */
/*                                                         */
/* created 16.07.2002                                      */
/* by Christian Reimann  empgodot@gmx.de                   */
/*                                                         */
/* Copyright    Christian Reimann                          */
/*              empgodot@gmx.de                            */
/*              www.infinitecode.com                       */
/*                                                         */
/* This code may be freely modified and redistributed.     */
/* If you use parts of this code in a project,             */
/* I'd appreciate a mention in the credits.                */
/*                                                         */
/***********************************************************/

#ifndef WORLD_H
#define WORLD_H

#include "D3DStruct.h"
#include "TankLib.h"
#include "QuadTree.h"
#include "Renderable.h"
#include "Terrain.h"
#include "DynamicArray.h"
#include "Tree.h"
#include "PhysicsStruct.h"
#include "SpawnInterface.h"
#include "ParticleEffect.h"
#include "ParticleFXRenderer.h"
#include "WorldPFXInterface.h"
#include "MissionObject.h"
#include "ProjectileLib.h"
#include "Projectile.h"
#include "Tank.h"
#include "Way.h"
#include "LUT2D.h"
#include "TerrainLine.h"
#include "TerrainLineRenderer.h"
#include "SoundManager.h"
#include "AccMapRenderer.h"
#include "MOPropertiesIf.h"
#include "Airfield.h"
#include "Decal.h"
#include "DecalLib.h"
#include "Minimap.h"
#include "EventSender.h"
#include "ForceFeedbackInterface.h"

// www.infinitecode.com game library
namespace IGAME
{

#define GAME_TREES	512

class World: public IEV::EventSender<const char *>, public I3D::ForceFeedbackInterface
{
public:
	typedef ISL::DynamicArray<Tank *> SelectionType;

	enum
	{
		NOTHING = 0,
		STREET_VERT = 1,
		STREET_HORZ = 2,
		STREET_CROSS = 3,
		WAY = 4,
		STREET_MASK = 7,
		TERRAIN,
		VEHICLE,
		DECAL,
		AIRFIELD
	};

	class Plain
	{
		friend World;

		IMATH::Vector2DXZ Pos;
		float Size;

	public:
		Plain() {}
		Plain(IMATH::Vector2DXZ &V, float f)
			:Pos(V), Size(f) {}
	};

	class WorldSpecs
	{
	public:
		WorldSpecs(void): DS(0), SndMgr(0), MOPropIf(0), Seed(0), EditMode(false) {}

		I3D::D3DStruct *DS;
		ISL::DynamicArray<I3D::Way *> Ways;
		ISL::DynamicArray<I3D::TerrainDsc *> TerrainDscs;
		ISL::DynamicArray<IGAME::Airfield *> Airfields;
		ISL::DynamicArray<IGAME::Decal *> Decals;
		ISL::DynamicArray<Plain> Plains;
		IAU::SoundManager *SndMgr;
		MOPropertiesIf *MOPropIf;
		int Seed;
		bool EditMode;
	};

	enum MouseTargets
	{
		MOUSE_TARGET_NOTHING = 0,
		MOUSE_TARGET_IMPASSABLE = 1,
		MOUSE_TARGET_FRIEND = 2,
		MOUSE_TARGET_ENEMY = 3,
	};

private:
	bool Initialized;
	WorldSpecs Specs;
	float Size;
	// world objects
	I3D::QuadTree QTree;
	I3D::Terrain Surface;
	I3D::LUT2D LUT;
	// actors
	ISL::DynamicArray<IGAME::Tree *, 0, 256> Trees;
	ISL::DynamicArray<IGAME::Tank *, 0, 256> Tanks;
	ISL::DynamicArray<IGAME::Tank *> TankDeleteList;
	ISL::DynamicArray<IGAME::Projectile *, 0, 256> Projectiles;
	ISL::DynamicArray<IGAME::Projectile *, 0, 256> ProjectileDeleteList;
	ISL::DynamicArray<IGAME::Airfield *> Airfields;
	//ISL::DynamicArray<IGAME::Decal *> Decals;
	ISL::DynamicArray<I3D::TerrainLine *> TerrainLines;
	SelectionType Selection;
	// camera
	I3D::Vector CameraPos, CameraAngle;
	// particle effects
	ISL::DynamicArray<I3D::ParticleEffect *> PFX;
	ISL::DynamicArray<I3D::ParticleEffect *> PFXDeleteList;
	I3D::ParticleFXRenderer PFXRenderer;
	I3D::ParticleFXLib PFXLib;
	I3D::TerrainLineRenderer TLR, TLRAcc;
	TankLib TLib;
	ProjectileLib ProjLib;
	DecalLib DLib;
	int RenderFlags;
	// Accessibility Map Renderer
	I3D::AccMapRenderer AccRnd;
	// lines for terraindscs in missioneditor, ...
	ISL::DynamicArray<I3D::TerrainLine *> Lines;
	bool DrawMELines;
	Minimap MMap;
	bool MMapDrag;
	ITL::Profiler Prf;
	ITL::Profiler PhysicsPrf;
	// force feedback
	IMATH::Vector ShakeVector;
	// drag box
	I3D::RenderStruct<I3D::Vertex, I3D_FVFVERTEX, 16, 16> DragRS;
	LPDIRECT3DTEXTURE9 DragTexture;
	IMATH::Vector2D DragPos[2];
	enum DragStates
	{
		DRAGGING_NOTHING,
		DRAGGING_VEHICLE,
		DRAGGING_WAY,
		DRAGGING_TERRAIN,
		DRAGGING_AIRFIELD,
		DRAGGING_DECAL,
		DRAGGING_AREA,
		DRAGGING_MARK
	};
	struct DragData
	{
		int Type, ID;
		Tank *Vehicle;
		I3D::Way *Way;
		I3D::TerrainDsc *Terrain;
	} Dragging;

	void CreateView(I3D::D3DStruct *DS);
	void CameraConstraints();
	void KillTrees(const IMATH::Vector2DXZ &From, const IMATH::Vector2DXZ &To);
	void CreateQuadTree(World::WorldSpecs *WS);
	void CreateTrees(World::WorldSpecs *WS);
	void CreateMinimap();

	// interface for spawning, deleting and accessing actors
	class CISpawn: public SpawnInterface
	{
		friend World;

		World *Wrld;

	public:
		virtual void Spawn(SpawnInterfaceSpecs *SS);
		virtual void CISpawn::GetNearObjects(const IMATH::Vector &Pos, const float Area, I3D::ManagedObject::MgdObjArray *Objects) const;
		virtual void CISpawn::GetNearObjects(const IMATH::Vector &Pos, const float Area, I3D::ManagedObject::MgdObjArray *Objects, const void *Type) const;
		//virtual void World::CISpawn::GetNearTanks(const IMATH::Vector &Pos, const float Area, ISL::DynamicArray<Tank *> *Tanks);
		virtual void DeleteTank(Tank *Tnk);
		virtual void DeleteProjectile(Projectile *Prj);
		//virtual void Display(SpawnInterfaceSpecs *SS);
	} ISpawn;

	class CIWorldPFX: public I3D::WorldPFXInterface
	{
		friend World;

		World *Wrld;

	public:
		virtual void Delete(I3D::ParticleEffect *PFX);
	} IWorldPFX;

	friend CIWorldPFX;

	bool DrawAccMap;

public:

	World(): Initialized(false) {}

	friend CISpawn;

	void SetProfShow(bool Value)
	{
		Prf.SetShow(Value);
	}

	bool GetProfShow(void)
	{
		return Prf.GetShow();
	}

	void SetPhysicsProfShow(bool Value)
	{
		PhysicsPrf.SetShow(Value);
	}

	bool GetPhysicsProfShow(void)
	{
		return PhysicsPrf.GetShow();
	}

	SpawnInterface *GetSpawnInterface(void)
	{
		return &ISpawn;
	}

	IAU::SoundManager *GetSoundManager(void)
	{
		return Specs.SndMgr;
	}

	float GetSize(void)
	{
		return Size;
	}

	bool GetDrawMELines()
	{
		return DrawMELines;
	}

	I3D::TerrainInterface *GetTerrainInterface()
	{
		return &Surface;
	}

	I3D::TerrainLineRenderer *GetTerrainLineRenderer()
	{
		return &TLR;
	}

	I3D::TerrainLineRenderer *GetTerrainLineRendererAcc()
	{
		return &TLRAcc;
	}

	float GetSurfaceY(float x, float z)
	{
		return Surface.GetY(x, z);
	}

	IMATH::Vector GetSurfaceNormal(float x, float z)
	{
		return Surface.GetNormal(x, z);
	}

	IMATH::Vector2DXZ GetSurfaceXZ(const IMATH::Vector2D &Pos)
	{
		return QTree.GetXZPos(Pos).Pos;
	}

	SelectionType GetSelection(void)
	{
		return Selection;
	}

	void SetQTreeProfShow(bool Value)
	{
		QTree.SetProfShow(Value);
	}

	bool GetQTreeProfShow(void)
	{
		return QTree.GetProfShow();
	}

	const I3D::Vector &GetCameraPos() const
	{
		return CameraPos;
	}

	const I3D::Vector &GetCameraAngle() const
	{
		return CameraAngle;
	}

	void SetCameraPos(const IMATH::Vector &V)
	{
		CameraPos = V;
		CameraConstraints();
	}

	void SetCameraAngle(const IMATH::Vector &A)
	{
		CameraAngle = A;
		CameraConstraints();
	}

	bool GetDrawAccMap(void)
	{
		return DrawAccMap;
	}

	void SetDrawAccMap(bool Value)
	{
		DrawAccMap = Value;
	}

	void SetDrawMELines(bool Value)
	{
		DrawMELines = Value;
		CreateLines();
	}

	ISL::DynamicArray<I3D::Way> *GetWays()
	{
		return Surface.GetWays();
	}

	~World()
	{
		Destroy();
	}

	void CreateHMap(int Seed)
	{
		Specs.Seed = Seed;
		Surface.CreateHMap(Seed);
		//CompileDecals();
		CreateQuadTree(&Specs);
		CreateTrees(&Specs);
		CreateLines();
		CreateMinimap();
	}

	void MinimapUpdate()
	{
		MMap.Update();
	}

	void SplitWay(const IMATH::Vector2DXZ &Pos)
	{
		Surface.SplitWay(Pos);
	}

	virtual void SetShakeVector(const IMATH::Vector &Shake)
	{
		ShakeVector = Shake;
	}

	virtual void AddShakeVector(const IMATH::Vector &Pos, const IMATH::Vector &Shake)
	{
		float Dist = 0.1f * (CameraPos - Pos).Length();
		if(Dist < 1.0f)
			Dist = 1.0f;
		if(Dist)
			ShakeVector += Shake / Dist;
	}

	void RenderPhysicsProfiler(I3D::D3DStruct *DS)
	{
		PhysicsPrf.Render(DS);
	}

	virtual void Render(I3D::D3DStruct *DS);
	void Init3D(I3D::D3DStruct *DS, bool recurse);
	void Create(WorldSpecs *Specs);
	void Destroy(void);
	void PhysicsPFX(I3D::D3DStruct *DS);
	void Physics(I3D::D3DStruct *DS);
	void ManageDeletionLists(void);
	void InitLight(I3D::D3DStruct *DS);
	void DestroyLight(I3D::D3DStruct *DS);
	void OnViewChange(void);
	void AddWay(I3D::Way::WaySpecs *WS);
	void AddWayNoHMap(const I3D::Way &W);
	void AddWayNoHMap(I3D::Way::WaySpecs *WS);
	void AddAirfieldPre(Airfield *AF);
	void AddAirfield(Airfield *AF);
	void AddDecalPre(Decal *D);
	void CompileDecalsAdditive();
	void CompileDecals();
	void World::UpdateSurfaceTextures();
	void AddDecal(Decal *D);
	void AddTerrain(I3D::TerrainDsc::TerrainDscSpecs *TS);
	void AddTerrainNoHMap(I3D::TerrainDsc::TerrainDscSpecs *TS);
	int GetTankNumber(int Owner);
	int GetTankNumber(int Owner, const IMATH::Vector2DXZ *Area);
	void CreateLines();

	void MoveCamera(const I3D::Vector &V);
	void MoveCameraDirect(const I3D::Vector &V);
	void RotateCamera(const I3D::Vector &V);
	void Select(I3D::BoundingBox2D *BBox);
	void Click(const IMATH::Vector2D &Pos);
	void RightClick(const IMATH::Vector2D &Pos);
	void Drag(const IMATH::Vector2D &Pos, const IMATH::Vector2D &Pos2);
	void DragStart(const IMATH::Vector2D &Pos);
	void DragEnd(const IMATH::Vector2D &Pos, const IMATH::Vector2D &Pos2);
	int GetMouseTarget(const IMATH::Vector2D &Pos);
	bool GetStreetAccessible(const IMATH::Vector2DXZ &Pos);
	int GetLandscape(const IMATH::Vector2DXZ &Pos);
	Decal *GetDecal(const IMATH::Vector2DXZ &Pos);
	IMATH::Vector2DXZ GetXZMousePos(const IMATH::Vector2D &Pos);
	Tank *GetTank(const char *Name);
	Tank *World::GetTankAtPos(const IMATH::Vector2DXZ &Pos);
};

}

#endif
